/*
 * Copyright © 2006 - 2007 Claudio Corsi, Paolo Ferragina
 *
 * This file is part of SmallText library.
 *
 * SmallText is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SmallText is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package it.unipi.di.graph;

import it.unipi.di.textdb.TextDB;
import it.unipi.di.util.Range;

import java.io.IOException;
import java.util.Arrays;
import java.util.Iterator;
import java.util.Random;
import java.util.SortedMap;
import java.util.Map.Entry;

/**
 * A graph labeler for vertices and edges. 
 * The graph is represented as a list of N vertices identified by integers in the
 * range [0, N-1] and a list of edges identified by the two vertices they connect. 
 * Each vertex can be associated with a <i>textual label</i>, and each edge can 
 * be associated with <i>one or more</i> textual attributes, each identified by a 
 * unique name. This class offers access to these informations by means of a 
 * {@link GraphLabelerConfig} object, which defines how the labels and the 
 * attribute values are stored on disk. Concrete access to these data is implemented 
 * by means of the {@link TextDB} interface.
 * <br><br>
 * We note that the structure of the graph, namely its vertices and edges, is 
 * not stored in this class.
 *  
 * @see GraphLabelerConfig 
 * @see TextDB
 * 
 * @author Claudio Corsi
 * @author Paolo Ferragina
 * 
 */
public class GraphLabeler {
	
	private GraphLabelerConfig conf;
	
	/**
	 * Create a new GraphLabeler using the provided configuration.
	 * 
	 * @param conf the configuration defining the data sources.
	 * @throws IOException
	 */
	public GraphLabeler(GraphLabelerConfig conf) throws IOException {
		this.conf = conf;
		
		open();
	}
	
	/**
	 * Create a new GraphLabeler loading its configuration from
	 * a textual file.
	 * 
	 * @see GraphLabelerConfig#load(String)
	 * 
	 * @param confFile the name of the configuration file
	 * @return a new GraphLabeler object
	 * @throws IOException
	 */
	public static GraphLabeler load(String confFile) throws IOException {
		GraphLabelerConfig conf = GraphLabelerConfig.load(confFile);
		return new GraphLabeler(conf);
	}
	
	/**
	 * Returns the current configuration.
	 * 
	 * @return the current graph configuration.
	 */
	public GraphLabelerConfig getConfig() {
		return conf;
	}

	/**
	 * This is the preliminary operation to open, and then use, this labeler.
	 * 
	 * @throws IOException
	 */
	private void open() throws IOException {
		conf.open();
	}
	
	/**
	 * Closes the labeler and releases all the resources associated to it.
	 * 
	 * @throws IOException
	 */
	public void close() throws IOException {
		conf.close();
	}
	
	/**
	 * Returns the label associated to the input vertex, or null if the label is not defined. 
	 * 
	 * @param v vertex id.
	 * @return the label for the requested vertex.
	 * @throws IOException if some I/O error occur.
	 * @throws ArrayIndexOutOfBoundsException if the requested vertex is outside the current configuration.
	 */
	public String getVertexLabel(int v) throws IOException {
		Range span = conf.getSpanningRange();
		if (!span.containsPoint(v)) throw new ArrayIndexOutOfBoundsException("Out of bound vertex id: " + v);
		Range r = conf.getRangeFor(v);
		TextDB db = conf.getVertexDB(v);
		return (db != null) ? db.get(v - r.getLowerBound()) : null;
	}
	
	/**
	 * Returns the labels for the vertices in range <code>[i, j]</code>
	 * (bounds included). The returned array will has the same length
	 * of the passed range. The array will contains null references
	 * for vertices not defined in this
	 * labeler. The requested range must be positive with a lower bound greater or equals than
	 * zero and an upper bound no bigger than the defined maximum position for this labeler.
	 * The bounds are checked by this method and an {@link ArrayIndexOutOfBoundsException}
	 * will be thrown in case the bounds doesn't respect such limits.
	 *  
	 * 
	 * @see GraphLabelerConfig#getSpanningRange()
	 * 
	 * @param i the lower bound of the range.
	 * @param j the upper bound of the range.
	 * @return the labels for the vertices in range <code>[i, j]</code>.
	 * @throws IOException if some I/O errors occur.
	 * @throws ArrayIndexOutOfBoundsException in case the requested ids are out of the bounds
	 * defined in the current configuration.
	 */
	public String[] getVertexLabels(int i, int j) throws IOException {
		
		Range span = conf.getSpanningRange();
		
		if (i < span.getLowerBound()) throw new ArrayIndexOutOfBoundsException("Out of bound range: " + i + " < " + span.getLowerBound());
		if (j > span.getUpperBound()) throw new ArrayIndexOutOfBoundsException("Out of bound range: " + j + " > " + span.getUpperBound());

		Range r = new Range(i, j);	// the requested range

		SortedMap<Range, TextDB> map = conf.getRangeDBMap(r);
		
		String[] res = new String[r.length()];
		int pos = 0;
		Iterator<Entry<Range, TextDB>> iter = map.entrySet().iterator();
		
		while(iter.hasNext()) {
			Entry<Range, TextDB> e = iter.next();
			Range r1 = e.getKey();
			TextDB db = e.getValue();
			
			int max = Math.max(r1.getLowerBound(), i);
			int min = Math.min(r1.getUpperBound(), j);
			
			Range range = new Range(min, max);
			
			pos = range.getLowerBound() - r.getLowerBound();
			
			int start = range.getLowerBound() - r1.getLowerBound();
			int stop = range.getUpperBound() - r1.getLowerBound();
			String[] tmp = db.getRange(start, stop);
			
			System.arraycopy(tmp, 0, res, pos, tmp.length);
		}
		
		return res;
	}
	
	/**
	 * Returns the labels for a <i>sorted</i> list of vertices. 
	 * 
	 * @param vertices sorted list of vertex ids.
	 * @return the labels for those vertices.
	 * @throws IOException if some I/O errors occur.
	 * @throws ArrayIndexOutOfBoundsException in case the requested ids are out of the bounds
	 * defined in the current configuration.
	 */
	public String[] getVertexLabels(int[] vertices) throws IOException {
		return getVertexLabels(vertices, 0, vertices.length);
	}
	
	/**
	 * Returns the labels for a <i>sorted</i> list of vertices. The passed
	 * array is considered from position <code>offset</code> to 
	 * <code>length-1</code>. A {@link ArrayIndexOutOfBoundsException} will be
	 * thrown if the input vertex ids are out of the bounds of the current configuration.
	 * <br>
	 * The returned array of StringS will contains exactly length references. Null references
	 * are used in case no labels are defined for the corresponding requested vertex.
	 * 
	 * @see GraphLabelerConfig#getSpanningRange()
	 * 
	 * @param vertices sorted list of vertex ids (at least in the positions from <code>0</code> to <code>length</code>).
	 * @param offset the starting position of the input array.
	 * @param length the number of positions to consider in the passed array starting 
	 * from 0. Must be non negative.
	 * @return the labels for those vertices.
	 * @throws IOException if some I/O errors occur.
	 * @throws ArrayIndexOutOfBoundsException in case the requested ids are out of the bounds
	 * defined in the current configuration.
	 */
	public String[] getVertexLabels(int[] vertices, int offset, int length) throws IOException {
		
		Range r = new Range(vertices[0], vertices[length - 1]);	// the requested range
		
		Range span = conf.getSpanningRange();

		if (r.getLowerBound() < span.getLowerBound()) throw new ArrayIndexOutOfBoundsException("Out of bound range: " + r.getLowerBound() + " < " + span.getLowerBound());
		if (r.getUpperBound() > span.getUpperBound()) throw new ArrayIndexOutOfBoundsException("Out of bound range: " + r.getUpperBound() + " > " + span.getUpperBound());
		
		String[] res = new String[length];
		
		SortedMap<Range, TextDB> map = conf.getRangeDBMap(span);
		Iterator<Entry<Range, TextDB>> iter = map.entrySet().iterator();
		
		// no ranges has been defined for this graph labeler, return a list of null
		if (!iter.hasNext()) return res;
		
		int[] tmp = new int[length];
		int k = 0;
		int resPos = 0;
		
		Entry<Range, TextDB> e = iter.next();
		Range r1 = e.getKey();
		TextDB db = e.getValue();
		
		for(int pos = offset; pos < length; pos++) {
			
			if (r1.containsPoint(vertices[pos]))
				tmp[k++] = vertices[pos] - r1.getLowerBound();  // id relative to the current TDB
			else {
				
				if (k > 0) {
					
					String[] labels = db.getSequential(tmp, 0, k);
					
					System.arraycopy(labels, 0, res, resPos, labels.length);
					
					k = 0;
					pos--;
					resPos += labels.length;
					
					if (iter.hasNext()) {
						e = iter.next();
						r1 = e.getKey();
						db = e.getValue();
					}
					else return res;	// finish, no more ranges to check
				}
				else 
					resPos++;
			}
		}

		// last one
		if (k > 0) {
			String[] labels = db.getSequential(tmp, 0, k);
			System.arraycopy(labels, 0, res, resPos, labels.length);
		}
		
		return res;
	}
	
	/**
	 * Returns a String object identifying the type of a vertex. Vertices
	 * of the same type will returns the same String object (i.e. String objects
	 * resulting equal respect the "equals()" method). 
	 * <br><br>
	 * The {@link TextDB} storing the vertex label determines
	 * the type of the vertex. Two vertices with labels stored into the same {@link TextDB}
	 * will result of the same type.
	 * 
	 * @see TextDB#getName()
	 * 
	 * @param v a vertex ID
	 * @return a String representing the vertex type or null if the vertex is not labeled
	 * @throws IOException
	 */
	public String getVertexType(int v) throws IOException {
		TextDB tdb = conf.getVertexDB(v); 
		return tdb == null ? null : tdb.getName();
	}
	
	/**
	 * Returns the attribute value of an edge. The edge is defined
	 * by means of the pair <i>(v, rank)</i>  where v is the source vertex 
	 * of the edge and rank is the cardinal position of the target vertex of this 
	 * edge among the neighbors of v (counting from 0). For instance, if v has 
	 * neighbors u,w,z, then (v,2) identifies (v,z). 
	 * A null value is returned in case the pair (v,rank) does not define any edge.
	 * <br> 
	 * @see #getOutgoingAttribute(int, String)
	 * 
	 * @param v source vertex of the edge
	 * @param rank the rank of the target node defining the edge among 
	 * the neighbors of v
	 * @param attribute the attribute name
	 * @return the value for the requested attribute of a given edge
	 * @throws IOException
	 */
	public String getEdgeAttribute(int v, int rank, String attribute) throws IOException {
		TextDB db = conf.getEdgeDB(attribute);
		return db.get(v, rank);
	}
	
	/**
	 * Returns the attribute values defined for the outgoing edges starting from 
	 * a vertex. That values will be returned ordered accordingly to the
	 * adjacency list of the vertex.
	 * 
	 * @param vertex the starting vertex
	 * @param attribute the name of the attribute to extract from the outgoing edges 
	 * @return the attribute values for all the outgoing edges
	 * @throws IOException
	 */
	public String[] getOutgoingAttribute(int vertex, String attribute) throws IOException {
		TextDB db = conf.getEdgeDB(attribute);
		String record = db.get(vertex);
		return db.getRecordFields(record);
	}
	
	/**
	 * Given a multi-value attribute, this method returns its individual
	 * values in an array of strings.
	 * <br>
	 * <br>
	 * This implementation is based on 
	 * the method {@link TextDB#getFieldValues(String, String)}.
	 * 
	 * @param attribute	the attribute to be parsed
	 * @param sep the character-sequence used to separate the values in the input attribute
	 * @return the array of values
	 */
	public String[] splitAttributeValues(String attribute, String sep) {
		return conf.getEdgeDB(attribute).getFieldValues(attribute, sep);
	}
	
	public static void main(String[] args) throws Exception {
		
		int i = 0;
		String filename = args[i++];
		
		GraphLabelerConfig conf =  GraphLabelerConfig.load(filename);
		GraphLabeler labeler = new GraphLabeler(conf);
		
		int[] vertices = new int[10];
		for(int v = 0; v < vertices.length; v++) vertices[v] = 4811645 + v;
		
		// shuffle
		Random rnd = new Random();
		for(int v = 0; v < vertices.length; v++) {
			int pos = rnd.nextInt(vertices.length);
			int tmp = vertices[v];
			vertices[v] = vertices[pos];
			vertices[pos] = tmp;
		}

		System.out.println(" >> VERTEX LABELS TEST 1: random access");
		
		for(int k = 0; k < vertices.length; k++) {
			int v = vertices[k];
			String label = labeler.getVertexLabel(v);
			System.out.println(v + ": " + label);
		}
		System.out.println();
		
		System.out.println(" >> VERTEX LABELS TEST 2: sequential access");
		Arrays.sort(vertices);
		String[] labels = labeler.getVertexLabels(vertices, 0, vertices.length);
		int v = 0;
		for (String label : labels) {
			System.out.println(vertices[v++] + ": " + label);
		}
		System.out.println();
		
//		System.out.println(" >> EDGE FIELD VALUES TEST");
//		
//		rnd = new Random();
//		for(v = 0; v < 2000; v++) {
//			int trgt = rnd.nextInt(10);
//			String value = labeler.getEdgeAttribute(v, trgt, "weight");
//			System.out.println("(v=" + v + ", rank=" + trgt + ")\t" + value);
//		}
	}
}
