/*
 * Copyright ©  2006 - 2007 Claudio Corsi, Paolo Ferragina
 *
 * This file is part of SmallText library.
 *
 * SmallText is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SmallText is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package it.unipi.di.graph;

import it.unipi.di.textdb.TextDB;
import it.unipi.di.util.Range;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Set;
import java.util.SortedMap;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Map.Entry;

/**
 * Configuration class for a {@link GraphLabeler}. Objects of this class
 * are used to store information about how the vertex labels and
 * the edge-attributes values are stored on disk.
 * <br><br>
 * Once a configuration is defined, it can be saved in a permanent way 
 * in a text file using the method {@link #store(String)}. A previous 
 * configuration can be loaded using
 * the static method {@link #load(String)}. The configuration file is 
 * written in a text form and can be easily modified with a text editor. 
 * 
 * <h1>Vertex Labels</h1>
 * The main assumption is that vertices are identified by <i>id</i> which are 
 * integers in the range [0, N-1], where N is the total number of 
 * vertices in the underlying graph. Labels can be associated to vertices by
 * creating a textual file containing one label per line, so that 
 * the n-th line contains the label for the vertex id <i>n</i>. 
 * The labels can then be efficiently associated to their vertices by using an 
 * instance of a {@link TextDB}. A TextDB stores on disk 
 * a textual file in a compressed way and offers fast access to its 
 * records (i.e. lines, and thus labels) 
 * without the need to uncompress the entire file. Among all implementations of a TextDB,
 * the programmer can choose the one that best fits the type of vertex labels. Then one 
 * can invoke a <i>build</i> method, or issue
 * it from the command line. Once the build stage
 * is complete the plain textual file can be thrown away.
 * <br><br>
 * For maximum flexibility it is possible to define multiple labels 
 * subranges (<i>not</i> overlapping)
 * and associate to them different types of TextDBs. This gives the possibility to choose
 * the best TextDB (in time ad space) considering the nature of the textual labels.  
 * Indeed, given a subrange [i, j] of vertices 
 * (possibly i=0 and j=N-1), one has to form first a textual file containing their 
 * labels, then build a TextDB 
 * on this file, and finally use the method {@link #setVertexDB(Range, TextDB)} to tie this TextDB
 * with its vertex subrange.<br>
 * <br>
 * 
 * <h1>Edges Attributes</h1>
 * The information relative to the edges can be divided into attributes, each identified by a 
 * unique name. The definition
 * of an attribute is global or, in other words, an attribute is defined over all edges of 
 * the graph. Using the method
 * {@link #setEdgeDB(String, TextDB)} it is possible to define an attribute (name) and declare 
 * the TextDB object used to access the values of that attribute. As in the case of vertex labels, 
 * the attribute values are stored on disk and 
 * accessed using a TextDB. This means that first a textual file containing the attribute values 
 * must be built, and then a (proper) TextDB has to be constructed on this file. Once the building
 * stage is over, the plain textual file can be thrown away.
 * <br><br>
 * As the TextDB provides access to lines (i.e. records) of the underlying file, a convention must 
 * be defined
 * in order to store the edges values. Suppose you want to define the attribute "weight" over the 
 * edges of your graph. A textual file called "weight" should be created where at the i-th line you 
 * store the weight-values for the outgoing edges of vertex "i". If K is the number of these outgoing 
 * edges, then K different values should be stored in this line, separated by a special sequence of chars
 * called the "attributes separator" (by default the single char '\t'). The attribute values stored in a 
 * line (or record) should be placed accordingly to the numerical order of the neighbors of "i".
 * For example suppose you are defining the weights w1, w2 and w4 for the edges (i, 1), (i, 2) and (i, 4). 
 * Then the textual file has to store at line i the string: w1 + '\t' + w2 + '\t' + w4.  
 * <br>
 * If you want to define values for a multi-attribute (that is, if you have more than one edge 
 * connecting the same source and target nodes) then you need to put these values separated 
 * by a single space (i.e. '\b') into one string which will form the weight-attribute corresponding to 
 * that edge. For instance, if you have two edges connecting the vertices i and j where j is the 
 * k-th successors of i, then you have to define the string w1 + '\b' + w2 and put it into the "weight" 
 * attribute file at i-th row and k-th position.  
 * 
 * <h1>Configuration File</h1>
 * Another way to bind TextDBs to vertex ranges or edge attributes is to define a configuration file
 * and load it using the {@link #load(String)} method. The syntax to use is
 * simple and intuitive. For instance, suppose to have two label ranges [0, 99] and [100, 199]
 * for a graph of 200 vertices and two edge attributes, namely "weight" and "color". Suppose to
 * store that informations in four TextDBs stored respectively into the files "label0.tdb",   
 * "labels1.tdb", "weight.tdb" and "color.tdb". Then the configuration file will be:
 * 
 * <blockquote>
 * <code>
 * [0, 99]: label0.tdb<br> 
 * [100, 199]: label1.tdb<br>
 * weight: weight.tdb<br>
 * color: color.tdb
 * </code>
 * </blockquote>
 * 
 * The first two lines defines two vertex ranges through the usual syntax <code>[i, j]</code> and 
 * bind to them the TextDBs storing its labels. The same happens in the last two lines, but
 * the values stored in the referred TextDBs are the values of the edge attributes "weight" and 
 * "color". Note that range definitions and edge attribute names are separated from the TextDB
 * file name by a ":" character. Remember also that the defined ranges cannot be overlapping and
 * the attribute names must be unique.
 * 
 * @author Claudio Corsi
 * @author Paolo Ferragina
 *
 */
public class GraphLabelerConfig {

	protected SortedMap<Range, TextDB> vertices = new TreeMap<Range, TextDB>();
	protected SortedMap<String, TextDB> edges = new TreeMap<String, TextDB>();
	
	/**
	 * Creates a new configuration.
	 */
	public GraphLabelerConfig() {}
	
	/**
	 * Returns true if the range [i, j] has been already defined as vertex range in this configuration. 
	 * 
	 * @param i lower bound of the range
	 * @param j upper bound of the range
	 * @return true if the range [i, j] has already associated a TextDB in this 
	 * configuration, false otherwise.
	 */
	public boolean vertexRangeAlreadyDefined(int i, int j) {
		Range r = new Range(i, j);
		return vertices.get(r) != null;
	}
	
	/**
	 * Returns true if the given edge attribute has been already configured. 
	 * 
	 * @param attribute the edge attribute to check
	 * @return true if this edge attribute has already a TextDB associated to it, false otherwise.
	 */
	public boolean attributeAlreadyDefined(String attribute) {
		return edges.get(attribute) != null;
	}
	
	/**
	 * Sets a TextDB storing the labels for the vertices in the range [i, j].
	 * 
	 * @param i the lower bound of the range
	 * @param j the upper bound of the range
	 * @param db the TextDB to associate to the range [i, j]
	 * @throws IOException
	 */
	public void setVertexDB(int i, int j, TextDB db) throws IOException {
		Range r = new Range(i, j);
		if (vertices.get(r) != null) throw new IOException("Unable to define the range [" + i + ", " + j + "]. It overlaps a previously defined range.");
		vertices.put(r, db);
	}
	
	/**
	 * Sets a TextDB storing the vertex labels. If N is the
	 * size of the passed DB then
	 * vertices from 0 to N-1 will be labeled.
	 * 
	 * @param db
	 * @throws IOException
	 */
	public void setVertexDB(TextDB db) throws IOException {
		if (db.size() == 0) return;
		setVertexDB(0, db.size() - 1, db);
	}
	
	/**
	 * Sets a TextDB storing the labels for the vertices in the given range.
	 * 
	 * @param r	a range of vertices
	 * @param db the TextDB associated to the input range of vertices
	 * @throws IOException
	 */
	public void setVertexDB(Range r, TextDB db) throws IOException {
		setVertexDB(r.getLowerBound(), r.getUpperBound(), db);
	}
	
	
	/**
	 * Defines an edge attribute and its TextDB for accessing its values. If the attribute
	 * has a previously associated TextDB, this will be substituted
	 * and returned.
	 * 
	 * @param attribute the edge attribute to define
	 * @param db the TextDB used to access the values of this attribute
	 * @return a previously associated TextDB, if any
	 * @throws IOException
	 */
	public TextDB setEdgeDB(String attribute, TextDB db) throws IOException {
		return edges.put(attribute, db);
	}
	
	/**
	 * Returns the TextDBs covering the passed range.
	 * 
	 * @param r a range
	 * @return an array of TextDBs
	 */
	public TextDB[] getVertexDBs(Range r) {
		
		// get all the defined ranges with the associated DBs 
		Set<Entry<Range, TextDB>> entrySet = vertices.entrySet();

		// select all the overlapping ranges
		SortedSet<Range> ranges = new TreeSet<Range>();
		for (Entry<Range, TextDB> entry : entrySet) {
			Range defRange = entry.getKey();
			if (defRange.compareTo(r) == 0) ranges.add(defRange);
		}
		
		// collect all the associated DBs to return
		TextDB[] res = new TextDB[ranges.size()];
		int i = 0;
		for (Range range : ranges) {
			res[i++] = vertices.get(range);	
		}
		
		return res;
	}
	
	/**
	 * Given an input range, this method collect all the Range/TextDB pairs
	 * for all the subranges defined in this configuration and contained in the input one.  
	 * 
	 * @param r the input range
	 * @return the Range/TextDB pairs
	 */
	public SortedMap<Range, TextDB> getRangeDBMap(Range r) {
		SortedMap<Range, TextDB> map = new TreeMap<Range, TextDB>();
		
		Set<Entry<Range, TextDB>> entrySet = vertices.entrySet();

		// select all the overlapping ranges (in order)
		for (Entry<Range, TextDB> entry : entrySet) {
			Range defRange = entry.getKey();
			if (defRange.compareTo(r) == 0) map.put(defRange, entry.getValue());
		}
		
		return map;
	}
	
	/**
	 * Returns the TextDB associated to a vertex.
	 * 
	 * @param v the vertex id
	 * @return the TextDB containing the label of v
	 */
	public TextDB getVertexDB(int v) {
		Range r = new Range(v, v);
		return vertices.get(r);
	}
	
	/**
	 * Returns the range containing the input vertex, as defined in this configuration. If no range
	 * is found, then the range [vertex, vertex] is returned.
	 * 
	 * @param vertex the input vertex
	 * @return the range containing the input vertex, as defined by this configuration
	 */
	public Range getRangeFor(int vertex) {
		Range r = new Range(vertex, vertex);
		Set<Range> ranges = vertices.keySet();	// is a sorted set of ranges
		for (Range range : ranges) {
			if (range.compareTo(r) == 0) return range;
		}
		return r;
	}
	
	/**
	 * Returns the spanning range containing all the vertex ranges defined in this configuration.
	 * If no vertex range are defined then the range [-1, -1] is returned.
	 * 
	 * @return the spanning range over the vertex domain.
	 */
	public Range getSpanningRange() {
		Range first = vertices.firstKey();
		Range last = vertices.lastKey();
		
		if (vertices.size() == 0) return new Range(-1, -1);
		
		return new Range(first.getLowerBound(), last.getUpperBound());
	}

	/**
	 * Returns the TextDB storing the values of an edge attribute.
	 * 
	 * @param attribute the attribute name
	 * @return the TextDB providing access to the attribute values
	 */
	public TextDB getEdgeDB(String attribute) {
		return edges.get(attribute);
	}
	
	/**
	 * Writes on a file the current configuration.
	 * 
	 * @param file the output file
	 * @throws IOException
	 */
	public void store(String file) throws IOException {
		PrintStream out = new PrintStream(file);
		for (Entry<Range, TextDB> e : vertices.entrySet()) {
			out.println(e.getKey() + ":" + e.getValue().getClass().getCanonicalName());
		}
		
		for (Entry<String, TextDB> e : edges.entrySet()) {
			out.println(e.getKey() + ":" + e.getValue().getClass().getCanonicalName());
		}
		out.close();
	}
	
	/**
	 * Loads a configuration previously stored in a file.
	 * 
	 * @param file the configuration file to load
	 * @return an instance of LabellerConfig representing the loaded configuration
	 * @throws IOException
	 */
	public static GraphLabelerConfig load(String file) throws IOException {

		try {
			GraphLabelerConfig conf = new GraphLabelerConfig();
			
			BufferedReader br = new BufferedReader(new FileReader(file));
			for(String line = br.readLine(); line != null; line = br.readLine()) {
				
				if (line.length() == 0 || line.startsWith("#")) continue;
				
				String[] cols = line.split(":");

				String tdbFile = cols[1].trim();
				
				File f = new File(tdbFile);
				String parent = f.getParent();
				if (f.getParent() == null) {
					File infile = new File(file);
					parent = infile.getParent();
				}
				
				if (parent == null) 
					parent = "";
				else
					parent += System.getProperty("file.separator");
				
				TextDB db = TextDB.fromTDBFile(parent + tdbFile);
				
				if (cols[0].startsWith("[")) {
					// vertex range definition
					Range r = Range.parseRange(cols[0].trim());
					conf.setVertexDB(r, db);
				}
				else {
					// edge attribute definition
					String attribute = cols[0].trim();
					conf.setEdgeDB(attribute, db);
				}
			}
			br.close();
			
			return conf;
		}
		catch(IOException ex) {
			throw ex;
		}
		catch(Exception ex) {
			throw new IOException("Unable to load the LabellerConfig from " + file);
		}
	}
	
	/**
	 * Opens all the TextDBs in this configuration. 
	 * 
	 * @throws IOException
	 */
	protected void open() throws IOException {
		for (Entry<Range, TextDB> e : vertices.entrySet()) {
			//System.out.print("Opening " + e.getValue().getName() + "... ");
			e.getValue().open();
			//System.out.println("done.");
		}
		
		for (Entry<String, TextDB> e : edges.entrySet()) {
			//System.out.print("Opening " + e.getValue().getName() + "... ");
			e.getValue().open();
			//System.out.println("done.");
		}
	}
	
	/**
	 * Closes all the TextDBs in this configuration.  
	 * 
	 * @throws IOException
	 */
	public void close() throws IOException {
		for (Entry<Range, TextDB> e : vertices.entrySet()) {
			e.getValue().close();
		}
		
		for (Entry<String, TextDB> e : edges.entrySet()) {
			e.getValue().close();
		}
	}
	
	/**
	 * Returns the available edge attributes defined in this configuration.
	 * 
	 * @return the available edge attributes.
	 */
	public Set<String> getEdgeAttributes() {
		return edges.keySet();
	}
}
