/*
 * Copyright © 2006 - 2007 Claudio Corsi, Paolo Ferragina
 *
 * This file is part of SmallText library.
 *
 * SmallText is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SmallText is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package it.unipi.di.textdb;

import it.unimi.dsi.fastutil.objects.Object2IntOpenHashMap;
import it.unimi.dsi.mg4j.compression.Coder;
import it.unimi.dsi.mg4j.compression.Decoder;
import it.unimi.dsi.mg4j.compression.HuffmanCodec;
import it.unimi.dsi.mg4j.io.FastBufferedReader;
import it.unimi.dsi.mg4j.io.FastByteArrayOutputStream;
import it.unimi.dsi.mg4j.io.InputBitStream;
import it.unimi.dsi.mg4j.io.OutputBitStream;
import it.unimi.dsi.mg4j.util.Fast;
import it.unimi.dsi.mg4j.util.MutableString;
import it.unipi.di.tokenizer.FixedTokenizer;
import it.unipi.di.tokenizer.TermTokenizer;
import it.unipi.di.tokenizer.Tokenizer;
import it.unipi.di.tokenizer.URLTokenizer;
import it.unipi.di.util.Directory;
import it.unipi.di.util.Utils;
import jargs.gnu.CmdLineParser;

import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintStream;
import java.nio.ByteBuffer;
import java.util.Comparator;
import java.util.PriorityQueue;
import java.util.Properties;
import java.util.Map.Entry;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

/**
 * {@link TextDB} implementation that uses a combination of the Huffword data compressor
 * and a bucketing scheme. Each bucket consists of a fixed-number of 
 * contiguous records, and has variable length. The whole file is compressed
 * with Huffword, and pointers to compressed buckets (also called <i>jumpers</i>) 
 * are kept in a file on disk.<br><br> 
 * At query time the bucket containing the
 * requested record is identified, using its corresponding jumper, loaded in memory
 * and read sequentially until the requested record is met. Given the properties of 
 * Huffword, only the requested portion of a bucket must be uncompressed.
 * <br>
 * The tokens constituting the Huffword's alphabet are computed using a 
 * {@link Tokenizer}. Codewords are assigned to each token by considering
 * their frequencies in the input file. Optionally, the <i>ZetaCompressor</i>
 * can be used to assign codewords to tokens: sort tokens by decreasing 
 * frequency and Zeta-encode their ranks (in this order). This avoids the need 
 * to explicitly store the dictionary of codewords (but the output is sub-optimal).
 * 
 * <br><br>
 * The Huffman and ZetaCode implementations are provided by the library MG4J
 * (<a>http://mg4j.dsi.unimi.it/</a>).
 * 
 * @author Claudio Corsi
 * @author Paolo Ferragina
 */
public class BucketedHuffword extends TextDB {
	
	public static final int ZETA_CODEC = 0;
	public static final int HUFFMAN_CODEC = 1;
	
	private static final String[] codecNames = new String[] {"Zeta", "Huffman"};
	
	public static final int DEFAULT_BUCKET_LENGTH = 100;
	public static final int DEFAULT_ZETA_K = 3;
	
	/** The default size of the in memory buffer where to store the loaded record*/
	public static final int RECORD_BUFFER_LENGTH = 1024;
	
	private int[] jumpers;
	
	private int bucketLen;
	private ByteBuffer content;
	private int numOfBuckets;
	private int zetaK = DEFAULT_ZETA_K;	// used at runtime
	
	private static int codec = HUFFMAN_CODEC;	// use Huffman as the default buckets compression method
	private static int zetaKParameter = ZETA_CODEC;	// used at build time
	
	private int numOfLines = 0;
	private int currentCodec = HUFFMAN_CODEC;
	private Decoder huffDecoder;
	
	private Directory tdb;
	
	private int newLineCodeword = -1;
	//private int dictSize;
	private long maxBucketSize;
	private byte[] b;			// buffer where to load the buckets at runtime
	
	private int[] dictOffsets;
	private MutableString dictBuffer;	// in memory dictionary (one single char[])
	//private MutableString[] dict;
	
	private MutableString rec = new MutableString(RECORD_BUFFER_LENGTH); // the buffer where to build the loaded record
	
	public static void setCodec(int codec) {
		BucketedHuffword.codec = codec;
	}
	
	public static void setZetaKParameter(int k) {
		BucketedHuffword.zetaKParameter = k;
	}
	
	private static class EntryComparator implements Comparator<Entry<MutableString, Integer>> {
		public int compare(Entry<MutableString, Integer> e1, Entry<MutableString, Integer> e2) {
			return e2.getValue() - e1.getValue();	// descending order
		}
	}
	
	/**
	 * Create a new BucketedHuffword object loading the needed data structures
	 * from the provided file.
	 * 
	 * @param filename the TDB file containing the needed data structures and the compressed content
	 */
	public BucketedHuffword(String filename) {
		super(filename);
	}
	
	public void close() throws IOException {
		//dict = null		
		jumpers = null;
		dictOffsets = null;
		b = null;
		dictBuffer = null;
		rec = null;
		huffDecoder = null;

		tdb.close();
		
		Fast.compactMemory();
	}
	
	/**
	 * Compresses the input file with BucketedHuffword. 
	 * The default bucket length is {@value #DEFAULT_BUCKET_LENGTH}, and the tokens 
	 * are identified via a {@link TermTokenizer}.
	 * 
	 * <br><br>
	 * See the static method {@link #build(Tokenizer, String, String, int, PrintStream)} 
	 * to compress with a set of customized parameters.
	 * 
	 * @param log a PrintStream where log messages are print out. A <i>null</i> value will suppress any output message
	 * @param outfile The output file name.
	 */
	public TextDB build(String outfile, PrintStream log) throws IOException {
		return BucketedHuffword.build(new TermTokenizer(filename), filename, outfile, DEFAULT_BUCKET_LENGTH, log);
	}
	
	/**
	 * Compresses the input file with Bucketed Huffword using 
	 * a set of custom parameters.
	 * 
	 * @param tokenizer the tokenizer used to define the Huffoword tokens from the input file
	 * @param inputfile	the input file name
	 * @param bucketLen	the number of records per bucket
	 * @param log a PrintStream where log messages are print out. A <i>null</i> 
	 * value will suppress any output message
	 * 
	 * @return A TextDB instance to access the built database.
	 * 
	 * @throws IOException
	 */
	public static TextDB build(Tokenizer tokenizer, String inputfile, String outfile, int bucketLen, PrintStream log) throws IOException {
		
		boolean verbose = log != null;
		
		if (verbose) {
			System.out.println(" >> Input file: " + inputfile);
			System.out.println(" >> Bucket size: " + bucketLen);
			System.out.println(" >> Tokenizer: " + tokenizer.toString());
			System.out.println(" >> Using " + codecNames[codec] + " codec");
			
			if (codec == ZETA_CODEC) log.println(" >> Zeta K parameter: " + zetaKParameter);
			
			log.println(" >> Counting terms frequencies");
		}
		
		// collect the frequencies
		Object2IntOpenHashMap<MutableString> codewordsMap = new Object2IntOpenHashMap<MutableString>();
		
		codewordsMap.defaultReturnValue(-1);
		for(MutableString term = tokenizer.next(); term != null; term = tokenizer.next()) {
			int freq = codewordsMap.put(term, 1);
			if (freq != -1) codewordsMap.put(term, freq + 1);
		}
		
		// sort by freq (desc)
		if (verbose) log.println(" >> Sorting terms by frequencies (desc)");
		Comparator<Entry<MutableString, Integer>> c = new EntryComparator();
		PriorityQueue<Entry<MutableString, Integer>> heap = new PriorityQueue<Entry<MutableString,Integer>>(codewordsMap.size(), c);
		for(Entry<MutableString, Integer> e : codewordsMap.entrySet()) {
			heap.add(e);
		}
		
		Directory tdb = new Directory(outfile, Directory.WRITE_MODE);

		// the textdb-type file
		tdb.newFile("textdb-type");
		tdb.write(BucketedHuffword.class.getCanonicalName().getBytes());
		
		if (verbose) log.println(" >> Generating the model");
		
		tdb.newFile("revdict");
		GZIPOutputStream gzip = new GZIPOutputStream(tdb.getOutputStream());
		OutputStreamWriter osw = new OutputStreamWriter(gzip);
		
		int[] freq = null;
		HuffmanCodec huffman = null;
		Coder huffCoder = null;
		if (codec == HUFFMAN_CODEC) freq = new int[codewordsMap.size()];
		
		// the lengths of the dictionary terms
		int[] lengths = new int[codewordsMap.size()];
		
		// creating the codewords and the reverse dictionary
		int code = 0;
		for (Entry<MutableString, Integer> e = heap.poll(); e != null; e = heap.poll(), code++) {
			
			MutableString term = e.getKey();
			
			if (codec == HUFFMAN_CODEC) freq[code] = e.getValue();	// collect the frequencies for huffman 
			
			// filling the dict
			codewordsMap.put(term, code);  // update old values
			
			// filling the reverse dict directly on disk
			lengths[code] = term.length();
			osw.write(term.array(), 0, term.length());
			//if (!term.equals("\n")) osw.write('\n');
		}	
		
		osw.close();
		gzip.close();
		
		tdb.newFile("dict-term-lengths");
		for (int l : lengths) tdb.write(Utils.intToBytes(l));		// to compress it?
		
		
		if (codec == HUFFMAN_CODEC) {
			if (verbose) log.println(" >> Generating the Huffman tree");
			huffman = new HuffmanCodec(freq);
			huffCoder = huffman.coder();
			
			if (verbose) log.println(" >> Saving the Huffman decoder");
			tdb.newFile("decoder");
			ObjectOutputStream oos = new ObjectOutputStream(new GZIPOutputStream(tdb.getOutputStream()));
			oos.writeObject(huffman.decoder());
			oos.close();
		}
		
		// write the codewords in the output file
		if (verbose) log.println(" >> Compressing the file");
		
		FastByteArrayOutputStream jumpers = new FastByteArrayOutputStream();
		
		tdb.newFile("content");
		OutputStream out = tdb.getOutputStream();
		
		// write the bucket directly on disk
		OutputBitStream obs = new OutputBitStream(out);
		
		int bucketSize = 0;
		int numberOfLines = 0;
		int bucketCounter = 0;
		int newLineCodeword = codewordsMap.getInt(new MutableString("\n"));
		
		// init the jumpers
		jumpers.write(Utils.intToBytes(0));
		
		tokenizer.reset();
		
		long maxBucketSize = 0;
		long ptr = 0;
		
		for(MutableString term = tokenizer.next(); term != null; term = tokenizer.next()) {
			
			code = codewordsMap.getInt(term);
			
			switch (codec) {
			case ZETA_CODEC:
				obs.writeZeta(code, zetaKParameter);
				break;
			case HUFFMAN_CODEC:
				huffCoder.encode(code, obs);
				break;
			default:
				throw new IOException("Invalid coder!");
			}
			
			if (code == newLineCodeword) {
				bucketSize++;	// respect to the number of lines
				numberOfLines++;
			}
			
			if (bucketSize == bucketLen) {
				obs.flush();
				
				long currPtr = tdb.getFilePointer();
				jumpers.write(Utils.intToBytes((int)currPtr));
				bucketSize = 0;
				bucketCounter++;
				
				long bs = currPtr - ptr;
				if (bs >= maxBucketSize) maxBucketSize = bs;
				ptr = currPtr;
			}
		}
		
		if (bucketSize != 0) {	// flush last bucket
			obs.flush();
			long currPtr = tdb.getFilePointer();
			jumpers.write(Utils.intToBytes((int)currPtr));
			bucketSize = 0;
			bucketCounter++;
			
			long bs = currPtr - ptr;
			if (bs >= maxBucketSize) maxBucketSize = bs;
			ptr = currPtr;
		}

		// flush
		switch (codec) {
		case HUFFMAN_CODEC:
			huffCoder.flush(obs);
			break;
		}
		
		obs.close();
		
		// write the jumpers file
		tdb.newFile("jumpers");
		tdb.write(jumpers.array, 0, jumpers.length);
		
		int dictsize = codewordsMap.size();
		
		codewordsMap = null;	// in case of memory shortage...
		
		Properties prop = new Properties();
		prop.put("NEW_LINE_CODEWORD", "" + newLineCodeword);
		prop.put("DICT_SIZE", "" + dictsize);
		prop.put("BUCKETS", "" + bucketCounter);
		prop.put("BUCKET_LENGTH", "" + bucketLen);
		prop.put("CODEC_NAME", "" + codecNames[codec]);
		prop.put("CODEC_ID", "" + codec);
		prop.put("ZETA_K", "" + zetaKParameter);
		prop.put("NUMBER_OF_LINES", "" + numberOfLines);
		prop.put("TOKENIZER", "" + tokenizer.toString());
		prop.put("MAX_BUCKET_SIZE", "" + maxBucketSize);
		
		tdb.newFile("properties");
		prop.store(tdb.getOutputStream(), "developed by Claudio Corsi");
		
		if (verbose) {
			File f = new File(inputfile);
			int dictSize = tdb.sizeOf("revdict");
			int dictTermLengths = tdb.sizeOf("dict-term-lengths");
			
			log.println(" >> Number of lines: " + numberOfLines);
			log.println(" >> Dict size: " + dictsize);
			log.println(" >> Buckets: " + bucketCounter);
			log.println(" >> Original file size: " + Utils.formatSize(f.length()));
			log.println(" >> Jumpers file size: " + Utils.formatSize(jumpers.length));
			log.println(" >> Dict file size: " + Utils.formatSize(dictSize));
			log.println(" >> Term-lengths file size: " + Utils.formatSize(dictTermLengths));

			int huffSize = tdb.sizeOf("content");
			log.println(" >> Compressed file size: " + Utils.formatSize(huffSize));
			
			long total = dictSize + jumpers.length + huffSize + dictTermLengths;
			
			if (codec == HUFFMAN_CODEC) {
				log.println(" >> Decoder file size: " + Utils.formatSize(tdb.sizeOf("decoder")));
				total += tdb.sizeOf("decoder");
			}
			
			log.println(" >> Total compressed size: " + Utils.formatSize(total));
			
			double ratio = total / (double)f.length();
			log.println(" >> Compression ratio: " + Utils.round(ratio, 100));
		}
		
		jumpers.close();
		out.close();
		
		// write the TDB file on disk
		tdb.close();
		
		return TextDB.fromTDBFile(outfile);
	}
	
	@SuppressWarnings("unchecked")
	public void open() throws IOException {
		
		// open the directory in read-only mode
		tdb = new Directory(filename, Directory.READ_MODE);
		
		// load the properties
		Properties prop = new Properties();
		prop.load(tdb.getInputStream("properties"));
		
		bucketLen = Integer.parseInt(prop.getProperty("BUCKET_LENGTH"));
		numOfBuckets = Integer.parseInt(prop.getProperty("BUCKETS"));
		numOfLines = Integer.parseInt(prop.getProperty("NUMBER_OF_LINES"));
		currentCodec = Integer.parseInt(prop.getProperty("CODEC_ID"));
		zetaK = Integer.parseInt(prop.getProperty("ZETA_K"));
		newLineCodeword = Integer.parseInt(prop.getProperty("NEW_LINE_CODEWORD"));
		//dictSize = Integer.parseInt(prop.getProperty("DICT_SIZE"));
		maxBucketSize = Long.parseLong(prop.getProperty("MAX_BUCKET_SIZE"));
		
		// init the buffer where to load the buckets
		b = new byte[(int)maxBucketSize];
		
		// load the jumpers
		jumpers = new int[numOfBuckets + 1];
		ByteBuffer bb = tdb.mapFile("jumpers"); 
		bb.asIntBuffer().get(jumpers);
		
		// load the dict in memory
		InputStream is = tdb.getInputStream("revdict");
		GZIPInputStream gzip = new GZIPInputStream(is);
		FastBufferedReader fbr = new FastBufferedReader(new InputStreamReader(gzip));
		
		// collecting all the chars into one single memory buffer
		dictBuffer = new MutableString();							// is NOT compact
		
		char[] buff = new char[102400];   // 100KB
		for (int len = fbr.read(buff); len != -1; len = fbr.read(buff)) {
			dictBuffer.append(buff, 0, len);
		}

		fbr.close();
		gzip.close();
		
		dictBuffer.compact();		// now is COMPACT
		
		// loading the terms offsets....
		int size = tdb.sizeOf("dict-term-lengths") / 4;
		dictOffsets = new int[size + 1];
		bb = tdb.mapFile("dict-term-lengths");
		bb.asIntBuffer().get(dictOffsets, 0, size);
		dictOffsets[size] = 0;
		
		int off = 0;
		for(int k = 0; k < dictOffsets.length; k++) {
			int tmp = dictOffsets[k]; 
			dictOffsets[k] = off;
			off += tmp;
		}
		
		if (currentCodec == HUFFMAN_CODEC) {
			try {
				ObjectInputStream ois = new ObjectInputStream(new GZIPInputStream(tdb.getInputStream("decoder")));
				huffDecoder = (Decoder)ois.readObject();
			}
			catch(ClassNotFoundException cne) {
				throw new IOException(cne.getMessage());
			}
		}
		
		content = tdb.mapFile("content");
	}
	
	public int size() {
		return numOfLines;
	}
	
	/**
	 * Returns the record for a given position in the compressed file, <i>null</i>
	 * if this position is out of range.
	 * 
	 * @param record The record number in the range [0, N-1]
	 * @return the record or <i>null</i>
	 * @throws IOException
	 */
	public String get(int record) throws IOException {
		
		if (record < 0 || record >= numOfLines) return null;
		
		int pos = record/bucketLen;
		int start = jumpers[pos];
		int stop = jumpers[pos + 1];
		int size = stop - start;
		
		content.position(start);
		content.get(b, 0, size);
		
		InputBitStream ibs = new InputBitStream(b);

		rec.length(0);
		
		for(int count = record - (pos * bucketLen); count >= 0;) {
			int code = -1;
			switch (currentCodec) {
			case HUFFMAN_CODEC:
				code = huffDecoder.decode(ibs);
				break;
			case ZETA_CODEC:
				code = ibs.readZeta(zetaK);
				break;
			default:
				throw new IOException("Unsupported codec");	
			}
			
			if (code == newLineCodeword)
				count--;
			else 
				if (count == 0) {
					//rec.append(dict[code]);
					int len = dictOffsets[code + 1] - dictOffsets[code];
					rec.append(dictBuffer.array(), dictOffsets[code], len); 
				}
		}
		
		return rec.toString();
	}
	
	public void getSequential(int[] records, int field, BufferedWriter out) throws IOException {
		getSequentialImplementation(records, 0, records.length, field, out, false);
	}
	
	public String[] getSequential(int[] records) throws IOException {
		return getSequentialImplementation(records, 0, records.length, -1, null, false);
	}
	
	public String[] getSequential(int[] records, int pos, int length) throws IOException {
		return getSequentialImplementation(records, pos, length, -1, null, false);
	}
	
	
	public String[] getRange(int i, int j) throws IOException {
		return getSequentialImplementation(null, i, j - i + 1, -1, null, true);
	}
	
	public String[] getRange(int i, int j, int field) throws IOException {
		return getSequentialImplementation(null, i, j - i + 1, field, null, true);
	}
	
	public void getRange(int i, int j, int field, BufferedWriter out) throws IOException {
		getSequentialImplementation(null, i, j - i + 1, field, out, true);
	}

	/*
	 * if range == true, records == null, pos = i, length = range size
	 */
	private String[] getSequentialImplementation(int[] records, int pos, int length, int field, BufferedWriter out, boolean range) throws IOException {
		
		if (range) {
			if (pos < 0 || (pos + length - 1) >= numOfLines) throw new IOException("Out of bound records");
		}
		else { 
			if (records[0] < 0 || records[pos + length - 1] >= numOfLines) throw new IOException("Out of bound records");
		}
		
		String[] res = (out == null) ? new String[length] : null;
		
		int seenSoFar = 0;
		InputBitStream ibs = null;
		int lastRank = -1;
		int lastBucket = -1;
		MutableString lastOne = null;
		int k = 0;		// current position in res[]
		
		for(int curr = pos; curr < pos + length; curr++) {
			
			int currRank = (range) ? curr : records[curr];
			
			if (lastRank == currRank) {		// in case of duplicate ranks
				
				if (out == null) 
					res[k++] = lastOne.toString();
				else {
					if (lastOne != null) out.write(lastOne.array(), 0, lastOne.length());
					out.write('\n');
				}

				continue;
			}
			
			lastRank = currRank;
			lastOne = null;
			
			int currBucket = currRank/bucketLen;
			
			if (currBucket != lastBucket) {		
				
				// load and cache the current bucket
				
				int start = jumpers[currBucket];
				int stop = jumpers[currBucket + 1];
				int size = stop - start;
				
				content.position(start);
				content.get(b, 0, size);
				
				lastBucket = currBucket;
				ibs = new InputBitStream(b);
				seenSoFar = 0;
			}
			
			rec.length(0);	// reset the buffer
			
			// sequential scan of the bucket
			for(int count = currRank - (currBucket * bucketLen) - seenSoFar; count >= 0;) {
				
				int code = -1;
				switch (currentCodec) {
				case HUFFMAN_CODEC:
					code = huffDecoder.decode(ibs);
					break;
				case ZETA_CODEC:
					code = ibs.readZeta(zetaK);
					break;
				default:
					throw new IOException("Unsupported codec");
				}
				
				if (code == newLineCodeword) {
					count--;
					seenSoFar++;
				}
				else 
					if (count == 0) {
						int len = dictOffsets[code + 1] - dictOffsets[code];
						rec.append(dictBuffer.array(), dictOffsets[code], len); 
						
						//rec.append(dict[code]);
					}
			}

			lastOne = Utils.getField(rec, field, fieldSeparator);
			
			if (out == null)
				res[k++] = lastOne.toString();
			else {
				if (lastOne != null) out.write(lastOne.array(), 0, lastOne.length());
				out.write('\n');
			}
		}
		
		return res;
	}
	
	private static void printUsage() {
		System.out.println();
		System.out.println(
				"Build a BucketedHuffword database. It splits the input file into fixed length\n" +
				"buckets and compress them with huffword. The huffword dictionary is built\n" +
				"by default using a TermTokenizer.");
		System.out.println();
		System.out.println("Input: the name of the file to store.");
		System.out.println();
		System.out.println("Optional arguments:");
		System.out.println();
		System.out.printf("   %1$-30s   %2$s\n", "-h/--help", "Print this message.");
		System.out.printf("   %1$-30s   %2$s\n", "-v/--verbose", "Print log messages on stdout.");
		System.out.printf("   %1$-30s   %2$s\n", "-b/--bucket-size", "The maximum number of records to put into a single bucket.");
		System.out.printf("   %1$-30s   %2$s\n", "--url-tokenizer", "Use the URLTokenizer to build the huffword dictionary.");
		System.out.printf("   %1$-30s   %2$s\n", "--fixed-tokenizer", "Use the FixedTokenizer to build the huffword dictionary.");
		System.out.printf("   %1$-30s   %2$s\n", "--fixed-tokenizer-length <len>", "Set the length parameter of the FixedTokenizer.");
		System.out.printf("   %1$-30s   %2$s\n", "--huffman", "Encode the huffword codewords with the Huffman encodign. (Default)");
		System.out.printf("   %1$-30s   %2$s\n", "--zeta", "Encode the huffword codewords with the Zeta encoding.");
		System.out.printf("   %1$-30s   %2$s\n", "--zeta-k <val>", "Set the K value for the zeta encoding. Default value is 3.");
		System.out.printf("   %1$-30s   %2$s\n", "-o/--outfile <filename>", "Set the output file name. Default is '<inputfilename>.tdb'.");
		System.out.println();
	}
	
	public static void main(String[] args) throws Exception {
		
		CmdLineParser parser = new CmdLineParser(); 
		
		CmdLineParser.Option help0 = parser.addBooleanOption('h', "help");
		CmdLineParser.Option verbose0 = parser.addBooleanOption('v', "verbose");
		CmdLineParser.Option urlTokenizer0 = parser.addBooleanOption("url-tokenizer");
		CmdLineParser.Option fixedTokenizer0 = parser.addBooleanOption("fixed-tokenizer");
		CmdLineParser.Option fixedLen0 = parser.addIntegerOption("fixed-tokenizer-length");
		CmdLineParser.Option bucketSize0 = parser.addIntegerOption('b', "bucket-size");
		CmdLineParser.Option k0 = parser.addIntegerOption("zeta-k");
		CmdLineParser.Option huff0 = parser.addBooleanOption("huffman");
		CmdLineParser.Option zeta0 = parser.addBooleanOption("zeta");
		CmdLineParser.Option out0 = parser.addStringOption('o', "outfile");
		
		try {
			parser.parse(args);
		} 
		catch (CmdLineParser.OptionException e) {
			System.err.println(e.getMessage());
			printUsage();
			System.exit(1);
		}
		
		boolean help = (Boolean)parser.getOptionValue(help0, Boolean.FALSE);
		if (help) {
			printUsage();
			return;
		}
		
		String file = null;
		String[] otherArgs = parser.getRemainingArgs();
		if (otherArgs.length == 1) {
			file = otherArgs[0];
			File f = new File(file);
			if (!f.exists()) {
				System.err.println("Input file \"" + otherArgs[0] + "\" not found.");
				System.exit(1);
			}
		}
		else {
			printUsage();
			System.exit(1);
		}
		
		boolean verbose = (Boolean)parser.getOptionValue(verbose0, Boolean.FALSE);
		boolean urlsTokenizer = (Boolean)parser.getOptionValue(urlTokenizer0, Boolean.FALSE);
		int bucketSize = (Integer)parser.getOptionValue(bucketSize0, BucketedHuffword.DEFAULT_BUCKET_LENGTH);
		boolean fixedTokenizer = (Boolean)parser.getOptionValue(fixedTokenizer0, Boolean.FALSE);

		if (fixedTokenizer && urlsTokenizer) {
			System.err.println("Incompatible options --fixed-tokenizer and --url-tokenizer");
			System.exit(1);
		}
		
		Tokenizer tokenizer = new TermTokenizer(file);
		if (urlsTokenizer) tokenizer = new URLTokenizer(file);
		if (fixedTokenizer) {
			int len = (Integer)parser.getOptionValue(fixedLen0, FixedTokenizer.DEFAULT_LENGTH);
			
			if (len <= 0) {
				System.err.println("The value for --fixed-tokenizer-length option must be > 0");
				System.exit(1);
			}
			
			tokenizer = new FixedTokenizer(file, len);
		}
		
		boolean useHuffman = (Boolean)parser.getOptionValue(huff0, Boolean.FALSE);
		boolean useZeta = (Boolean)parser.getOptionValue(zeta0, Boolean.FALSE);
		
		int k = (Integer)parser.getOptionValue(k0, BucketedHuffword.DEFAULT_ZETA_K);
		
		if (useHuffman) BucketedHuffword.setCodec(HUFFMAN_CODEC);
		
		if (useZeta) {
			BucketedHuffword.setCodec(ZETA_CODEC);
			BucketedHuffword.setZetaKParameter(k);
		}
		
		String outfile = (String)parser.getOptionValue(out0);
		if (outfile == null) outfile = file + ".tdb";
		
		BucketedHuffword.build(tokenizer, file, outfile, bucketSize, verbose ? System.out : null);
	}
}
