/*
 * Copyright © 2006 - 2007 Claudio Corsi, Paolo Ferragina
 *
 * This file is part of SmallText library.
 *
 * SmallText is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SmallText is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package it.unipi.di.textdb;

import it.unimi.dsi.mg4j.io.FastBufferedReader;
import it.unimi.dsi.mg4j.io.FastByteArrayOutputStream;
import it.unimi.dsi.mg4j.util.MutableString;
import it.unipi.di.util.Directory;
import it.unipi.di.util.ExternalSort;
import it.unipi.di.util.Utils;
import jargs.gnu.CmdLineParser;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.nio.ByteBuffer;
import java.util.Properties;

/**
 * This is a {@link TextDB} which uses a combination of a bucketing scheme 
 * and the Zip data compression technique. A bucket is defined
 * as a fixed-number of contiguous records. Each bucket is compressed with 
 * Zip (thus it has variable length),
 * and may be accessed via a pointer (also called <i>jumper</i>) kept in a 
 * file on disk.
 * <br><br>  
 * At query time the bucket containing the requested record is identified, 
 * using its corresponding jumper, loaded in memory and (fully-)uncompressed until
 * the requested record is met.
 * 
 * @see ExternalSort
 * 
 * @author Claudio Corsi
 * @author Paolo Ferragina
 *
 */
public class BucketedZip extends TextDB {

	public final static int DEFAULT_BUCKET_SIZE = 100;
	public final static int DEFAULT_COMPRESSION_LEVEL = 6;
	
	private int[] jumpers; // offsets of the compressed buckets
	private int numBuckets; // total number of buckets
	private int bucketSize = DEFAULT_BUCKET_SIZE; // number of strings per bucket
	private int numStrings; // number of strings overall
	private ByteBuffer content;
	private Directory tdb;
	private long maxBucketSize;
	private byte[] block;
	
	/**
	 * Create a new BucketedZip object loading
	 * the needed data structures from the provided file. 
	 * 
	 * @param filename the file containing the content and the data structures to load, stored in TDB format 
	 */
	public BucketedZip(String filename) {
		super(filename);
	}
	
	public void close() throws IOException {
		tdb.close();
	}

	public int size() {
		return numStrings;
	}
	
	public String get(int record) throws IOException {
		String[] res = select(new int[] {record}, 0, 1, -1, null, false);
		if (res == null) return null;
		if (res.length == 0) throw new IOException("You have found a bug! Please contact the authors.");
		return res[0];
	}
	
	public String[] getRange(int i, int j) throws IOException {
		return select(null, i, j - i + 1, -1, null, true);
	}
	
	public void getRange(int i, int j, int field, BufferedWriter out) throws IOException {
		select(null, i, j - i + 1, field, out, true);
	}

	public String[] getSequential(int[] records) throws IOException {
		return select(records, 0, records.length, -1, null, false);
	}
	
	public String[] getSequential(int[] records, int pos, int length) throws IOException {
		return select(records, pos, length, -1, null, false);
	}
	
	public void getSequential(int[] records, int field, BufferedWriter out) throws IOException {
		select(records, 0, records.length, field, out, false);
	}
	
	/*
	 * records[] == array of 0-based rec IDs
	 * if range == true, records == null, pos = i, length = range size
	 */
	private String[] select(int[] records, int pos, int length, int field, BufferedWriter out, boolean range) throws IOException {
		
		if (range) {
			if (pos + 1 <= 0 || (pos + length) > numStrings) throw new IOException("Out of bound records");
		}
		else { 
			if (records[0] + 1 <= 0 || records[pos + length - 1] + 1 > numStrings) throw new IOException("Out of bound records");
		}
		
		String[] res = (out == null) ? new String[length] : null;
		MutableString lastOne = null;
		
		int k = 0;		// current position in res[]
		int lastRank = -1;
		int lastBucketNum = -1;
		int bucketNum = -1;
		
		byte[] b = null;
		int j = 0;
		FastByteArrayOutputStream fbs = new FastByteArrayOutputStream((int)maxBucketSize);
		
		int leftRanks = 0;
		int size = 0;
		int seenSoFar = 0;
		
		// init the buffers
		MutableString rec = new MutableString(1024);
		
//		byte[] buffer = new byte[(int)maxBucketSize];  <-- use the max record size (to compute)
//		ZInputStream zIn = null;
//		FastBufferedInputStream fbis = null;

		for(int curr = pos; curr < pos + length; curr++) {
			
			int rank = (range) ? curr + 1 : records[curr] + 1;		// the rank to fetch (1-based)
			
			if (lastRank == rank) {		// in case of duplicate ranks
				
				if (out == null) 
					res[k++] = lastOne.toString();
				else {
					if (lastOne != null) out.write(lastOne.array(), 0, lastOne.length());
					out.write('\n');
				}
				
				continue;
			}
			
			lastRank = rank;
			lastOne = null;
			
			bucketNum = (int) (rank / bucketSize);	// the containing bucket
			
			// load a new bucket if needed
			if (bucketNum != lastBucketNum) {	
				
				// first string in bucket 0 is dummy
				leftRanks = (int) (rank % bucketSize) + 1;
				
				// candidate bucket delimiters
				int start = jumpers[bucketNum];
				int end = jumpers[bucketNum + 1];
				size = end - start;
		
				// load the bucket
				content.position(start);
				content.get(block, 0, size);
				
				// uncompress the entire block
				size = Utils.fastUnzip(block, 0, size, fbs);
				b = fbs.array;
				j = 0;	// starts from the beginning
				
				// sequential block access
//				zIn = new ZInputStream(new FastByteArrayInputStream(block, 0, size));
//				fbis = new FastBufferedInputStream(zIn, 16 * 1024);

				rec.length(0);		// reset the current record
				seenSoFar = 0;
				
				lastBucketNum = bucketNum;
			}
			else {
				leftRanks = (int)(rank % bucketSize) + 1  - seenSoFar;
			}
			
			int len = 0;
			int start = j;
			for(; (j < size) && (leftRanks > 0); leftRanks--, j++) {
				start = j;
				len = 0;
				for (; b[j] != '\n'; j++) len++;
				seenSoFar++;
			}
			rec.length(0);
			rec.append(new String(b, start, len));	// decoding the byte[] through a String object

//			int len = 0;
//			while(leftRanks > 0) {
//				len = fbis.readLine(buffer);
//				leftRanks--;
//				seenSoFar++;
//			}
//			rec.length(0);
//			rec.append(new String(buffer, 0, len));	// decoding the byte[] through a String object
			
			lastOne = Utils.getField(rec, field, fieldSeparator);
			
			if (out == null)
				res[k++] = lastOne.toString();
			else {
				if (lastOne != null) out.write(lastOne.array(), 0, lastOne.length());
				out.write('\n');
			}
		}
		
		return res;
	}

	public void open() throws IOException {
		
		tdb = new Directory(filename, Directory.READ_MODE);

		Properties prop = new Properties();
		prop.load(tdb.getInputStream("properties"));

		numStrings = Integer.parseInt(prop.getProperty("NUM_STRINGS"));
		numBuckets = Integer.parseInt(prop.getProperty("NUM_BUCKETS"));
		bucketSize = Integer.parseInt(prop.getProperty("BUCKET_SIZE"));
		maxBucketSize = Long.parseLong(prop.getProperty("MAX_BUCKET_SIZE"));
		
		jumpers = new int[numBuckets + 1];
		block = new byte[(int)maxBucketSize];
		
		// load jumpers
		ByteBuffer jumpBuff = tdb.mapFile("jumpers");
		int i = 0;
		for(; i < jumpers.length; i++) {
			jumpers[i] = jumpBuff.getInt();
		}
		
		content = tdb.mapFile("content");
	}
	
	/**
	 * Build a BucketedZip over an input file. Custom parameters are related
	 * to the bucket size and to the search support. This last option need a <i>sorted</i>
	 * file in input.
	 * 
	 * @param inputfile the file to compress
	 * @param outfile the output file name
	 * @param bucketSize the maximum size (in the number of records) of each bucket 
	 * @param level the compression level (from 0 = FASTEST to 9 = BEST COMPRESSION)
	 * @param log a {@link PrintStream} where to send the log messages. If null that messages will be suppressed
	 * @return the TextDB to access the build database
	 * @throws IOException
	 */
	public static TextDB build(String inputfile, String outfile, int bucketSize, int level, PrintStream log) throws IOException {
		
		boolean verbose = log != null;
		
		if (verbose) {
			log.println(" >> Input file: " + inputfile);
			log.println(" >> Buckets size: " + bucketSize);
			log.println(" >> Compression level: " + level);
		}
		
		FastBufferedReader in = new FastBufferedReader(new FileReader(inputfile));
		FastByteArrayOutputStream memJumpers = new FastByteArrayOutputStream();
		
		Directory tdb = new Directory(outfile, Directory.WRITE_MODE);
		
		// the textdb-type file
		tdb.newFile("textdb-type");
		tdb.write(BucketedZip.class.getCanonicalName().getBytes());
		
		FastByteArrayOutputStream bucket = new FastByteArrayOutputStream();
		
		int numRecords = 0;
		int numBuckets = 0;

		long fileSize = (new File(inputfile)).length();
	
		tdb.newFile("content");
		
		// init the beginners & jumpers
		memJumpers.write(Utils.intToBytes(0));
		
		long maxBucketSize = 0;
		
		if (verbose) log.println(" >> Compressing the file");
		
		FastByteArrayOutputStream zout = new FastByteArrayOutputStream(1024);
		
		MutableString rec = new MutableString();
		
		while(in.readLine(rec) != null) {
			
			numRecords++;
			bucket.write(rec.toString().getBytes());
			bucket.write('\n');
			
			if (numRecords % bucketSize == 0) { // Bucket start
				
				// compress the current bucket
				Utils.fastZip(bucket.array, 0, bucket.length, zout, level);
				
				// write the compressed data
				tdb.write(zout.array, 0, zout.length);
				
				// one more bucket
				numBuckets++;
				
				// reset the bucket
				bucket.reset();

				// write the starting position of the next one
				memJumpers.write(Utils.intToBytes((int)tdb.getFilePointer()));
				
				if (zout.length >= maxBucketSize) maxBucketSize = zout.length;
			}
		}

		// finalize latest bucket
		if (bucket.length > 0) {
			
			// compress the current bucket
			Utils.fastZip(bucket.array, 0, bucket.length, zout, level);
			
			// write the compressed data
			tdb.write(zout.array, 0, zout.length);
			
			// one more bucket
			numBuckets++;
			
			// reset the bucket
			bucket.reset();

			// write the starting position of the next one
			memJumpers.write(Utils.intToBytes((int)tdb.getFilePointer()));
			
			if (zout.length >= maxBucketSize) maxBucketSize = zout.length;
		}
		
		tdb.newFile("jumpers");
		tdb.write(memJumpers.array, 0, memJumpers.length);
		
		// Properties
		Properties prop = new Properties();
		prop.put("NUM_STRINGS", "" + numRecords);	// don't count the dummy string
		prop.put("NUM_BUCKETS", "" + numBuckets);
		prop.put("BUCKET_SIZE", "" + bucketSize);
		prop.put("MAX_BUCKET_SIZE", "" + maxBucketSize);
		
		tdb.newFile("properties");
		prop.store(tdb.getOutputStream(), "Developed by Paolo Ferragina and Claudio Corsi");
		
		// close everything
		bucket.reset();
		memJumpers.close();
		tdb.close();
		
		if (verbose) {
			
			log.println(" >> Original file size: " + Utils.formatSize(fileSize));
			log.println(" >> Number of buckets: " + numBuckets);
			log.println(" >> Number of strings: " + (numRecords-1));
			
			int tot = tdb.sizeOf("jumpers");
			log.println(" >> Jumpers file size: " + Utils.formatSize(tdb.sizeOf("jumpers")));
			
			tot += tdb.sizeOf("content");
			log.println(" >> Compressed file size: " + Utils.formatSize(tdb.sizeOf("content")));
			
			log.println(" >> Total compressed space: " + Utils.formatSize(tot));
			
			double ratio = tot / (double)fileSize;
			log.println(" >> Compression ratio: " + Utils.round(ratio, 100));
		}
		
		return TextDB.fromTDBFile(outfile);
	}
	
	public TextDB build(String outfile, PrintStream log) throws IOException {
		return BucketedZip.build(filename, outfile, DEFAULT_BUCKET_SIZE, DEFAULT_COMPRESSION_LEVEL, log);
	}
	
	public static void main(String[] args) throws Exception {

		CmdLineParser parser = new CmdLineParser(); 
		CmdLineParser.Option help0 = parser.addBooleanOption('h', "help");
		CmdLineParser.Option bucketSize0 = parser.addIntegerOption('b', "bucket-size");
		CmdLineParser.Option verbose0 = parser.addBooleanOption('v', "verbose");
		CmdLineParser.Option out0 = parser.addStringOption('o', "outfile");
		CmdLineParser.Option level0 = parser.addIntegerOption('l', "level");
		
		try {
			parser.parse(args);
		} 
		catch (CmdLineParser.OptionException e) {
			System.err.println(e.getMessage());
			printUsage();
			System.exit(1);
		}
		
		boolean help = (Boolean)parser.getOptionValue(help0, Boolean.FALSE);
		if (help) {
			printUsage();
			return;
		}
		
		boolean verbose = (Boolean)parser.getOptionValue(verbose0, Boolean.FALSE);
		Integer bucketSize = (Integer)parser.getOptionValue(bucketSize0);
		
		Integer level = (Integer)parser.getOptionValue(level0, DEFAULT_COMPRESSION_LEVEL);
		if (level < -1 || level > 9) {
			System.err.println("The compression level must be a value from 0 to 9.");
			System.exit(1);
		}
		
		String inputfile = null;
		String[] otherArgs = parser.getRemainingArgs();
		if (otherArgs.length == 1) {
			// input file
			File f = new File(otherArgs[0]);
			if (!f.exists()) {
				System.err.println("Input file \"" + otherArgs[0] + "\" not found.");
				System.exit(1);
			}
			else {
				inputfile = otherArgs[0];
			}
		}
		else {
			printUsage();
			System.exit(1);
		}
		
		String outfile = (String)parser.getOptionValue(out0);
		if (outfile == null) outfile = inputfile + ".tdb";
		
		// =================== START HERE ========================= //
		
		long start = System.currentTimeMillis();

		if (bucketSize != null)
			BucketedZip.build(inputfile, outfile, bucketSize, level, verbose ? System.out : null);
		else
			BucketedZip.build(inputfile, outfile, DEFAULT_BUCKET_SIZE, level, verbose ? System.out : null);
		
		if (verbose) System.out.println(" >> Done in " + Utils.elapsedTime(start, System.currentTimeMillis()));
	}
	
	private static void printUsage() {
		System.out.println();
		System.out.println("Build a BucketedZip database. It splits the input file into fixed\nlength buckets and compress them with Zip.");
		System.out.println();
		System.out.println("Input: the name of the file to store.");
		System.out.println();
		System.out.println("Optional arguments:");
		System.out.println();
		System.out.printf("   %1$-30s   %2$s\n", "-h/--help", "Print this message.");
		System.out.printf("   %1$-30s   %2$s\n", "-v/--verbose", "Print log messages on stdout.");
		System.out.printf("   %1$-30s   %2$s\n", "-b/--bucket-size", "The maximum number of records to put into a single bucket.");
		System.out.printf("   %1$-30s   %2$s\n", "-l/--level", "Set the compression level to use (default is 6). Values range from -1 (faster) to 9 (better).");
		System.out.printf("   %1$-30s   %2$s\n", "-o/--outfile <filename>", "Set the output file name. Default is '<inputfilename>.tdb'.");
		System.out.printf("   %1$-30s   %2$s\n", "--search-support", "Build the data structures to allow searches into this database.");
		System.out.println();
	}
}
