/*
 * Copyright © 2006 - 2007 Claudio Corsi, Paolo Ferragina
 *
 * This file is part of SmallText library.
 *
 * SmallText is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SmallText is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package it.unipi.di.textdb;

import it.unimi.dsi.mg4j.io.FastByteArrayOutputStream;
import it.unimi.dsi.mg4j.util.MutableString;
import it.unipi.di.util.Directory;
import it.unipi.di.util.ExternalSort;
import it.unipi.di.util.Range;
import it.unipi.di.util.Utils;
import jargs.gnu.CmdLineParser;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.io.StringReader;
import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.Properties;

/**
 * A {@link TextDB} that uses the Front-Coding technique to compress its records 
 * and a bucketing scheme to provide efficient access to them. 
 * A bucket is defined as a fixed-number of contiguous records, and
 * has variable length. 
 * In each bucket the first record is stored explicitly, whereas any other record is
 * front-encoded by squeezing out the longest prefix (LP) it shares with the 
 * preceding record in its bucket. The squeezing consists of substituting LP 
 * with its length (encoded using 1 or 3 bytes). Optionally a front-coded 
 * bucket can be further compressed using GZip.
 * The pointer to the beginning of each bucket (also called <i>jumper</i>)
 * is stored in a separate file on disk.<br><br> 
 * 
 * At query time the bucket containing the requested record is identified,
 * using its jumper, loaded in memory and accessed sequentially until the 
 * record is met.
 * 
 * <br><br><br>
 * To achieve higher compression, this scheme should be applied over a 
 * TextDB whose records
 * are <i>sorted lexicographically</i> (see {@link ExternalSort}).
 * 
 * @author Claudio Corsi
 * @author Paolo Ferragina
 *
 */
public class FrontCoding extends TextDB implements SearchableDB {

	public final static int DEFAULT_BUCKET_SIZE = 100;
	public final static int DEFAULT_COMPRESSION_LEVEL = 6;
	
	public final static char SMALLER_CHAR = '\u0000';
	public final static char BIGGER_CHAR = '\uffff';
	public final static byte ESCAPE_VALUE_BYTE = 127;
	
	private String[] bucketBeginners; // strings starting the buckets
	private int[] jumpers; // offsets of the compressed buckets
	private int numBuckets; // total number of buckets
	private int bucketSize = DEFAULT_BUCKET_SIZE; // number of strings per bucket
	private long numStrings; // number of strings overall
	private long dictSize; // byte size of the dictionary (including \n)
	private boolean frontCoded = true;
	private boolean zipped = true;
	private ByteBuffer content;
	private Directory tdb;
	private boolean searchSupport = true;
	private long maxBucketSize;
	private byte[] block;
	
	public String dictName() {
		return filename;
	}

	public long dictSize() {
		return dictSize;
	}

	public long numBuckets() {
		return numBuckets;
	}

	public long bucketSize() {
		return bucketSize;
	}

	public long numStrings() {
		return numStrings;
	}

	public long fcBeginSize() {
		return (new File(filename + "-fc.beginners")).length();
	}

	public long fcDictSize() {
		return (new File(filename + "-fc.dict")).length();
	}

	public long fcJumpSize() {
		return (new File(filename + "-fc.jumpers")).length();
	}

	public long fcTotalSize() {
		return fcDictSize() + fcJumpSize() + fcBeginSize();
	}

	public float compressRatio() {
		return ((float) fcTotalSize()) / dictSize();
	}

	private static int lcp(String s1, String s2) {
		int len;
		for (len = 0; (len < s1.length()) && (len < s2.length())
				&& (s1.charAt(len) == s2.charAt(len)); len++);
		return len;
	}
	
	/**
	 * Given a front-coded string represented on a byte array, this 
	 * method first reads the length of the squeezed prefix (1 or 3 bytes)
	 * and then writes the remaining suffix in the input StringBuffer. 
	 * It finally returns the length of the front-coded string (i.e. 1 or 3 bytes
	 * plus byte-length of the remaining suffix). 
	 * <br> <br>
	 * NOTE: the input StringBuffer is not reseted by this method and 
	 * the string suffix is appended starting at the current position. 
	 * 
	 * @param b the array containing the front-coded string
	 * @param offset the starting position in b of the front-compressed string
	 * @param out the StringBuffer where the string suffix is written out
	 * @return
	 */
	private int getSuffixSkip(byte[] b, int offset, StringBuilder out) {
		
		int skip = 1;
		int len = b[offset++];
		if (len == ESCAPE_VALUE_BYTE) {
			len = b[offset++] * ESCAPE_VALUE_BYTE;
			len += b[offset++];
			skip += 2;
		}
		
		out.append(new String(b, offset, len));
		
		return len + skip;
	}
	
	private int getSuffixSkip(byte[] b, int offset, MutableString out) {
		
		int skip = 1;
		int len = b[offset++];
		if (len == ESCAPE_VALUE_BYTE) {
			len = b[offset++] * ESCAPE_VALUE_BYTE;
			len += b[offset++];
			skip += 2;
		}
		
		out.length(0);	// reset the out buffer
		out.append(new String(b, offset, len));	// use the String to decode the byte[] with the right Charset encoding
			
		return len + skip;
	}

	/**
	 * Created a new FrontCoding object from the provided file.
	 * 
	 * @param filename the TDB file containing the front coded records and the data
	 * structures needed to access them
	 */
	public FrontCoding(String filename) {
		super(filename);
		this.dictSize = (new File(filename)).length();
	}
	
	/**
	 * Created a new FrontCoding object from the provided file.
	 * 
	 * @param filename the TDB file containing the front coded records and the data
	 * structures needed to access them
	 * @param searchSupport if false don't load the data structures to support
	 * the searches defined in the {@link SearchableDB} interface 
	 */
	public FrontCoding(String filename, boolean searchSupport) {
		super(filename);
		this.dictSize = (new File(filename)).length();
		this.searchSupport = searchSupport;
	}
	
	/**
	 * Compresses the TextDB via (plain) front-coding with buckets 
	 * of {@value #DEFAULT_BUCKET_SIZE} records. 
	 * There is no support for the interface {@link SearchableDB}. 
	 * 
	 * @param log a PrintStream where to print log messages. A <i>null</i> value will suppress any output message
	 * @param outfile the output file name
	 */
	public TextDB build(String outfile, PrintStream log) throws IOException {
		return FrontCoding.build(filename, outfile, DEFAULT_BUCKET_SIZE, false, true, true, DEFAULT_COMPRESSION_LEVEL, log);
	}
	
	/**
	 * Compresses the input file with a customized compression scheme, which
	 * combines frontcoding and Zip. 
	 * 
	 * @param inputfile	the file to compress
	 * @param outfile the output file name
	 * @param bucketSize the number of contiguous records forming a bucket
	 * @param compress if true, ZIP is used onto the bucket (possibly front-compressed).
	 * @param frontCoding if true, front-coding is used over the buckets (before applying Gzip, if any)
	 * @param searchSupport build the support for the {@link SearchableDB} interface
	 * @param level compression level if compress=true. An integer between 0 and 9 (inclusive)
	 * @param log a PrintStream where to print log messages. A <i>null</i> value 
	 * will suppress any output message
	 * @throws IOException
	 * 
	 * @return A TextDB instance to access the built database.
	 */
	public static TextDB build(String inputfile, String outfile, int bucketSize, boolean compress, boolean frontCoding, boolean searchSupport, int level, PrintStream log) throws IOException {
		
		boolean verbose = log != null;
		
		if (verbose) {
			log.println(" >> Input file: " + inputfile);
			log.println(" >> Buckets size: " + bucketSize);
			
			if (compress) {
				log.println(" >> ZIP compression enabled");
				log.println(" >> Compression level: " + level);
			}
			
			if (frontCoding) 
				log.println(" >> FrontCoding enabled");
		}
		
		BufferedReader in = new BufferedReader(new FileReader(inputfile));
		ByteArrayOutputStream memBeginners = new ByteArrayOutputStream();
		ByteArrayOutputStream memJumpers = new ByteArrayOutputStream();
		
		Directory tdb = new Directory(outfile, Directory.WRITE_MODE);
		
		// the textdb-type file
		tdb.newFile("textdb-type");
		tdb.write(FrontCoding.class.getCanonicalName().getBytes());
		
		FastByteArrayOutputStream bucket = new FastByteArrayOutputStream();
		
		// dummy largest string
		String sLargest = new String(new char[] {BIGGER_CHAR, BIGGER_CHAR});

		if (verbose) log.println(" >> Analyzing the file");
		
		// It computes the info about dict size and num strings, so numBuckets
		long numStrings = 1;	// count the first dummy string
		for (String s = in.readLine(); s != null; s = in.readLine()) {
			numStrings++;
			if (s.compareTo(sLargest) == 0) throw new IOException("It does exist \uffff string");
		}
		in.close();

		int numBuckets = (int) Math.ceil(((float) numStrings) / bucketSize);

		long fileSize = (new File(inputfile)).length();
	
		// reopen the input file
		in = new BufferedReader(new FileReader(inputfile));
		
		String currS, oldS;
		numStrings = 0;
		int lcp = 0;
		oldS = "" + SMALLER_CHAR;	// the dummy string
		
		tdb.newFile("dict");
		
		// init the beginners & jumpers
		if (searchSupport) memBeginners.write((oldS + "\n").getBytes());
		memJumpers.write(Utils.intToBytes(0));
		
		if (frontCoding) bucket.write(0);	// fake lcp for the smaller dummy string
		bucket.write(1); 					// dummy suffix length = 1 byte
		bucket.write(SMALLER_CHAR);			// dummy suffix
		
		numStrings++;
		
		int maxLcp = lcp;
		int avgLcp = lcp;
		
		long maxBucketSize = 0;
		
		if (verbose) log.println(" >> Compressing the file");
		
		FastByteArrayOutputStream out = new FastByteArrayOutputStream(1024);
		
		int size = 0;
		for (currS = in.readLine(); currS != null; currS = in.readLine()) {
			
			if (numStrings % bucketSize == 0) { // Bucket start
				
				if (searchSupport) memBeginners.write((currS + "\n").getBytes());
				
				// finalize current bucket
				byte[] bc = bucket.array;
				size = bucket.length;
				
				if (compress) {
					Utils.fastZip(bc, 0, size, out, level);
					bc = out.array;
					size = out.length;
				}
				
				// reset the bucket
				bucket.reset();
				
				// write the previous bucket
				tdb.write(bc, 0, size);

				// write the starting position of the current one
				memJumpers.write(Utils.intToBytes((int)tdb.getFilePointer()));
				
				if (size >= maxBucketSize) maxBucketSize = size;
				
				oldS = "";
			}
			
			numStrings++;

			lcp = 0;
			
			if (frontCoding) {
				
				lcp = lcp(currS, oldS);
				
				// dropped length (bytes) + suffix + -1
				if (lcp < ESCAPE_VALUE_BYTE) {
					bucket.write((byte)lcp);
				} 
				else {
					bucket.write(ESCAPE_VALUE_BYTE);				// escape value for the lcp field
					bucket.write((byte)(lcp / ESCAPE_VALUE_BYTE));
					bucket.write((byte)(lcp % ESCAPE_VALUE_BYTE));
				}
			}
			
			if (lcp >= maxLcp) maxLcp = lcp;
			avgLcp += lcp;

			byte[] suffix = currS.substring(lcp).getBytes();
			
			// suffix length
			int suffixLen = suffix.length; 
			if (suffixLen < ESCAPE_VALUE_BYTE)
				bucket.write(suffix.length);
			else {
				bucket.write(ESCAPE_VALUE_BYTE);
				bucket.write((byte)(suffixLen / ESCAPE_VALUE_BYTE));
				bucket.write((byte)(suffixLen % ESCAPE_VALUE_BYTE));
			}
			
			// write the remaining suffix
			bucket.write(suffix);
			
			oldS = currS;
		}

		avgLcp /= numStrings;
			
		// Add a dummy largest string
		if (frontCoding) bucket.write(0);			// fake lcp
		bucket.write(sLargest.getBytes().length); 	// dummy largest string length
		bucket.write(sLargest.getBytes());			// dummy largest

		// Flush last bucket
		if (searchSupport) memBeginners.write((currS + "\n").getBytes());
		memJumpers.write(Utils.intToBytes((int)tdb.getFilePointer()));
		byte[] bc = bucket.array;
		size = bucket.length;
		if (compress) {
			Utils.fastZip(bc, 0, size, out, level);
			bc = out.array;
			size = out.length;
		}
		tdb.write(bc, 0, size);
		
		if (size >= maxBucketSize) maxBucketSize = size;
		
		if (searchSupport) {
			tdb.newFile("beginners");
			tdb.write(memBeginners.toByteArray());
		}
		
		tdb.newFile("jumpers");
		tdb.write(memJumpers.toByteArray());
		
		// Properties
		Properties prop = new Properties();
		prop.put("BUCKET_COMPRESSION", "" + compress);
		prop.put("FRONT_CODING", "" + frontCoding);
		prop.put("DICT_SIZE", "" + fileSize);
		prop.put("NUM_STRINGS", "" + (numStrings - 1));	// don't count the dummy string
		prop.put("NUM_BUCKETS", "" + numBuckets);
		prop.put("BUCKET_SIZE", "" + bucketSize);
		prop.put("MAX_BUCKET_SIZE", "" + maxBucketSize);
		
		tdb.newFile("properties");
		prop.store(tdb.getOutputStream(), "Developed by Paolo Ferragina and Claudio Corsi");
		
		// close everything
		bucket.reset();
		memBeginners.close();
		memJumpers.close();
		tdb.close();
		
		if (verbose) {
			
			log.println(" >> Original file size: " + Utils.formatSize(fileSize));
			log.println(" >> Number of buckets: " + numBuckets);
			log.println(" >> Number of strings: " + (numStrings-1));
			log.println(" >> Max LCP: " + maxLcp);
			log.println(" >> Avg LCP: " + avgLcp);
			
			long tot = tdb.sizeOf("beginners");
			log.println(" >> Beginners file size: " + Utils.formatSize(tot));
			
			tot += tdb.sizeOf("jumpers");
			log.println(" >> Jumpers file size: " + Utils.formatSize(tdb.sizeOf("jumpers")));
			
			tot += tdb.sizeOf("dict");
			log.println(" >> Compressed file size: " + Utils.formatSize(tdb.sizeOf("dict")));
			
			log.println(" >> Total compressed space: " + Utils.formatSize(tot));
			
			double ratio = tot / (double)fileSize;
			log.println(" >> Compression ratio: " + Utils.round(ratio, 100));
		}
		
		return TextDB.fromTDBFile(outfile);
	}

	/**
	 * It opens and loads from disk all data structures 
	 * needed to access the compressed TextDB. 
	 * 
	 * @throws IOException
	 */
	public void open() throws IOException {

		tdb = new Directory(filename, Directory.READ_MODE);

		// beginners
		ByteBuffer buff = ByteBuffer.allocate(0);
		
		if (searchSupport) {
			
			if (tdb.exist("beginners")) {
				buff = tdb.mapFile("beginners");
				buff.rewind();
			}
			else {
				searchSupport = false;
			}
		}
		
		
		Properties prop = new Properties();
		prop.load(tdb.getInputStream("properties"));

		zipped = Boolean.parseBoolean(prop.getProperty("BUCKET_COMPRESSION"));
		frontCoded = Boolean.parseBoolean(prop.getProperty("FRONT_CODING"));	
		dictSize = Long.parseLong(prop.getProperty("DICT_SIZE"));
		numStrings = Long.parseLong(prop.getProperty("NUM_STRINGS"));
		numBuckets = Integer.parseInt(prop.getProperty("NUM_BUCKETS"));
		bucketSize = Integer.parseInt(prop.getProperty("BUCKET_SIZE"));
		maxBucketSize = Long.parseLong(prop.getProperty("MAX_BUCKET_SIZE"));
		
		bucketBeginners = searchSupport ? new String[numBuckets + 1] : null;

		jumpers = new int[numBuckets + 1];
		block = new byte[(int)maxBucketSize];
		
		int i = 0;
		String s;
		
		if (searchSupport) {
			String big = new String(buff.array(), 0, buff.capacity());
			StringReader sr = new StringReader(big);
			BufferedReader inFCbegin = new BufferedReader(sr);
			
			for (s = inFCbegin.readLine(); s != null; s = inFCbegin.readLine()) {
				bucketBeginners[i] = s;
				i++;
			}
			
			// dummy rightmost bucket
			bucketBeginners[i] = "" + BIGGER_CHAR;
		}
		
		// load jumpers
		ByteBuffer jumpBuff = tdb.mapFile("jumpers");
		for(i = 0; i < jumpers.length - 1; i++) {
			jumpers[i] = jumpBuff.getInt();
		}
		
		// dummy rightmost bucket
		jumpers[i] = tdb.sizeOf("dict");
		
		content = tdb.mapFile("dict");
	}

	public void close() throws IOException {
		tdb.close();
	}
	
	/**
	 * Returns the position of the first record having prefix p (here records
	 * are assumed to be sorted), or the alphabetical position of p among 
	 * the records in the TextDB.
	 * 
	 * @param p The prefix to search.
	 * @return The position of (the first record having prefix) p.
	 * @throws IOException
	 */
	protected int locate(String p) throws IOException {
		return Math.abs(rank(p));
	}
	
	/**
	 * Returns the range [i, j) of positions identifying the records in 
	 * the (ordered) TextDB which are prefixed by p. If no such record does exist, 
	 * an empty range [i, i) is returned, where i is the alphabetic position 
	 * of p within the ordered TextDB.
	 * 
	 * @param p the prefix to search
	 * @return the range [i, j) represented in a Range object
	 */
	public Range prefix(String p) throws IOException {
		return new Range(locate(p), locate(p + BIGGER_CHAR));
	}
	
//	/**
//	 * @see #select(int[], int, int, int, PrintStream, boolean)
//	 * @deprecated
//	 */
////	@Deprecated
//	private String[] extract(int i, int j, int field, PrintStream out) throws IOException {
//
//		if ((i > j) || (i <= 0)) throw new IOException("Out of bounds input values: " + i + ", " + j);
//		
//		int start_bucket = (int)((i) / bucketSize);
//		int end_bucket = (int)((j-1) / bucketSize);
//		int span = end_bucket - start_bucket + 1; // how many buckets to span
//
//		int k = 0;
//		
//		String[] res = (out == null) ? new String[(int)(j - i)] : null;
//		
//		MutableString rec = new MutableString(1024);
//		MutableString suffix = new MutableString(1024);
//		
//		for(int bucketCounter = 0; bucketCounter < span; i = (bucketCounter + start_bucket) * bucketSize) {
//		
//			int currBucket = (int) ((i) / bucketSize);
//			
//			int startInBucket = (int)((i) % bucketSize);
//			int endInBucket = (int)bucketSize - 1;
//			if (bucketCounter + 1 == span) endInBucket = (int)((j-1) % bucketSize);
//			
//			// bucket loading...
//			int start_pos = jumpers[currBucket];
//			int end_pos = jumpers[currBucket + 1];
//			int size = (int)(end_pos - start_pos);
//
//			//byte[] c = new byte[size];
//			content.position(start_pos);
//			content.get(block, 0, size);
//			
//			byte[] b = gzipped ? Utils.fastUnzip(block, 0, size) : block;
//			
//			size = (gzipped) ? b.length : size;
//			
//			int count = 0;
//			
////			String sTmp = "";
////			StringBuilder sb = new StringBuilder();
//			
//			for (int pos = 0; (pos < size) && count <= endInBucket; count++) {
//				
//				// shared prefix of new string with tmpS
//				int lcp = frontCoded ? (int) b[pos++] : 0;
//				if (lcp == ESCAPE_VALUE_BYTE) { // represented on two bytes (assumed < 65000)
//					lcp = b[pos++] * ESCAPE_VALUE_BYTE;
//					lcp += b[pos++];
//				}
//	
//				// create the new string
//				//prefix = (lcp > 0 ? rec.substring(0, lcp) : prefix.length(0));
//				
//				// extract the suffix
//				suffix.length(0);
//				int skip = getSuffixSkip(b, pos, suffix);
//				
//				//String suffix = sb.toString();
//				
//				// advance the iterator to the following [lcp | len | suffix]
//				pos += skip;
//				
////				sTmp = prefix + suffix;
//				rec.length(lcp);	// start form 0 to the current lcp
//				rec.append(suffix);
//				
//				if (count >= startInBucket) {
//					
//					String f = Utils.getField(rec, field, fieldSeparator);
//					
//					if (out == null)
//						res[k++] = f;
//					else
//						out.println((f == null) ? "" : f);
//				}
//			}
//			
//			bucketCounter++;
//		}
//		
//		return res;
//	}

	/**
	 * It returns the rank r of the smallest record in the current TextDB
	 * which is larger than or equal to s. 
	 * Note that the rank is counted from 1, thus guaranteeing
	 * that if s is not found, the absolute value of the returned (negative) 
	 * result is the alphabetic position of s among the records in 
	 * the current TextDB.
	 * 
	 * @param s	The string to be ranked.
	 */
	public int rank(String s) throws IOException {

		// TODO improve this method using the MutableString
		
		if (s == null || s.startsWith("" + SMALLER_CHAR) || s.startsWith("" + BIGGER_CHAR)) throw new IOException("Illegal input string");
		
		// rightward check is useless being bucketBeginners[numBuckets] the largest

		// binary search over sampled strings
		int startIndex = Arrays.binarySearch(bucketBeginners, s);		
		if (startIndex < 0) startIndex = Math.abs(startIndex + 1) - 1; // previous beginner
		
		// absolute position of the bucket beginner
		int rank = startIndex * ((int) bucketSize);
		
		// candidate bucket delimiters
		int start = jumpers[startIndex];
		int end = jumpers[startIndex + 1]; // added dummy last bucket
		int size = (int)(end - start);
		
		content.position(start);
		content.get(block);
		
		byte[]  b = block;
		
		FastByteArrayOutputStream fbs = new FastByteArrayOutputStream((int)maxBucketSize);
		
		if (zipped) {
			Utils.fastUnzip(block, 0, size, fbs);
			b = fbs.array;
			size = fbs.length;
		}

		int j = 0;
		String sTmp = "";
		StringBuilder sb = new StringBuilder();
		for (; (j < size) && (s.compareTo(sTmp) > 0); ) {

			// shared prefix of new string with tmpS
			int lcp = frontCoded ? (int) b[j++] : 0;
			
			if (lcp == ESCAPE_VALUE_BYTE) { // represented on two bytes (assumed < 65000)
				lcp = b[j++] * ESCAPE_VALUE_BYTE;
				lcp += b[j++];
			}

			// create the new string to compare
			String prefix = (lcp > 0 ? sTmp.substring(0, lcp) : "");
			
			// extract the suffix
			sb.setLength(0);
			int skip = getSuffixSkip(b, j, sb);
			String suffix = sb.toString();
			
			// advance the iterator to the following [lcp | len | suffix]
			j += skip;
			
			sTmp = prefix + suffix;
			
			rank++;
		}
		
		rank--;

		if (s.compareTo(sTmp) == 0) {
			return rank;
		} 
		else 
			if (s.compareTo(sTmp) < 0) {
				return rank;
			} 
			else {
				return -(rank + 1); // s is just before the beginning of the next bucket
			}
	}

	public int size() {
		return (int)numStrings;
	}
	
	public String get(int record) throws IOException {
		String[] res = select(new int[] {record}, 0, 1, -1, null, false);
		if (res == null) return null;
		if (res.length == 0) throw new IOException("You have found a bug! Please contact the authors.");
		return res[0];
	}
	
	public String[] getRange(int i, int j) throws IOException {
		//return extract(i+1, j+2, -1, null);
		return select(null, i, j - i + 1, -1, null, true);
	}
	
	public void getRange(int i, int j, int field, BufferedWriter out) throws IOException {
		//extract(i+1, j+2, field, out);
		select(null, i, j - i + 1, field, out, true);
	}

	public String[] getSequential(int[] records) throws IOException {
		return select(records, 0, records.length, -1, null, false);
	}
	
	public String[] getSequential(int[] records, int pos, int length) throws IOException {
		return select(records, pos, length, -1, null, false);
	}
	
	public void getSequential(int[] records, int field, BufferedWriter out) throws IOException {
		select(records, 0, records.length, field, out, false);
	}
	
	/*
	 * records[] == array of 0-based rec IDs
	 * if range == true, records == null, pos = i, length = range size
	 */
	private String[] select(int[] records, int pos, int length, int field, BufferedWriter out, boolean range) throws IOException {
		
		if (range) {
			if (pos + 1 <= 0 || (pos + length) > numStrings) throw new IOException("Out of bound records");
		}
		else { 
			if (records[0] + 1 <= 0 || records[pos + length - 1] + 1 > numStrings) throw new IOException("Out of bound records");
		}
		
		String[] res = (out == null) ? new String[length] : null;
		MutableString lastOne = null;
		
		int k = 0;		// current position in res[]
		int lastRank = -1;
		int lastBucketNum = -1;
		int bucketNum = -1;
		byte[] b = null;
		int leftRanks = 0;
		int j = 0;
		int size = 0;
		int seenSoFar = 0;
		
		// init the buffers
		MutableString rec = new MutableString(1024);
		MutableString suffix = new MutableString(1024);
		
		FastByteArrayOutputStream fbs = (zipped) ? new FastByteArrayOutputStream((int)maxBucketSize) : null;

		for(int curr = pos; curr < pos + length; curr++) {
			
			int rank = (range) ? curr + 1 : records[curr] + 1;		// the rank to fetch (1-based)
			
			if (lastRank == rank) {		// in case of duplicate ranks
				
				if (out == null) 
					res[k++] = lastOne.toString();
				else {
					if (lastOne != null) out.write(lastOne.array(), 0, lastOne.length());
					out.write('\n');
				}
				
				continue;
			}
			
			lastRank = rank;
			lastOne = null;
			
			bucketNum = (int) (rank / bucketSize);	// the containing bucket
			
			// load a new bucket if needed
			if (bucketNum != lastBucketNum) {	
				
				// first string in bucket 0 is dummy
				leftRanks = (int) (rank % bucketSize) + 1;
				
				// candidate bucket delimiters
				int start = jumpers[bucketNum];
				int end = jumpers[bucketNum + 1];
				size = end - start;
		
				// load the bucket
				content.position(start);
				content.get(block, 0, size);
				
				b = block;
				
				if (zipped) {
					Utils.fastUnzip(block, 0, size, fbs);
					b = fbs.array;
					size = fbs.length;
				}
				
				rec.length(0);		// reset the current record
				seenSoFar = 0;
				
				lastBucketNum = bucketNum;
				j = 0;
			}
			else {
				leftRanks = (int)(rank % bucketSize) + 1  - seenSoFar;
			}
			
			for (; (j < size) && (leftRanks > 0); leftRanks--) {
	
				// shared prefix of new string with tmpS
				int lcp = frontCoded ? b[j++] : 0;
				
				if (lcp == ESCAPE_VALUE_BYTE) { // represented on two bytes (assumed < 65000)
					lcp = b[j++] * ESCAPE_VALUE_BYTE;
					lcp += b[j++];
				}
				
				// extract the suffix (the method resets the suffix buffer)
				int skip = getSuffixSkip(b, j, suffix);
				
				// advance the iterator to the following [lcp | len | suffix]
				j += skip;
				
				rec.length(lcp);
				rec.append(suffix);
				
				seenSoFar++;
			}
			
			lastOne = Utils.getField(rec, field, fieldSeparator);
			
			if (out == null)
				res[k++] = lastOne.toString();
			else {
				if (lastOne != null) out.write(lastOne.array(), 0, lastOne.length());
				out.write('\n');
			}
		}
		
		return res;
	}

	public static void main(String[] args) throws Exception {

		CmdLineParser parser = new CmdLineParser(); 
		CmdLineParser.Option help0 = parser.addBooleanOption('h', "help");
		CmdLineParser.Option bucketSize0 = parser.addIntegerOption('b', "bucket-size");
		CmdLineParser.Option compress0 = parser.addBooleanOption("compress");
		CmdLineParser.Option search0 = parser.addBooleanOption("search-support");
		CmdLineParser.Option skipFC0 = parser.addBooleanOption("no-front-coding");
		CmdLineParser.Option verbose0 = parser.addBooleanOption('v', "verbose");
		CmdLineParser.Option out0 = parser.addStringOption('o', "outfile");
		CmdLineParser.Option level0 = parser.addIntegerOption('l', "level");
		
		try {
			parser.parse(args);
		} 
		catch (CmdLineParser.OptionException e) {
			System.err.println(e.getMessage());
			printUsage();
			System.exit(1);
		}
		
		boolean help = (Boolean)parser.getOptionValue(help0, Boolean.FALSE);
		if (help) {
			printUsage();
			return;
		}
		
		boolean verbose = (Boolean)parser.getOptionValue(verbose0, Boolean.FALSE);
		boolean searchSupport = (Boolean)parser.getOptionValue(search0, Boolean.FALSE);
		boolean skipFC = (Boolean)parser.getOptionValue(skipFC0, Boolean.FALSE);
		
		Integer bucketSize = (Integer)parser.getOptionValue(bucketSize0);
		
		String inputfile = null;
		String[] otherArgs = parser.getRemainingArgs();
		if (otherArgs.length == 1) {
			// input file
			File f = new File(otherArgs[0]);
			if (!f.exists()) {
				System.err.println("Input file \"" + otherArgs[0] + "\" not found.");
				System.exit(1);
			}
			else {
				inputfile = otherArgs[0];
			}
		}
		else {
			printUsage();
			System.exit(1);
		}
		
		String outfile = (String)parser.getOptionValue(out0);
		if (outfile == null) outfile = inputfile + ".tdb";
		
		boolean compress = (Boolean)parser.getOptionValue(compress0, Boolean.FALSE);
		
		Integer level = (Integer)parser.getOptionValue(level0, DEFAULT_COMPRESSION_LEVEL);
		if (level < -1 || level > 9) {
			System.err.println("The compression level must be a value from 0 to 9.");
			System.exit(1);
		}
		
		// =================== START HERE ========================= //
		
		long start = System.currentTimeMillis();

		if (bucketSize != null)
			FrontCoding.build(inputfile, outfile, bucketSize, compress, !skipFC, searchSupport, level, verbose ? System.out : null);
		else
			FrontCoding.build(inputfile, outfile, DEFAULT_BUCKET_SIZE, compress, !skipFC, searchSupport, level, verbose ? System.out : null);
		
		if (verbose) System.out.println(" >> Done in " + Utils.elapsedTime(start, System.currentTimeMillis()));
	}
	
	private static void printUsage() {
		System.out.println();
		System.out.println(
				"Build a FrontCoding database. The input file is splitted in fixed-length\n" +
				"buckets of records encoded with the front coding technique.\n" +
				"Optionally the produced buckets can be compressed using gzip.");
		System.out.println();
		System.out.println("Input: the name of the sorted file to store.");
		System.out.println();
		System.out.println("Optional arguments:");
		System.out.println();
		System.out.printf("   %1$-30s   %2$s\n", "-h/--help", "Print this message.");
		System.out.printf("   %1$-30s   %2$s\n", "-v/--verbose", "Print log messages on stdout.");
		System.out.printf("   %1$-30s   %2$s\n", "-b/--bucket-size", "The maximum number of records to put into a single bucket.");
		System.out.printf("   %1$-30s   %2$s\n", "--search-support", "Build the data structures to allow searches into this database.");
		System.out.printf("   %1$-30s   %2$s\n", "--compress", "Apply zip over the buckets (front-coded by default).");
		System.out.printf("   %1$-30s   %2$s\n", "-l/--level", "Set the compression level to use (default is 6). Values range from -1 (faster) to 9 (better).");
		System.out.printf("   %1$-30s   %2$s\n", "--no-front-coding", "Don't front-code the buckets, just store them as-is.");
		System.out.printf("   %1$-30s   %2$s\n", "-o/--outfile <filename>", "Set the output file name. Default is '<inputfilename>.tdb'.");
		System.out.println();
	}
}
