/*
 * Copyright © 2006 - 2007 Claudio Corsi, Paolo Ferragina, Alessandro Barilari
 *
 * This file is part of SmallText library.
 *
 * SmallText is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SmallText is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package it.unipi.di.textdb;

import it.unimi.dsi.fastutil.objects.Object2IntOpenHashMap;
import it.unimi.dsi.fastutil.objects.ObjectIterator;
import it.unimi.dsi.mg4j.compression.Coder;
import it.unimi.dsi.mg4j.compression.Decoder;
import it.unimi.dsi.mg4j.compression.HuffmanCodec;
import it.unimi.dsi.mg4j.io.FastBufferedReader;
import it.unimi.dsi.mg4j.io.InputBitStream;
import it.unimi.dsi.mg4j.io.OutputBitStream;
import it.unimi.dsi.mg4j.util.MutableString;
import it.unipi.di.rs.IRankSelect;
import it.unipi.di.rs.Select;
import it.unipi.di.rs.SparseSelect;
import it.unipi.di.tokenizer.FixedTokenizer;
import it.unipi.di.tokenizer.TermTokenizer;
import it.unipi.di.tokenizer.Tokenizer;
import it.unipi.di.tokenizer.URLTokenizer;
import it.unipi.di.util.Directory;
import it.unipi.di.util.Utils;
import jargs.gnu.CmdLineParser;

import java.io.BufferedWriter;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintStream;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.util.Properties;
import java.util.Map.Entry;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;


/**
 * A {@link TextDB} that compress the source file with the Huffword
 * technique and access it through Rank and Select operations.
 * The compressed file is marked by arrays of bits delimiting the fields and the records.
 * At query time Rank&Select operations over these vectors will be performed in order to find the
 * positions of the requested fields/records to be uncompressed and returned.
 * <br>
 * The Huffword dictionary is built using a {@link Tokenizer} instance in order to split
 * the source text into terms. 
 * To each of them is assigned a codeword depending on its frequency in the text.
 * The Huffman prefix-free code is then used to encode that codewords.
 * 
 * @author Claudio Corsi
 * @author Paolo Ferragina
 * @author Alessandro Barilari
 *
 */
public class RSHuffword extends TextDB {
	
	public static final int DEFAULT_BUCKET_LENGTH = 100;
	
	private int[] dictOffsets;
	private MutableString dictBuffer;	// in memory dictionary (one single char[])
	
	private ByteBuffer content;
	
	private int numOfLines = 0;
	private Decoder huffDecoder;
	
	private Directory tdb;
	
	private int newLineCodeword = -1;
	private int newFieldCodeword = -1;
	
	private IRankSelect fieldIndexer;
	private IRankSelect recordsIndexer;
	
	private boolean withFields = true;
	
	private boolean isFSparse, isRSparse;
	
	
	
	/**
	 * Create a new RSHuffword object loading the needed data structures
	 * from the provided file.
	 * 
	 * @param filename the TDB file containing the needed data structures and the compressed content
	 */
	public RSHuffword(String filename) {
		super(filename);
	}
	

	/**
	 * Compress with a standard bucketed huffword technique. This method will use a {@link TermTokenizer} 
	 * to build the huffword dictionary.
	 * <br>
	 * The TermTokenizer use the default separator (the blank space character) to recognize fields.
	 * <br><br>
	 * See the static method {@link #build(Tokenizer, String, String, PrintStream, boolean, char)} to compress
	 * with customized parameters.
	 * 
	 * @param log a PrintStream where to print log messages. A null value will suppress any output message
	 */
	
	public TextDB build(String outfile, PrintStream log) throws IOException {
		TermTokenizer tokenizer = new TermTokenizer(filename, ' ');
		return RSHuffword.build(tokenizer, filename, outfile, log, false, '\t');
	}
	
	/**
	 * Compress the input file with the bucketed huffword technique using customized parameters.
	 * @param tokenizer the tokenizer used to parse the input file
	 * @param inputfile	the input file name
	 * @param log a PrintStream where to print log messages. A null value will suppress any output message
	 * @return the TextDB to access the built database (to be opened)
	 * @throws IOException
	 */
	public static TextDB build(Tokenizer tokenizer, String inputfile, String outfile, PrintStream log, boolean withFields, char separator) throws IOException {
		boolean verbose = log != null;
		
		if (verbose) {
			System.out.println(" >> Input file: " + inputfile);
			System.out.println(" >> Tokenizer: " + tokenizer.toString());
			System.out.println(" >> Using Huffman codec with R&S vectors");
			log.println(" >> Counting terms frequencies");
		}
		//Runtime run = Runtime.getRuntime();
		
		// collect the frequencies
		Object2IntOpenHashMap<MutableString> codewordsMap = new Object2IntOpenHashMap<MutableString>();
		codewordsMap.defaultReturnValue(-1);
		for(MutableString term = tokenizer.next(); term != null; term = tokenizer.next()) {
			int freq = codewordsMap.put(term, 1);
			if (freq != -1) codewordsMap.put(term, freq + 1);
		}
		
		//the textdb-type file
		Directory tdb = new Directory(outfile, Directory.WRITE_MODE);
		tdb.newFile("textdb-type");
		tdb.write(RSHuffword.class.getCanonicalName().getBytes());
		
		tdb.newFile("revdict");
		GZIPOutputStream gzip = new GZIPOutputStream(tdb.getOutputStream());
		OutputStreamWriter osw = new OutputStreamWriter(gzip);
		
		// build the codewords map
		int[] freq = new int[codewordsMap.size()];
		//the lengths of the dictionary terms
		int[] lengths = new int[codewordsMap.size()];
		
		
		ObjectIterator<Entry<MutableString, Integer>> iterator = codewordsMap.entrySet().iterator();
		
		Entry<MutableString, Integer> e;
		for (int codeword = 0; codeword < codewordsMap.size(); codeword++) 
		{
			e = iterator.next();
			MutableString term = e.getKey();
			
			freq[codeword] = e.getValue();
			e.setValue(codeword);
			lengths[codeword] = term.length();
			osw.write(term.array(), 0, term.length());
			//revdict.put(codeword, iterator.key());
		}
		iterator = null;
		osw.close();
		gzip.close();
		
		tdb.newFile("dict-term-lengths");
		for (int l : lengths) tdb.write(Utils.intToBytes(l));		// to compress it?
		lengths = null;
		System.gc();
		
		if (verbose) log.println(" >> Generating the Huffman tree");
		HuffmanCodec huffman = new HuffmanCodec(freq);
		Coder huffCoder = huffman.coder();
		
		if (verbose) log.println(" >> Saving the Huffman decoder");
		tdb.newFile("huffman.dec");
		ObjectOutputStream oos = new ObjectOutputStream(new GZIPOutputStream(tdb.getOutputStream()));
		oos.writeObject(huffman.decoder());
		oos.close();
		
		// write the codewords in the output file
		if (verbose) log.println(" >> Compressing file");

		tdb.newFile("content");
		OutputStream out = tdb.getOutputStream();
		
		//vectors R and F
		final int STREAM_LIMIT = 8192;
		
		File tempField = new File("fields");
		//tempField.deleteOnExit();
		FileOutputStream fosFields = new FileOutputStream(tempField);
		ByteArrayOutputStream bosFields = new ByteArrayOutputStream();
		OutputBitStream obsFields = new OutputBitStream(bosFields);
		
		File tempRecord = new File("records");
		tempRecord.deleteOnExit();
		FileOutputStream fosRecords = new FileOutputStream(tempRecord);
		ByteArrayOutputStream bosRecords = new ByteArrayOutputStream();
		OutputBitStream obsRecords = new OutputBitStream(bosRecords, STREAM_LIMIT);
		
		// write the bucket in a memory buffer
		ByteArrayOutputStream bucket = new ByteArrayOutputStream();
		OutputBitStream obs = new OutputBitStream(bucket);
		
		tokenizer.reset();
		
		//building the compressed document
		int numberOfLines = 0;
		int newLineCodeword = codewordsMap.get(new MutableString("\n"));
		int newFieldCodeword = -1;
		if( withFields ) {
			try {
				//TODO Fix this part when changing the separator to a MutableString
				newFieldCodeword = codewordsMap.get( new MutableString(separator) );
			} catch(Exception ex) {
				throw new IOException("Field separator term not present, choose another one or change a specific tokenizer");
			}
		}
		boolean newField = true, newRecord = true;
		long RNum1 = 1, FNum1 = 1;
		for(MutableString term = tokenizer.next(); term != null; term = tokenizer.next()) {
			int code = codewordsMap.get(term);

			int written = huffCoder.encode(code, obs);
			obsFields.writeBit(newField);
			written--;
			while( written > 0 ) {
				obsFields.writeBit(false);
				written--;
			}
			if( newField && withFields )
				obsRecords.writeBit(newRecord);
			newField = false;
			newRecord = false;
			
			if (code == newLineCodeword) {
				newRecord = true;
				newField = true;
				numberOfLines++;
				RNum1++;
				FNum1++;
			}
			else
				if( code == newFieldCodeword && withFields ) {
					newField = true; FNum1++; }
			
			if( withFields )
				if ( obsRecords.writtenBits() == STREAM_LIMIT ) {
					obsRecords.flush();
					fosRecords.write(bosRecords.toByteArray());
					bosRecords.reset();
				}
			
			if ( obs.writtenBits() % 512 == 0 ) {
				obs.flush();
				out.write(bucket.toByteArray());
				bucket.reset();
				
				obsFields.flush();
				fosFields.write(bosFields.toByteArray());
				bosFields.reset();
			}
		}
		
		if ( obs.writtenBits() != 0) {	// flush last bucket
			obs.flush();
			out.write(bucket.toByteArray());
			bucket.reset();
			obsFields.flush();
			fosFields.write(bosFields.toByteArray());
			bosFields.reset();
		}
		if( withFields )
			if (obsRecords.writtenBits() != 0 ) {
				obsRecords.flush();
				fosRecords.write(bosRecords.toByteArray());
				bosRecords.reset();
			}

		huffCoder.flush(obs);
		
		obsFields.close();
		obsRecords.close();
		obs.close();
		
		bosFields.close();
		bosRecords.close();
		fosFields.close();
		fosRecords.close();
		
		freq = null;
		long mapSize = codewordsMap.size();
		codewordsMap.clear();
		codewordsMap = null;
		huffman = null;
		//System.gc();
		
		//checking which IRankSelect indexers will be used
		long FSize = (new RandomAccessFile(tempField.getAbsolutePath(), "r")).length();
		long RSize = (new RandomAccessFile(tempRecord.getAbsolutePath(), "r")).length();
		boolean isFSparse, isRSparse;
		isFSparse = ( ( (double)FNum1 / (double)(FSize * 8) ) <= SparseSelect.DENSITY_LIMIT );
		isRSparse = ( ( (double)RNum1 / (double)(RSize * 8) ) <= SparseSelect.DENSITY_LIMIT );
		
		if (verbose) log.println(" >> Saving the support vectors");
		//.fields
		if( !isFSparse ) {
			System.out.println(" >> Fields...");
			long FFileSize = (new RandomAccessFile(tempField.getAbsolutePath(), "r")).length();
			int BUFF_DIM = 2048;
			while( FFileSize % BUFF_DIM != 0 ) BUFF_DIM--;
			tdb.newFile("fields");
			FileInputStream FFile = new FileInputStream(tempField.getAbsolutePath());
			OutputStream output = tdb.getOutputStream();
			byte[] buf = new byte[BUFF_DIM];
			while( FFile.read(buf) != -1 )
				output.write(buf);
			FFile.close();
		}
		//.records
		if( !isRSparse && withFields ) {
			System.out.println(" >> Records...");
			long RFileSize = (new RandomAccessFile(tempRecord.getAbsolutePath(), "r")).length();
			int BUFF_DIM = 2048;
			while( RFileSize % BUFF_DIM != 0 ) BUFF_DIM--;
			tdb.newFile("records");
			FileInputStream RFile = new FileInputStream(tempRecord.getAbsolutePath());
			OutputStream output = tdb.getOutputStream();
			byte[] buf = new byte[BUFF_DIM];
			while( RFile.read(buf) != -1 )
				output.write(buf);
			RFile.close();
		}
		
		if (verbose) log.println(" >> Building index for fields vector");
		IRankSelect RSFields;
		if( isFSparse ) {
			RSFields = new SparseSelect( tempField.getAbsolutePath() );
			(new File(SparseSelect.getVectorsFileNames(tempField.getAbsolutePath())[0])).deleteOnExit();
			(new File(SparseSelect.getVectorsFileNames(tempField.getAbsolutePath())[1])).deleteOnExit();
		}
		else
			RSFields = new Select( tempField.getAbsolutePath() );
		RSFields.createIndex(IRankSelect.MEMORY);
		RSFields.store(tdb);

		Properties prop = new Properties();
		long RSRecordsSize = 0;
		
		if( withFields ) {
			IRankSelect RSRecords;
			if (verbose) log.println(" >> Building index for records vector");
			if( isRSparse ) {
				RSRecords = new SparseSelect( tempRecord.getAbsolutePath() );
				(new File(SparseSelect.getVectorsFileNames(tempRecord.getAbsolutePath())[0])).deleteOnExit();
				(new File(SparseSelect.getVectorsFileNames(tempRecord.getAbsolutePath())[1])).deleteOnExit();
			}
			else
				RSRecords = new Select( tempRecord.getAbsolutePath() );
			RSRecords.createIndex(IRankSelect.MEMORY);
			RSRecords.store(tdb);
			
			prop.put("RECORDS_INDEXER", RSRecords.getClass().getName());
			
			RSRecordsSize = RSRecords.ISize() / 8;
		}
		
		prop.put("NEW_LINE_CODEWORD", "" + newLineCodeword);
		prop.put("NEW_FIELD_CODEWORD", "" + newFieldCodeword);
		prop.put("DICT_SIZE", "" + mapSize);
		prop.put("NUMBER_OF_LINES", "" + numberOfLines);
		prop.put("TOKENIZER", "" + tokenizer.toString());
		prop.put("WITH_FIELDS", String.valueOf(withFields));
		prop.put("FIELDS_INDEXER", RSFields.getClass().getName());
		prop.put("IS_FIELDS_VECTOR_SPARSE", String.valueOf(isFSparse));
		prop.put("IS_RECORDS_VECTOR_SPARSE", String.valueOf(isRSparse));
		
		tdb.newFile("properties");
		prop.store(tdb.getOutputStream(), "developed by Alessandro Barilari");
		
		if (verbose) {
			File f = new File(inputfile);
			long dictSize = tdb.sizeOf("revdict");
			int dictTermLengths = tdb.sizeOf("dict-term-lengths");
			
			log.println(" >> Number of lines: " + numberOfLines);
			log.println(" >> Dict size: " + mapSize);
			log.println(" >> Original file size: " + Utils.formatSize(f.length()));
			log.println(" >> Dict file size: " + Utils.formatSize(dictSize));
			log.println(" >> Term-lengths file size: " + Utils.formatSize(dictTermLengths));

			long huffSize = tdb.sizeOf("content");
			log.println(" >> Huff file size: " + Utils.formatSize(huffSize));
			
			long fieldsSize = tdb.sizeOf("fields");
			log.println(" >> Fields file size: " + Utils.formatSize(fieldsSize));
			
			long RSFieldsSize = RSFields.ISize() / 8;
			log.println(" >> Fields index size: " + Utils.formatSize(RSFieldsSize));
			
			long recordsSize = tdb.sizeOf("records");
			log.println(" >> Records file size: " + Utils.formatSize(recordsSize));
			
			log.println(" >> Records index size: " + Utils.formatSize(RSRecordsSize));
			
			long total = dictSize + huffSize + fieldsSize + recordsSize + RSFieldsSize + RSRecordsSize;
			
			File huffDec = new File("huffman.dec");
			log.println(" >> Huffman decoder file size: " + Utils.formatSize(huffDec.length()));
			total += huffDec.length();
			
			log.println(" >> Total compressed size: " + Utils.formatSize(total));
			
			double ratio = total / (double)f.length();
			log.println(" >> Compression ratio: " + Utils.round(ratio, 100));
		}
		
		out.close();
		
		// write on disk the .TDB file
		tdb.close();
		
		return TextDB.fromTDBFile(inputfile + ".tdb");
	}
	
	@SuppressWarnings("unchecked")
	public void open() throws IOException {
		
		// open the direcotry in read-only
		//System.out.println(" >> Opening the Directory");
		tdb = new Directory(filename, Directory.READ_MODE);
		
		// load the properties
		//System.out.println(" >> Loading properties");
		Properties prop = new Properties();
		
		//prop.load(new FileInputStream(basename + "-huffword.properties"));
		prop.load(tdb.getInputStream("properties"));
		
		withFields = Boolean.parseBoolean(prop.getProperty("WITH_FIELDS"));
		this.numOfLines = Integer.parseInt(prop.getProperty("NUMBER_OF_LINES"));
		this.newLineCodeword = Integer.parseInt(prop.getProperty("NEW_LINE_CODEWORD"));
		this.newFieldCodeword = Integer.parseInt(prop.getProperty("NEW_FIELD_CODEWORD"));
		isFSparse = Boolean.parseBoolean(prop.getProperty("IS_FIELDS_VECTOR_SPARSE"));
		isRSparse = Boolean.parseBoolean(prop.getProperty("IS_RECORDS_VECTOR_SPARSE"));
		
		
		try {
			// load the dict in memory
			//System.out.println(" >> Loading dictionary");
			InputStream is = tdb.getInputStream("revdict");
			GZIPInputStream gzip = new GZIPInputStream(is);
			FastBufferedReader fbr = new FastBufferedReader(new InputStreamReader(gzip));
			
			// collecting all the chars into one single memory buffer
			dictBuffer = new MutableString();							// is NOT compact
			
			char[] buff = new char[102400];   // 100KB
			for (int len = fbr.read(buff); len != -1; len = fbr.read(buff)) {
				dictBuffer.append(buff, 0, len);
			}
			
			fbr.close();
			gzip.close();
			
			dictBuffer.compact();		// now is COMPACT

			// loading the terms offsets....
			int size = tdb.sizeOf("dict-term-lengths") / 4;
			dictOffsets = new int[size + 1];
			ByteBuffer bb = tdb.mapFile("dict-term-lengths");
			bb.asIntBuffer().get(dictOffsets, 0, size);
			dictOffsets[size] = 0;
			
			int off = 0;
			for(int k = 0; k < dictOffsets.length; k++) {
				int tmp = dictOffsets[k]; 
				dictOffsets[k] = off;
				off += tmp;
			}
			
			// load the Decoder
			//System.out.println(" >> Loading decoder");
			ObjectInputStream ois = new ObjectInputStream(new GZIPInputStream(tdb.getInputStream("huffman.dec")));

			huffDecoder = (Decoder)ois.readObject();
			ois.close();
		}
		catch (ClassNotFoundException e) {
			throw new IOException(e.getMessage());
		}
		
		//System.out.println(" >> Loading indexes");
		if( isFSparse ) {
			String[] names = SparseSelect.getVectorsFileNames("fields");
			this.fieldIndexer = new SparseSelect(names[0], names[1], tdb);
		}
		else
			this.fieldIndexer = new Select("fields", tdb);
		this.fieldIndexer.load(tdb);
		
		
		if( withFields ) {
			if( isRSparse ) {
				String[] names = SparseSelect.getVectorsFileNames("records");
				this.recordsIndexer = new SparseSelect(names[0], names[1], tdb);
			}
			else
				this.recordsIndexer = new Select("records", tdb);
			this.recordsIndexer.load(tdb);
			
			int len = dictOffsets[newFieldCodeword + 1] - dictOffsets[newFieldCodeword];
			this.setFieldSeparator(String.valueOf(dictBuffer.array(), dictOffsets[newFieldCodeword], len));
		}
		
		content = tdb.mapFile("content");
	}
	

	/* (non-Javadoc)
	 * @see it.unipi.di.textdb.TextDB#close()
	 */
	@Override
	public void close() throws IOException {
		//dict = null;
		dictBuffer = null;
		tdb.close();
	}


	/* (non-Javadoc)
	 * @see it.unipi.di.textdb.TextDB#get(int)
	 */
	@Override
	public String get(int record) throws IOException {
		if (record < 0 || record >= numOfLines) return null;
		
		boolean last = ( record == numOfLines - 1 );
		
		long start, stop;
		if( withFields ) {
			start = this.fieldIndexer.select1(this.recordsIndexer.select1(record));
			stop = ( last ) ? ( content.limit() * 8 - 1 ) : this.fieldIndexer.select1(this.recordsIndexer.select1(record + 1));
		}
		else {
			start = this.fieldIndexer.select1(record);
			stop = ( last ) ? ( content.limit() * 8 - 1 ) : this.fieldIndexer.select1(record + 1);
		}
		
		//now we have to load bits [startingBit, firstOutBit)
		//int size = (int)( ( ( stop / 8 ) -  ( start / 8 ) ) + 1 );
		byte[] b = new byte[(int)( ( ( stop / 8 ) -  ( start / 8 ) ) + 1 )];
		
		content.position((int)(start / 8));
		content.get(b);
		
		//input
		ByteArrayInputStream bis = new ByteArrayInputStream(b);
		InputBitStream ibs = new InputBitStream(bis);
		
		//align to the first needed bit
		ibs.skip(start % 8);
		
		//output
		MutableString out = new MutableString("");
		
		int code = huffDecoder.decode(ibs);
		while( code != newLineCodeword ) {
			int len = dictOffsets[code + 1] - dictOffsets[code];
			out.append(dictBuffer.array(), dictOffsets[code], len);
			
			try {
				code = huffDecoder.decode(ibs);
			} catch( EOFException ex ) {
				break;
			}
		}
		return out.toString();
	}
	
	/**
	 * 
	 * Returns the field of a record, given their ordinal positions, or <i>null</i> 
	 * if one is not present. <br><br>
	 * 
	 * @param record the position of a record
	 * @param field the position of the field to be retrieved
	 * @return the requested field for that record
	 * @throws IOException
	 */
	@Override
	public String get(int record, int field) throws IOException {
		if (record < 0 || record >= numOfLines) return null;
		if( !withFields )
			return getField(get(record), field);
		else {
			long startingField = this.recordsIndexer.select1(record);
			long start = this.fieldIndexer.select1(startingField + field);
			long stop = this.fieldIndexer.select1(startingField + field + 1);
			
			//now we have to load bits [startingBit, firstOutBit)
			int size = (int)( (stop - start) / 8 );
			byte[] b = new byte[size + 2];
			content.position((int)(start / 8));
			content.get(b);
			
			//input
			ByteArrayInputStream bis = new ByteArrayInputStream(b);
			InputBitStream ibs = new InputBitStream(bis);
			
			//align to the first needed bit
			ibs.skip(start % 8);
			
			//output
			MutableString out = new MutableString("");
			
			int code = huffDecoder.decode(ibs);
			while( code != newFieldCodeword ) {
				int len = dictOffsets[code + 1] - dictOffsets[code];
				out.append(dictBuffer.array(), dictOffsets[code], len); 
				
				try {
					code = huffDecoder.decode(ibs);
				} catch( EOFException ex ) {
					break;
				}
			}
			return out.toString();
		}
	}

	/* (non-Javadoc)
	 * @see it.unipi.di.textdb.TextDB#getRange(int, int)
	 */
	@Override
	public String[] getRange(int i, int j) throws IOException {
		int numOfRecords = j - i  + 1;
		String[] res = new String[ numOfRecords ];
		
		MutableString out = new MutableString("");
		
		int lastValidRecord = Math.min(j, numOfLines - 1);
		boolean last = ( lastValidRecord == numOfLines - 1 );
		
		long startingField, firstOutField;
		if( withFields ) {
			startingField = this.recordsIndexer.select1(i);
			firstOutField = ( last ) ? 0 : this.recordsIndexer.select1(lastValidRecord + 1);
		}
		else {
			startingField = i;
			firstOutField = lastValidRecord + 1;
		}
		
		long start = this.fieldIndexer.select1(startingField);
		long stop = ( last ) ? content.limit() : this.fieldIndexer.select1(firstOutField);
		
		//now we have to load bits [startingBit, firstOutBit)
		int size = (int)( (stop - start) / 8 );
		byte[] b = new byte[size + 2];
		
		content.position((int)(start / 8));
		content.get(b);
		
		InputBitStream ibs = new InputBitStream(b);

		//align to the first needed bit
		ibs.skip(start % 8);
		
		for(int curr = 0, index = i; curr < numOfRecords; curr++, index++) {
			if ( index > lastValidRecord ) {
				res[curr] = null;
				continue;
			}
			
			out.setLength(0);	// reset the string buffer
			int code = huffDecoder.decode(ibs);
			while( code != newLineCodeword ) {
				try {
					int len = dictOffsets[code + 1] - dictOffsets[code];
					out.append(dictBuffer.array(), dictOffsets[code], len); 
					code = huffDecoder.decode(ibs);
				}
				catch(EOFException eof) {
					res[curr] = null;	// finish
					break;
				}
			}

			res[curr] = out.toString();
		}
		return res;
	}
	
	
	public void getRange(int i, int j, int field, BufferedWriter out) throws IOException {
		//TODO this method
	}
	
	

	/* (non-Javadoc)
	 * @see it.unipi.di.textdb.TextDB#getSequential(int[])
	 */
	@Override
	public String[] getSequential(int[] records) throws IOException {
		return getSequentialImplementation(records, 0, records.length, -1, null);
	}
	
	/* (non-Javadoc)
	 * @see it.unipi.di.textdb.TextDB#getSequential(int[], int, int)
	 */
	@Override
	public String[] getSequential(int[] records, int pos, int length) throws IOException {
		return getSequentialImplementation(records, pos, length, -1, null);
	}

	/* (non-Javadoc)
	 * @see it.unipi.di.textdb.TextDB#getSequential(int[], int, java.io.PrintStream)
	 */
	@Override
	public void getSequential(int[] records, int field, BufferedWriter out) throws IOException {
		getSequentialImplementation(records, 0, records.length, field, out);
	}

	/**
	 *  Given a <i>sorted</i> array of ranks it returns
	 *  an array of String containing all the lines in the compressed file
	 *  at the specified positions. Each rank should be in the range [0, N-1].
	 * <br><br>
	 *  In case of a too big rank value (greater than the number of lines in the compressed file)
	 *  a null will be placed in the corresponding position of the returned array.
	 *  <br><br>
	 *  If a not null PrintStream is passed it will be used to write the result instead to
	 *  collect it in a in-memory String array. This is useful when
	 *  the output have to be dumped on disk without consuming memory. In this case
	 *  a null reference will be returned.
	 * 
	 * @param records The list of ranks to extract 
	 * @param pos the starting position of the subrange
	 * @param length the length of the subrange
	 * @param field The fields to return in case of a not null PrintStream
	 * @param out The output stream used to write the extracted lines (can be null)
	 * @return A String array containing at position i the line in the compressed file at position ranks[i] or null if a not null PrintStream is provided
	 * @throws IOException
	 */
	private String[] getSequentialImplementation(int[] records, int pos, int length, int field, BufferedWriter out) throws IOException {
		if( pos + length > records.length ) throw new IOException("Bad offset!");
		String[] res = (out == null) ? new String[length] : null;
		
		int lastRank = -1;
		String lastOne = null;
		int lastValidRecord = Math.min(records[pos + length - 1], numOfLines - 1);

		for(int curr = pos; curr < ( pos + length ); curr++) {
			if (lastRank == records[curr]) {	// in case of duplicate ranks
				if (out == null) 
					res[curr] = lastOne;
				else
					out.write(getField(lastOne, field));
				continue;
			}
			
			lastRank = records[curr];
			lastOne = null;
			
			if (records[curr] > lastValidRecord) {
				if (out == null)  
					res[curr] = null;
				else
					out.write("");
				continue;
			}
			
			lastOne = get( records[curr] );
			
			if (out == null)
				res[curr] = lastOne;
			else
				out.write(getField(lastOne, field));
		}
		return res;
	}
	
	/* (non-Javadoc)
	 * @see it.unipi.di.textdb.TextDB#getNumberOfRecords()
	 */
	@Override
	public int size() {
		return numOfLines;
	}
	
	private static void printUsage() {
		System.out.println();
		System.out.println("Build a RSHuffword database. It compress the text using the huffword technique. The huffword dictionary will be " +
				"built using a TermTokenizer. This database use the Rank/Select technology to access the compressed data.");
		System.out.println();
		System.out.println("Input: the name of the file to store.");
		System.out.println();
		System.out.println("Optional arguments:");
		System.out.println();
		System.out.printf("   %1$-30s   %2$s\n", "-h/--help", "Print this message.");
		System.out.printf("   %1$-30s   %2$s\n", "-v/--verbose", "Print log messages on stdout.");
		System.out.printf("   %1$-30s   %2$s\n", "--url-tokenizer", "Use the URLTokenizer instead of the FieldTermTokenizer to build the huffword dictionary.");
		System.out.printf("   %1$-30s   %2$s\n", "--fixed-tokenizer", "Use the FixedTokenizer instead of the FieldTermTokenizer to build the huffword dictionary.");
		System.out.printf("   %1$-30s   %2$s\n", "--fixed-tokenizer-length <len>", "Set the length parameter of the FixedTokenizer.");
		System.out.printf("   %1$-30s   %2$s\n", "--with-fields", "Create index with random access support to fields (force the use of a FieldTermTokenizer)");
		System.out.printf("   %1$-30s   %2$s\n", "--field-separator <char>", "Set a custom field separator character. Default is '\t'.");
		System.out.printf("   %1$-30s   %2$s\n", "-o/--outfile <filename>", "Set the output file name. Default is '<inputfilename>.tdb'.");
		System.out.println();
	}
	
	public static void main(String[] args) throws Exception {
		CmdLineParser parser = new CmdLineParser(); 
		
		CmdLineParser.Option help0 = parser.addBooleanOption('h', "help");
		CmdLineParser.Option verbose0 = parser.addBooleanOption('v', "verbose");
		CmdLineParser.Option urlTokenizer0 = parser.addBooleanOption("url-tokenizer");
		CmdLineParser.Option fixedTokenizer0 = parser.addBooleanOption("fixed-tokenizer");
		CmdLineParser.Option fixedLen0 = parser.addIntegerOption("fixed-tokenizer-length");
		CmdLineParser.Option withFields0 = parser.addBooleanOption("with-fields");
		CmdLineParser.Option fieldSeparator0 = parser.addStringOption("field-separator");
		CmdLineParser.Option out0 = parser.addStringOption('o', "outfile");
		
		try {
			parser.parse(args);
		} 
		catch (CmdLineParser.OptionException e) {
			System.err.println(e.getMessage());
			printUsage();
			System.exit(1);
		}
		
		boolean help = (Boolean)parser.getOptionValue(help0, Boolean.FALSE);
		if (help) {
			printUsage();
			return;
		}
		
		String file = null;
		String[] otherArgs = parser.getRemainingArgs();
		
		if (otherArgs.length >= 1) {
			file = otherArgs[0];
			File f = new File(file);
			if (!f.exists()) {
				System.err.println("Input file \"" + otherArgs[0] + "\" not found.");
				System.exit(1);
			}
		}
		else {
			printUsage();
			System.exit(1);
		}
		
		boolean verbose = (Boolean)parser.getOptionValue(verbose0, Boolean.FALSE);
		boolean urlsTokenizer = (Boolean)parser.getOptionValue(urlTokenizer0, Boolean.FALSE);
		boolean fixedTokenizer = (Boolean)parser.getOptionValue(fixedTokenizer0, Boolean.FALSE);
		boolean useFields = (Boolean)parser.getOptionValue(withFields0, Boolean.FALSE);
		String fieldSeparator = (String)parser.getOptionValue(fieldSeparator0);
		
		if ( (fixedTokenizer && urlsTokenizer) || (fixedTokenizer && useFields) || (urlsTokenizer && useFields) ) {
			System.err.println("Incompatible tokenizer options");
			System.exit(1);
		}
		
		Tokenizer tokenizer = new TermTokenizer(file);
		char separator = '\t';
		if( fieldSeparator != null )
		{
			//MutableString separator = new MutableString(fieldSeparator);
			separator = (fieldSeparator).charAt(0);
			tokenizer = new TermTokenizer(file, separator);
		}
		if( urlsTokenizer ) tokenizer = new URLTokenizer(file);
		if( fixedTokenizer ) {
			int len = (Integer)parser.getOptionValue(fixedLen0, FixedTokenizer.DEFAULT_LENGTH);
			if (len <= 0) {
				System.err.println("The value for --fixed-tokenizer-length option must be > 0");
				System.exit(1);
			}
			tokenizer = new FixedTokenizer(file, len);
		}
		
		String outfile = (String)parser.getOptionValue(out0);
		if (outfile == null) outfile = file + ".tdb";
		
		RSHuffword.build(tokenizer, file, outfile, verbose ? System.out : null, useFields, separator);
		
	}

}
