/*
 * Copyright © 2006 - 2007 Claudio Corsi, Paolo Ferragina
 *
 * This file is part of SmallText library.
 *
 * SmallText is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SmallText is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package it.unipi.di.textdb;

import it.unimi.dsi.mg4j.util.MutableString;
import it.unipi.di.util.Directory;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.regex.Pattern;

/**
 * Consider a textual file consisting of <i>records</i> (i.e. lines separated 
 * by <i>new lines</i>). Each record is composed of a variable number of <i>fields</i> (i.e. 
 * strings separated by '\t'). Possibly, each field may be composed by multiple 
 * <i>values</i> separated by a char-sequence specified by the user. The number of values and the number 
 * of fields may differ among the records, thus generalizing the classical relational approach.<br><br> 
 * A TextDB stores this input file on disk in a compressed 
 * form and offers efficient access to individual records and fields. 
 * A record and a field can be identified by means of an ordinal 
 * position, counted from 0 and starting from the beginning of the file (for the records) or 
 * the beginning of the record (for the fields).
 * <br><br>
 * Implementations of this class provide different compression techniques and 
 * different accessing methods, thus offering time/space trade-offs.
 * 
 * <h1>A note on the build interface</h1>
 * This interface specifies a standard {@link #build(String, PrintStream)} method that must be 
 * provided by each implementation.
 * This method will build the data structures using default values for any 
 * custom parameter. In order
 * to build the TextDB with customized parameters, you need to 
 * define a static "build" method accepting parameters and 
 * performing all the work in order to produce on disk the permanent data structures. 
 * This is only a suggestion on how to develop a better implementation, but there are no constraints to 
 * respect it (except the good practice to respect a suggested standard).
 * 
 * <h1>The output file format (TDB)</h1>
 * During the build process, the provided TextDB implementations will write the compressed file 
 * content into a single file on disk in a format called TDB. 
 * The output file name on disk will be the file name passed to the constructor followed by the suffix ".tdb".
 * That file will contains all the data structures needed at run time in order to access the content
 * of the built TextDB. Being a single file, the generated TDB file is easy to share among different users.  
 * From the programmer point of view a TDB file can be loaded through the static method {@link #fromTDBFile(String)}
 * that load a TDB file and returns the correct instance of the stored TextDB without knowing nothing
 * about it. 
 * 
 * @author Claudio Corsi
 * @author Paolo Ferragina
 * 
 */
public abstract class TextDB {
	
	protected String filename;
	
	/**
	 * Creates a new TextDB from an input textual file.
	 * <br>
	 * The input file name is used
	 * by the implementing class to load all data structures
	 * from disk. If the TextDB has yet to be built, the file name will
	 * be used to create the necessary file(s).
	 * 
	 * <br><br>
	 * Implementation of this interface are encouraged to store their files into 
	 * one single file using a {@link Directory} object and to respect the TDB format.
	 * <br>
	 *  <b>NOTE: more informations should be added here about the TDB format!</b> 
	 * 
	 * @param filename the file containing this TextDB
	 */
	public TextDB(String filename) {
		this.filename = filename;
	}
	
	/**
	 * Returns a TextDB from a TDB file.
	 * 
	 * @param tdbFile the TDB file
	 * @return an instance of a TextDB corresponding to that file
	 * @throws IOException if the input file is not a valid TDB file or some I/O errors occur
	 */
	public static TextDB fromTDBFile(String tdbFile) throws IOException {
		Directory tdb = new Directory(tdbFile, Directory.READ_MODE);
		
		if (!tdb.exist("textdb-type")) throw new IOException("Not a valid TDB file."); 
		
		BufferedReader br = new BufferedReader(new InputStreamReader(tdb.getInputStream("textdb-type")));
		String clazz = br.readLine();
		br.close();
		
		try {
			 return (TextDB)Class.forName(clazz).getConstructor(String.class).newInstance(tdbFile);
		}
		catch(Exception e) {
			throw new IOException ("Unable to restore the TextDB: " + e.getMessage());
		}
	}
	
	/**
	 * Returns the name of this TextDB. <br><br>
	 * This is a <i>not unique</i> string that identify this TextDB instance and is
	 * computed over the name of the file storing this database.
	 * It is the filename without the file system path and without the 
	 * canonical ".tdb" suffix (if any). As the ".tdb" is just a convention
	 * for the file names containing the TextDBs, if it is not present 
	 * this method will return the full filename without the file system path.
	 * Be aware that different TextDBs can clash into the same name if they are
	 * stored into files named in the same way but located in different file system 
	 * location. To avoid this case we encourage the use of meaningful filenames.   
	 *
	 * <br><br>
	 * As an example consider the file "/home/user/example.tdb". It store a TextDB named "example".
	 *  
	 * @return the name of this TextDB instance
	 */
	public String getName() {
		File f = new File(filename);
		String filename = f.getName();
		return filename.endsWith(".tdb") ? filename.split(".tdb")[0] : filename;
	}
	
	/**
	 * The default separator '\t' for fields.
	 * 
	 */
	public static final String DEFAULT_FIELD_SEPARATOR = "\t";

	/**
	 * The char sequence used to separate the fields within a record.
	 *  
	 */
	protected MutableString fieldSeparator = new MutableString(DEFAULT_FIELD_SEPARATOR);
	
	/**
	 * The pattern used to split a record into fields.
	 * 
	 */
	private Pattern p = Pattern.compile(fieldSeparator.toString());
	
	/**
	 * Set the sequence of chars used to separate fields.
	 * 
	 * @param sep the new separator for fields
	 */
	public void setFieldSeparator(String sep) {
		this.fieldSeparator = new MutableString(sep);
		this.p = Pattern.compile(fieldSeparator.toString());
	}
	
	/**
	 * Splits the input record into fields, using the separator specified 
	 * with {@link #setFieldSeparator(String)}, 
	 * and returns the field at the specified position, or null if that position
	 * is out-of-bound.
	 * <br><br>
	 * The entire record is returned if -1 is specified as field position.
	 * 
	 * <br><br>
	 * NOTE: Descending class are encouraged to overwrite this method with 
	 * a faster implementation.
	 * 
	 * @param record the record to parse
	 * @param field the ordinal position of the field to return, or -1 to return the entire record 
	 * @return the requested field or the content of the entire record 
	 */
	protected String getField(String record, int field) {
		if (record == null) return null;
		if (field == -1) return record;
		String[] fields = p.split(record);
		return (field < fields.length) ? fields[field] : null;
	}
	
	/**
	 * Returns the values of a <i>multi-valued</i> field, where values are 
	 * separated by a <i>user-defined</i> separator.
	 * 
	 * <br><br>
	 * This implementation splits the field using a regular expression built
	 * over the separator string <i>sep</i>. Subclasses are encouraged to 
	 * overwrite this method with a more efficient one. 
	 * 
	 * @param field	the field content
	 * @param sep the separator used to separate the values composing this field
	 * @return an array containing the values of this field
	 */
	public String[] getFieldValues(String field, String sep) {
		return field.split(sep);
	}

	/**
	 * Returns all fields forming the input record.
	 * The fields are extracted through a linear scan of the record itself.
	 * <br><br>
	 * NOTE: Descending classes are encouraged to overwrite this method with a 
	 * faster implementation.	 
	 * @param record the record to parse
	 * @return an array of strings, one per field (order is preserved).
	 */
	public String[] getRecordFields(String record) {
		return p.split(record);
	}
	
	/**
	 * Returns the record for a given position in the range [0, N-1], 
	 * where N is the number of records present in the TextDB.
	 * 
	 * @param record a position in the range [0, N-1]
	 * @return the requested record
	 * @throws IOException
	 */
	public abstract String get(int record) throws IOException;
	
	/**
	 * 
	 * Returns the field of a record, given their ordinal positions, or <i>null</i> 
	 * if one is not present. <br><br>
	 * 
	 * @param record the position of a record
	 * @param field the position of the field to be retrieved
	 * @return the requested field for that record
	 * @throws IOException
	 */
	public String get(int record, int field) throws IOException {
		return getField(get(record), field);
	}
	
	/**
	 * Returns the records having positions from i to j in the TextDB.
	 *  
	 * @param i the starting position of the records to retrieve (inclusive)
	 * @param j the ending position of the records to retrieve (inclusive)
	 * @return the records in the defined range
	 * @throws IOException
	 */
	public abstract String[] getRange(int i, int j) throws IOException;
	
	
	/**
	 * Returns the specified field for the records in the range [i,j].
	 * If not present, a <i>null</i> value is
	 * stored in the corresponding position of the returned array.
	 * <br>
	 * The default implementation is based on a sequential scan of the
	 * fetched records.
	 * 
	 * @param i the starting position of the records to be fetched (included) 
	 * @param j the ending position of the records to be fetched (included)
	 * @param field the position of the field to return for all those records
	 * @return the field of the records in the range [i,j]
	 * @throws IOException
	 */
	public String[] getRange(int i, int j, int field) throws IOException {
		String[] range = getRange(i, j);
		String[] values = new String[range.length];
		for (int k = 0; k < range.length; k++) {
			values[k] = getField(range[k], field);
		}
		return values;
	}
	
	/**
	 * Print on the passed {@link PrintStream} the specified field for the records in the range [i,j].
	 * If not present, an empty line will be dumped out.
	 * 
	 * @param i the starting position of the records to be fetched (included) 
	 * @param j the ending position of the records to be fetched (included)
	 * @param field the position (counting from 0) of the field to return for all the records in range, or -1 to retrieve the entire record
	 * @param out the output BufferedWriter
	 * @throws IOException
	 */
	public abstract void getRange(int i, int j, int field, BufferedWriter out) throws IOException;
	
	/**
	 * Given a <i>sorted</i> array of record positions, this method
	 * returns all of them.
	 * <br><br>
	 * If some of the requested records are not available, the 
	 * behavior is unspecified and depend on the underlying implementation. 
	 * 
	 * @param records a <i>sorted</i> array of record positions
	 * @return the records having these positions (order is preserved)
	 * @throws IOException
	 */
	public String[] getSequential(int[] records) throws IOException {
		return getSequential(records, 0, records.length);
	}
	
	/**
	 * Given an array of record positions containing a <i>sorted</i> subrange
	 * defined by the parameters <code>pos</code> and <code>length</code>,
	 * this method returns the records for such positions.
	 * <br>
	 * <br>
	 * The fetched positions are the ones in the range <code>records[pos]</code>
	 * (included) to <code>records[pos+length]</code> (exluded). 
	 * 
	 * @param records array with a <i>sorted</i> subrange of records positions
	 * @param pos the starting position of the subrange
	 * @param length the length of the subrange
	 * @return the records having these positions (order is preserved)
	 * @throws IOException
	 */
	public abstract String[] getSequential(int[] records, int pos, int length) throws IOException;
	
	/**
	 * Given a <i>sorted</i> array of record positions and the position of a field, 
	 * this method returns that field of those records. 
	 * A <i>null</i> value is stored in an array entry whenever the 
	 * field is not present in the corresponding record. The input 
	 * order is respected in the output array.
	 * <br><br>
	 * The default implementation is based on a sequential scan of the fetched 
	 * records.
	 * 
	 * @param records a <i>sorted</i> array of record positions
	 * @param field	the field to select into each of these records
	 * @return the requested fields
	 * @throws IOException
	 */
	public String[] getSequential(int[] records, int field) throws IOException {
		
		String[] range = getSequential(records);
		
		String[] values = new String[range.length];
		for (int k = 0; k < range.length; k++) {
			values[k] = getField(range[k], field);
		}
		return values;
	}
	
	/**
	 * Given a <i>sorted</i> array of record positions and the position of a field,  
	 * this method retrieves the specified field from those records.
	 * If a record doesn't contain the requested field, the behavior of the method 
	 * depends on its implementation (implementing classes are encouraged to dump 
	 * a new line in this case, i.e. empty string).  
	 * <br>
	 * In order to dump <i>all</i> fields of the specified records, you have
	 * to input the integer -1 as field position.
	 * <br><br>
	 * The retrieved records are not kept in memory but immediately dumped on 
	 * the provided <i>PrintStream</i> without wasting further memory. 
	 * <br><br>
	 * NOTE: implementations can use the method {@link #getField(String, int)} 
	 * provided by
	 * this abstract class that selects a field of a record through a sequential access
	 * to the record itself. The use of a more efficient implementation of this function
	 * is encouraged.
	 * 
	 * @param records a <i>sorted</i> array of record positions
	 * @param field the position of the field to extract, or -1 to dump all fields
	 * @param out the output BufferedWriter
	 * @throws IOException
	 */
	public abstract void getSequential(int[] records, int field, BufferedWriter out) throws IOException;
	
	/**
	 * Opens the TextDB.
	 * <br>
	 * This method has to be called before any other operation on the TextDB.
	 * 
	 * @throws IOException
	 */
	public void open() throws IOException {}
	
	/**
	 * Closes the TextDB and releases all of its resources.
	 * 
	 * @throws IOException
	 */
	public void close() throws IOException {}
	
	/**
	 * Returns the number of records contained in this TextDB. 
	 * If N is the returned value then records of this database 
	 * are numbered from 0 to N-1.
	 * 
	 * @return the size of this TextDB as the number of the contained records
	 */
	public abstract int size();
	
	/**
	 * Builds the TextDB over the textual file identified by
	 * the <i>filename</i> string used in the constructor (see {@link #TextDB(String)}). 
	 * This method runs a build process with default values for all input parameters.
	 * 
	 * <br><br>
	 * Log messages will be dumped into the passed {@link PrintStream}, or suppressed
	 * if the passed reference is <i>null</i>.
	 * 
	 * @param log a PrintStream for log messages. A null value will suppress any output message
	 * @param outfile The output file name. 
	 * @throws IOException
	 * 
	 * @return A TextDB instance to access the built database.
	 */
	public abstract TextDB build(String outfile, PrintStream log) throws IOException;
	
	/**
	 * Builds a TextDB over the textual file identified by
	 * the <i>filename</i> string used in the constructor (see {@link #TextDB(String)}). 
	 * This method runs a build process with default values for all input parameters.
	 * 
	 * @param outfile The output file name
	 * 
	 * @throws IOException
	 * 
	 * @return A TextDB instance to access the built database.
	 */
	public TextDB build(String outfile) throws IOException {
		return build(outfile, null);
	}
}
