/*
 * Copyright © 2006 - 2007 Claudio Corsi, Paolo Ferragina
 *
 * This file is part of SmallText library.
 *
 * SmallText is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SmallText is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package it.unipi.di.textdb;

import it.unimi.dsi.mg4j.io.FastBufferedReader;
import it.unimi.dsi.mg4j.util.MutableString;
import it.unipi.di.util.Directory;
import it.unipi.di.util.Utils;
import jargs.gnu.CmdLineParser;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintStream;
import java.util.Properties;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import com.jcraft.jzlib.ZInputStream;
import com.jcraft.jzlib.ZOutputStream;

/**
 * A forward only cursor over a <i>Zip</i> compressed file. The implementation is based on 
 * the JCraft library, a 100% pure Java implementation of ZLib (see the <a href=http://www.jcraft.com/jzlib/index.html>JCraft home page</a>).
 * <br><br>
 * GZip format is also supported but, in this case, the compression and access is made using the standard
 * java.util.zip package.
 * <br><br>
 * NOTE: The JCraft library doesn't support the GZip format instead.
 * 
 * @author Claudio Corsi
 * @author Paolo Ferragina
 *
 */
public class ZipCursor extends TextDB {
	
	protected FastBufferedReader cursor;
	protected int currentPos = 0;
	protected MutableString currentRec = new MutableString(1024);
	protected boolean isOpen = false;
	protected boolean isZip = true;
	
	private int dbSize = 0;
	
	/**
	 * Creates an instance of a ZipFileCursor over a zipped (or gzipped) compressed file.
	 * 
	 * @param filename the TDB file containing the compressed file
	 */
	public ZipCursor(String filename) {
		super(filename);
	}
	
	public void open() throws IOException {
		
		isOpen = true;
		currentPos = 0;

		Directory tdb = new Directory(filename, Directory.READ_MODE);
		
		Properties prop = new Properties();
		prop.load(tdb.getInputStream("properties"));
		String format = prop.getProperty("FORMAT");
		isZip = (format.equals("zip")) ? true : false;
		dbSize = Integer.parseInt(prop.getProperty("DBSIZE"));
		
		InputStream is = tdb.getInputStream("content");
		InputStream zis = (isZip) ? new ZInputStream(is) : new GZIPInputStream(is);
		cursor = new FastBufferedReader(new InputStreamReader(zis));
	}
	
	/**
	 * Moves the cursor on the next record and returns it.
	 * A <i>null</i> value is returned when the cursor is over. 
	 * 
	 * @return the next String
	 * @throws IOException
	 */
	public String next() throws IOException {
		skip(currentPos + 1);
		return currentRec.toString();
	}
	
	/**
	 * Moves the cursor on a specified record and returns it. The new record
	 * position have to be greater or equals than the current one.
	 * A null value is returned when the cursor is over.
	 * 
	 * @param pos the position of the record where to move this cursor
	 * @return the String relative to the new position or null if the 
	 * cursor is over
	 * @throws IOException
	 */
	public String next(int pos) throws IOException {
		skip(pos);
		return currentRec.toString();
	}
	
	/** 
	 * Forwards this cursor to the record at position <i>pos</i>, which 
	 * must be positive and greater than or equal to the previous one.
	 * 
	 * @param pos the new (positive) position for this cursor
	 * @throws IOException
	 */ 
	protected void skip(int pos) throws IOException {

		if (!isOpen) throw new IOException("The cursor is not opened");
				
		if (pos == currentPos) return;
		
		if (pos < currentPos) throw new IOException("Rank " + pos + " is smaller than previous rank " + currentPos);
		
		int i = currentPos;
		for(; i < pos; i++) {
			if (cursor.readLine(currentRec) == null) break;	// over
		}
		currentPos = i;
	}
	
	/**
	 * Returns the current record position.
	 * 
	 * @return the current position of this cursor.
	 */
	public int getCurrentPos() {
		return currentPos;
	}
	
	/**
	 * Returns the content of the current record.
	 * 
	 * @return the content of the current record
	 */
	public String getCurrentValue() {
		return currentRec.toString();
	}
	
	public void close() throws IOException {
		cursor.close();
		isOpen = false;
	}

	/**
	 * Returns the String in position <i>pos</i> for this compressed TextDB. As this
	 * TextDB is a forward-only cursor, it is possible to retrieve
	 * only consecutive positions. In other words, the required position must be
	 * greater than or equal to the previous one. 
	 * 
	 */
	public String get(int record) throws IOException {
		return next(record);
	}

	public int size() {
		return dbSize;
	}

	public String[] getRange(int i, int j) throws IOException {
		String[] res = new String[j - i + 1];
		for (int k = 0; k < j - i + 1; k++) res[k] = next(i + k);
		return res;
	}

	@Override
	public void getRange(int i, int j, int field, BufferedWriter out) throws IOException {
		for (int k = 0; k < j - i + 1; k++) {
			String res = next(i + k);
			if (res != null) out.write(res);
			out.write('\n');
		}
	}

	public String[] getSequential(int[] records) throws IOException {
		return getSequential(records, 0, records.length);
	}
	
	public String[] getSequential(int[] records, int pos, int length) throws IOException {
		String[] res = new String[records.length];
		for(int k = pos; k < pos + length; k++) res[k] = next(records[k]);
		return res;
	}

	public void getSequential(int[] records, int field, BufferedWriter out) throws IOException {
		for(int k = 0; k < records.length; k++) {
			String res = next(records[k]);
			if (res != null) out.write(res);
			out.write('\n');
		}
	}

	/**
	 * Create a TextDB compressing the input file with Zip or optionally with GZip. The built database
	 * will be accessed in a sequential manner a la zcat.
	 * 
	 * @param filename the file to compress
	 * @param outfile the output file name
	 * @param useGZip if true compress with gzip instead of zip
	 * @return the TextDB to access the built database
	 * @throws IOException
	 */
	public static TextDB build(String filename, String outfile,  boolean useGZip) throws IOException {
		
		Directory tdb = new Directory(outfile , Directory.WRITE_MODE);
		
		// the textdb-type file
		tdb.newFile("textdb-type");
		tdb.write(ZipCursor.class.getCanonicalName().getBytes());
		
		tdb.newFile("content");
		OutputStream out = useGZip ? new GZIPOutputStream(tdb.getOutputStream(), 1024) : new ZOutputStream(tdb.getOutputStream(), 6);
		
		FileInputStream fis = new FileInputStream(filename);

		int dbSize = 0;
		
		byte[] buff = new byte[1024];
		char last = '\n';
		for(int read = fis.read(buff); read != -1; read = fis.read(buff)) {
			out.write(buff, 0, read);
			
			// count how many lines this buffer contains
			for(int k = 0; k < read; k++) if (buff[k] == '\n') dbSize++;
			last = (char)buff[read-1];
		}
		
		// in case the latest line doesn't end with a \n
		if (last != '\n') dbSize++;
		
		out.close();
		fis.close();
		
		tdb.newFile("properties");
		OutputStream os = tdb.getOutputStream();
		Properties prop = new Properties();
		prop.setProperty("FORMAT", useGZip ? "gzip" : "zip");
		prop.setProperty("DBSIZE", "" + dbSize);
		prop.store(os, "Created by Claudio Corsi");
		
		tdb.close();
		
		return TextDB.fromTDBFile(outfile);	
	}
	/**
	 * Create a TextDB compressing the input file with Zip. The built database
	 * will be accessed in a sequential manner a la zcat.
	 * 
	 * @param log this parameter is not used
	 * @param outfile the output file name
	 * 
	 * @throws IOException
	 * 
	 * @return A TextDB instance to access the built database.
	 */
	public TextDB build(String outfile, PrintStream log) throws IOException {
		return ZipCursor.build(filename, outfile, false);
	}
	
	public static void main(String[] args) throws Exception {

		CmdLineParser parser = new CmdLineParser(); 
		CmdLineParser.Option help0 = parser.addBooleanOption('h', "help");
		CmdLineParser.Option verbose0 = parser.addBooleanOption('v', "verbose");
		CmdLineParser.Option zip0 = parser.addBooleanOption('g', "gzip");
		CmdLineParser.Option out0 = parser.addStringOption('o', "outfile");
		
		try {
			parser.parse(args);
		} 
		catch (CmdLineParser.OptionException e) {
			System.err.println(e.getMessage());
			printUsage();
			System.exit(1);
		}
		
		boolean help = (Boolean)parser.getOptionValue(help0, Boolean.FALSE);
		if (help) {
			printUsage();
			return;
		}
		
		boolean verbose = (Boolean)parser.getOptionValue(verbose0, Boolean.FALSE);
		
		boolean gzip = (Boolean)parser.getOptionValue(zip0, Boolean.FALSE);
		
		String inputfile = null;
		String[] otherArgs = parser.getRemainingArgs();
		if (otherArgs.length == 1) {
			// input file
			File f = new File(otherArgs[0]);
			if (!f.exists()) {
				System.err.println("Input file \"" + otherArgs[0] + "\" not found.");
				System.exit(1);
			}
			else {
				inputfile = otherArgs[0];
			}
		}
		else {
			printUsage();
			System.exit(1);
		}
		
		String outfile = (String)parser.getOptionValue(out0);
		if (outfile == null) outfile = inputfile + ".tdb";
		
		// =================== START HERE ========================= //
		
		if (verbose) {
			System.out.println(" >> Compressor: " + ((gzip) ? "GZIP" : "ZIP"));
			System.out.print(" >> Deflating... ");
		}
		
		long start = System.currentTimeMillis();

		ZipCursor.build(inputfile, outfile, gzip);
		
		if (verbose) System.out.println("done in " + Utils.elapsedTime(start, System.currentTimeMillis()));
	}
	
	private static void printUsage() {
		System.out.println();
		System.out.println(
				"Build a ZipCursor database. The input file is entirely compressed\n" +
				"with zip (without any bucketing structure) and accessed sequentially\n" +
				"a la zcat. Optionally the gzip format can be used instead of zip.");
		System.out.println();
		System.out.println("Input: the name of the file to store.");
		System.out.println();
		System.out.println("Optional arguments:");
		System.out.println();
		System.out.printf("   %1$-30s   %2$s\n", "-h/--help", "Print this message.");
		System.out.printf("   %1$-30s   %2$s\n", "-v/--verbose", "Print log messages on stdout.");
		System.out.printf("   %1$-30s   %2$s\n", "-g/--gzip", "Compress using gzip instead of zip.");
		System.out.printf("   %1$-30s   %2$s\n", "-o/--outfile <filename>", "Set the output file name. Default is '<inputfilename>.tdb'.");
		System.out.println();
	}
}
