/*
 * Copyright © 2006 - 2007 Claudio Corsi, Paolo Ferragina
 *
 * This file is part of SmallText library.
 *
 * SmallText is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SmallText is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package it.unipi.di.tokenizer;

import it.unimi.dsi.mg4j.util.MutableString;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

/**
 * Splits a text into a list of fixed-size tokens. Token "\n" is added.
 * 
 * @author Claudio Corsi
 * @author Paolo Ferragina
 *
 */
public class FixedTokenizer extends TermTokenizer {
	
	public static final int DEFAULT_LENGTH = 4;
	
	private int length = DEFAULT_LENGTH; 
	
	public String toString() {
		return this.getClass().getName() + ", length: " + length;
	}
	
	/**
	 * Create a new FixedTokenizer object over the given file using
	 * the default length value of {@value #DEFAULT_LENGTH}.
	 * 
	 * @param file the file to split in tokens
	 * @throws IOException
	 */
	public FixedTokenizer(String file) throws IOException {
		super(file);
	}
	
	/**
	 * Create a new FixedTokenizer over the given file using a custom
	 * token length. 
	 * 
	 * @param file the file to split in tokens
	 * @param length the (maximum) length of the tokens in the number of chars.
	 * @throws IOException
	 */
	public FixedTokenizer(String file, int length) throws IOException {
		super(file);
		this.length = length;
	}

	@Override
	protected MutableString[] split(MutableString line) {
		
		List<MutableString> tokens = new ArrayList<MutableString>();
		
		boolean wsRun = false;
		int bufflength = 0;		
		
		for (int i = 0; i < line.length(); i++) {
			
			char ch = line.charAt(i);
			
			if (Character.isWhitespace(ch)) {
				term[bufflength++] = ch;
				wsRun = true;
				continue;
			}
			
			if (bufflength >= length || wsRun) {
				tokens.add(new MutableString(term, 0, bufflength));
				bufflength = 0;
			}
			
			term[bufflength++] = ch;	// next term
			wsRun = false;
		}
		
		// last one
		if (bufflength > 0) tokens.add(new MutableString(term, 0, bufflength));
		
		tokens.add(new MutableString("\n"));	// add "\n" as term
		
		return (MutableString[])tokens.toArray(new MutableString[0]);
	}
}
