/*
 * Copyright © 2006 - 2007 Claudio Corsi, Paolo Ferragina
 *
 * This file is part of SmallText library.
 *
 * SmallText is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SmallText is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package it.unipi.di.tokenizer;

import it.unimi.dsi.mg4j.io.FastBufferedReader;
import it.unimi.dsi.mg4j.util.MutableString;

import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;

/**
 * Splits a text into a list of tokens, '\n' is added. 
 * A token is a maximal sequence of alphanumeric
 * characters or other symbols. White-spaces are considered part of
 * the previous token.
 * <br>
 * 
 * @author Claudio Corsi
 * @author Paolo Ferragina
 *
 */
public class TermTokenizer implements Tokenizer {
	
	// FIXME add support to char[] as separator (instead of a single char separator)
	
	private String file;
	private FastBufferedReader br;
	private MutableString[] tokens;
	private int i = 0;
	private MutableString line = new MutableString();
	
	private char separator;
	private boolean detectSeparator = false;
	
	// static buffer allocation used to split lines in terms (max term length is 100.000 chars)
	protected static char[] term = new char[100000];	// 200KB in memory
	
	/**
	 * Create a new TermTokenizer over the given file.
	 * 
	 * @param file the file to split in tokens
	 * @throws IOException
	 */
	public TermTokenizer(String file) throws IOException {
		this.file = file;
		this.br = new FastBufferedReader(new FileReader(file));
	}
	
	/**
	 * Create a new TermTokenizer over the given file. The passed character will be
	 * detected as distinct token (if any).
	 * 
	 * @param file the file to split in tokens
	 * @param separator the character to detect as distinct token
	 * @throws IOException
	 */
	public TermTokenizer(String file, char separator) throws IOException {
		this.file = file;
		this.br = new FastBufferedReader(new FileReader(file));
		this.detectSeparator = true;
		this.separator = separator;
	}
	
	public String toString() {
		return this.getClass().getName();
	}
	
	protected MutableString[] split(MutableString line) {
		
		ArrayList<MutableString> tokens = new ArrayList<MutableString>();
		
		int length = 0;
		
		boolean isChar = true;
		boolean whitespaceSeen = false;
		
		for(int i = 0; i < line.length(); i++) {
			
			char ch = line.charAt(i);
			
			if (detectSeparator && ch == separator) {
				
				// add the previous term if any
				if (length > 0) {
					tokens.add(new MutableString(term, 0, length));
					length = 0;
				}
				
				// add the separation term
				term[length++] = ch;
				tokens.add(new MutableString(term, 0, length));
				length = 0;
				
				continue;
			}
			
			// always add white spaces to the previous term
			if (Character.isWhitespace(ch)) {
				term[length++] = ch;
				whitespaceSeen = true;
				continue;
			}
			
			if (Character.isLetter(ch) || Character.isDigit(ch)) {
				
				if (!isChar && length != 0 || whitespaceSeen) {
					tokens.add(new MutableString(term, 0, length)); // add the previous term
					length = 0;
				}
				
				term[length++] = ch;
				isChar = true;
			}
			else {
				
				if (isChar && length != 0 || whitespaceSeen) {
					tokens.add(new MutableString(term, 0, length)); // add the previous term
					length = 0;
				}
				
				term[length++] = ch;
				isChar = false;
			}
			
			whitespaceSeen = false;
		}
		
		if (length != 0) {
			tokens.add(new MutableString(term, 0, length)); // add latest term
			length = 0;
		}
		
		tokens.add(new MutableString("\n"));	// add "\n" as term
		
		return (MutableString[])tokens.toArray(new MutableString[0]);
	}
	
	public MutableString next() throws IOException {
		if (tokens == null) {
			i = 0;
			if (br.readLine(line) != null)
				tokens = split(line);
			else 
				return null;	// EOF
		}
		
		MutableString next = tokens[i++];
		if (i == tokens.length) tokens = null;
		
		return next;
	}
	
	public void reset() throws IOException {
		this.br = new FastBufferedReader(new FileReader(file));
	}
	
	public void close() throws IOException {
		br.close();
	}
}