/*
 * Copyright © 2006 - 2007 Claudio Corsi, Paolo Ferragina
 *
 * This file is part of SmallText library.
 *
 * SmallText is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SmallText is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package it.unipi.di.util;

import jargs.gnu.CmdLineParser;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintStream;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.channels.FileChannel.MapMode;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

/**
 * A virtual directory of files identified by unique names and physically 
 * contained into one single file on disk.
 * <br> 
 * Through instances of this class it is possible to add new files to a 
 * brand-new directory (write mode) or to access files in the directory (read 
 * only mode).  It is possible to manage only one 
 * file at a time (the current file). This makes this implementation not 
 * thread-safe. Simultaneous read/write accesses are not supported. Once built 
 * on disk, the directory cannot be updated. 
 * 
 * <h1>Building a new Directory</h1>
 * 
 * To create a new directory just instantiate a new object of this class using 
 * the {@link #WRITE_MODE} constant in the constructor. Then create a new file 
 * by using either the method {@link #newFile(String)} and
 * appending new content with the method {@link #write(byte[], int, int)}, or 
 * by getting an {@link OutputStream} through {@link #getOutputStream()}
 * and filling sequentially the file.
 * There is no need to close the current file when you're done, that is automatic
 * each time the method {@link #newFile(String)} is called or this directory is 
 * closed.
 * Once all files have been added to the new directory, call {@link #close()} in 
 * order to flush its content into the file specified in the constructor. 
 * 
 * <h1>Reading a Directory</h1>
 * 
 * Instantiate a new object of this class by specifying the file on disk which 
 * contains this directory and by using the {@link #READ_MODE} flag. To get 
 * the list of the available files in this virtual directory, use the 
 * method {@link #list()}. 
 * <br><br>
 * To access a given virtual file, use
 * the method {@link #getInputStream(String)} for a sequential access, 
 * or the methods {@link #readFileBlock(String, long, byte[])} and
 * {@link #mapFile(String)} for a random access.
 *  
 * @author Claudio Corsi
 * @author Paolo Ferragina
 *
 */
public class Directory {

	public final static int READ_MODE = 0;
	public final static int WRITE_MODE = 1;
	
	private final static int HEADER_OFFSET_LEN = 8;	// the length of the initial header offset
	
	private RandomAccessFile raf;
	private FileChannel ch;
	private int mode;
	private String currentFile;	// remember the last file being written
	private String filename;
	
	private Map<String, VirtualFile> directory = new HashMap<String, VirtualFile>();
	
	private class VFileOutputStream extends OutputStream {

		public VFileOutputStream(VirtualFile file) throws IOException {
			ch.position(file.pos);
		}
		
		@Override
		public void write(int b) throws IOException {
			ch.write(ByteBuffer.wrap(new byte[] {(byte)b}));
		}

		@Override
		public void write(byte[] b, int off, int len) throws IOException {
			ch.write(ByteBuffer.wrap(b, off, len));
		}
	}
	
	private class VFileInputStream extends InputStream {

		private int size, currPos;
		private ByteBuffer buff = ByteBuffer.allocate(1);
		
		public VFileInputStream(VirtualFile file) throws IOException {
			this.size = file.size;
			this.currPos = 0;
			ch.position(file.pos);
		}
		
		@Override
		public int read() throws IOException {
			if (currPos >= size) return -1;
			buff.rewind();
			currPos += ch.read(buff);
			buff.rewind();
			return buff.get() & 0xff; 
		}
		
		public long size() {
            return size;
        }
		
		public int available() {
			int avail = size - currPos;
			return avail > 0 ? avail : 0;
		}

		@Override
		public int read(byte[] b, int off, int len) throws IOException {
			if (currPos >= size) return -1;
			ByteBuffer buff = ByteBuffer.wrap(b, off, Math.min(len, size));
			int read = ch.read(buff);
			currPos += read;
			return read;
		}
	}
	
	/**
	 * Represents an entity (file) contained in this directory.
	 * 
	 * @author Claudio Corsi
	 *
	 */
	private class VirtualFile {
		private long pos;
		private int size;
		public VirtualFile(long pos, int size) {
			this.pos = pos;
			this.size = size;
		}
	}
	
	/**
	 * Creates a new virtual directory in read or write mode.
	 * 
	 * @param filename the file on disk storing the directory, or where 
	 * to put the newly created directory 
	 * @param mode the open modality
	 * @throws IOException
	 */
	public Directory(String filename, int mode) throws IOException {
		switch (mode) {
		case READ_MODE:
			if (!(new File(filename)).exists()) throw new IOException("The file \"" + filename + "\" doesn't exists");
			raf = new RandomAccessFile(filename, "r");
			loadHeader();
			break;
		case WRITE_MODE:
			raf = new RandomAccessFile(filename, "rw");
			raf.setLength(0);	// reset
			break;
		default:
			throw new IOException("Open mode not supported");
		}
		
		this.ch = raf.getChannel();
		this.mode = mode;
		this.filename = filename;
	}
	
	private void loadHeader() throws IOException {
		
		long offset = raf.readLong();
		raf.seek(offset);
		
		for(long currPos = raf.readLong(); currPos != -1; currPos = raf.readLong()) {

			// record: [(long) pos, (int) size, (int) len, (byte[]) <name>] | -1 as long
			
			int size = raf.readInt();
			int len = raf.readInt();
			byte[] buf = new byte[len];
			raf.read(buf);
			String name = new String(buf);
			VirtualFile f = new VirtualFile(currPos, size);
			directory.put(name, f);
		}
	}
	
	/**
	 * Maps a virtual file into memory.
	 * <br><br>
	 * This method provide a fast random access to virtual files, but
	 * it may present problems/bugs on certain JVMs.
	 * 
	 * @param file the virtual file to map
	 * @return a {@link ByteBuffer} object to access randomly the file content, or null 
	 * if the file doesn't exist
	 * @throws IOException
	 */
	public ByteBuffer mapFile(String file) throws IOException {
		VirtualFile f = directory.get(file);
		if (f == null) return null;
		currentFile = file;
		return ch.map(MapMode.READ_ONLY, f.pos, f.size);
	}
	
	/**
	 * Returns an {@link InputStream} to read a virtual file sequentially.
	 * A <i>null</i> reference is returned if the file doesn't exist in 
	 * this directory.
	 * 
	 * @param file the file to access
	 * @return an {@link InputStream} to the content of the file, or <i>null</i> if the file doesn't exist
	 * @throws IOException
	 */
	public InputStream getInputStream(String file) throws IOException {
		VirtualFile f = directory.get(file);
		if (f == null) return null;
		currentFile = file;
		return new VFileInputStream(f);
	}
	
	/**
	 * Read a portion of a virtual file belonging to this directory, and starting at 
	 * the given file position. If S is the result of the method
	 * {@link #sizeOf(String)} for this file, then the portion to be extracted
	 * must start in the range [0, S-1].
	 * This method tries to read up to <i>block.length</i> bytes. If there are less
	 * bytes to read, the remaining bytes will be unspecified. 
	 * <br><br>
	 * This method returns the number of bytes effectively read.
	 * 
	 * @param file the virtual file 
	 * @param pos the starting position of the block to be fetched
	 * @param block the buffer in which the block has to be loaded
	 * @return the number of bytes effectively read, or -1 if the file 
	 * doesn't exist or the starting position is negative or greater than 
	 * the file size
	 * @throws IOException
	 */
	public int readFileBlock(String file, long pos, byte[] block) throws IOException {
		VirtualFile f = directory.get(file);
		if (f == null || pos >= f.size || pos  < 0) return -1;
		currentFile = file;
		long realPos = f.pos + pos;
		ch.position(realPos);
		int len = Math.min(block.length, (int)(f.size - pos));
		ByteBuffer buff = ByteBuffer.wrap(block, 0, len);
		return ch.read(buff);
	}
	
	/**
	 * Lists the names of the virtual files contained in this directory.
	 * 
	 * @return the names of the virtual files
	 */
	public String[] list() {
		return directory.keySet().toArray(new String[0]);
	}
	
	/**
	 * Returns the size in byte of a virtual file, 0 if the file 
	 * doesn't exist.
	 * <br>
	 * This method does not change the directory current file. 
	 * 
	 * @param file the virtual file to measure
	 * @return the size of the file in bytes
	 */
	public int sizeOf(String file) {
		VirtualFile f = directory.get(file);
		if (f == null) return 0;
		return f.size;
	}
	
	/**
	 * Returns the total size of this directory in bytes.
	 *  
	 * @return the size of this directory in bytes
	 */
	public long size() throws IOException {
		return new File(filename).length();
	}
	
	/**
	 * Closes the directory. If it has been opened in {@link #WRITE_MODE} then
	 * this method writes on disk a file containing all the virtual files of this 
	 * directory and the auxiliary data 
	 * structures needed to efficiently access them. The file name is the one specified at 
	 * construction time.
	 * 
	 * @throws IOException
	 */
	public void close() throws IOException {
		
		// dump the header
		if (mode == WRITE_MODE) {
			
			// close previous file, if any
			if (currentFile != null) {
				VirtualFile f = directory.get(currentFile);
				f.size = (int)(ch.position() - f.pos);
			}
			currentFile = null;
			
			long offset = raf.getFilePointer();
			raf.seek(0);
			raf.writeLong(offset);
			raf.seek(offset);
			
			Set<Entry<String, VirtualFile>> set = directory.entrySet();
			
			// record: [(long) pos, (int) size, (int) len, (byte[]) <name>] | -1 as long
			
			for (Entry<String, VirtualFile> e : set) {
				
				String name = e.getKey();
				VirtualFile f = e.getValue();

				raf.writeLong(f.pos);
				raf.writeInt(f.size);
				raf.writeInt(name.getBytes().length);
				raf.write(name.getBytes());
			}
			
			raf.writeLong(-1);	// the ending long
		}
		
		raf.close();
	}
	
	/**
	 * Returns TRUE if the passed virtual file is contained into this directory. 
	 * 
	 * @param file the file name to check
	 * @return TRUE if the file is contained into this directory
	 */
	public boolean exist(String file) {
		return directory.containsKey(file);
	}
	
	/**
	 * Creates an empty file. The specified name must be 
	 * not already in use into this directory.
	 * 
	 * @param file the file to add in this directory
	 * @throws IOException
	 */
	public void newFile(String file) throws IOException {

		if (mode != WRITE_MODE) throw new IOException("Directory opened in read only mode");
		
		if (exist(file)) throw new IOException("The file already exists in this directory");
		
		// close previous file, if any
		if (currentFile != null) {
			VirtualFile f = directory.get(currentFile);
			f.size = (int)(ch.position() - f.pos);
		}
		
		// leave an empty space at the beginning of the file directory
		if (ch.position() == 0) ch.position(HEADER_OFFSET_LEN);
		
		// create an empty file
		VirtualFile f = new VirtualFile(ch.position(), 0);
		directory.put(file, f);
		
		currentFile = file;
	}
	
	/**
	 * Appends content to the current file from a given byte array. 
	 * 
	 * @see #newFile(String)
	 * 
	 * @param b the content to append (as a byte array)
	 * @param off the offset from which the copy must start
	 * @param len the number of bytes to write
	 * @throws IOException
	 */
	public void write(byte[] b, int off, int len) throws IOException {
		if (mode != WRITE_MODE) throw new IOException("Directory opened in read only mode");
		ch.write(ByteBuffer.wrap(b, off, len));
	}
	
	/**
	 * Appends content to the current file from a given byte array.
	 * 
	 * @see #newFile(String)
	 * 
	 * @param b the content to append (as an entire byte array)
	 * @throws IOException
	 */
	public void write(byte[] b) throws IOException {
		write(b, 0, b.length);
	}
	
	/**
	 * Returns an {@link OutputStream} to fill sequentially the content
	 * of the current file, or <i>null</i> if the current file doesn't exist
	 * (i.e. this directory is empty). Writing into a file using both the returned
	 * stream and the method {@link #write(byte[], int, int)} can generate 
	 * undefined results.
	 * 
	 * @see #newFile(String)
	 * 
	 * @return an {@link OutputStream} to write into the current file, or 
	 * <i>null</i> if the directory is empty
	 * @throws IOException
	 */
	public OutputStream getOutputStream() throws IOException {
		if (mode != WRITE_MODE) throw new IOException("Directory opened in read only mode");
		
		VirtualFile f = directory.get(currentFile);
		if (f == null) return null;
		
		return new VFileOutputStream(f);
	}
	
	/**
	 * Prints the file names and their 
	 * sizes (in byte) for the current directory.
	 * 
	 * @param out the {@link PrintStream} where to list this directory content
	 */
	public void ls(PrintStream out) {
		out.println("Total: " + dirSize() + " bytes (" + filename + ")");
		String[] list = list();
		for (String file : list) {
			out.printf("%1$12s bytes   %2$s\n", sizeOf(file), file);
		}
	}

	/**
	 * Print on a {@link PrintStream} the content of a file.
	 * 
	 * @param file the virtual file to print
	 * @param out the {@link PrintStream} where to print the file content
	 * @throws IOException if the file is not contained into this directory.
	 */
	public void cat(String file, PrintStream out) throws IOException {
		if (!exist(file)) throw new IOException("The file \"" + file + "\" doesn't exits.");
		InputStream in = getInputStream(file);
		BufferedReader br = new BufferedReader(new InputStreamReader(in));
		for(String line = br.readLine(); line != null; line = br.readLine()) out.println(line);
	}
	
	/**
	 * Returns the total size in bytes of this directory. This is the length of 
	 * the file where this directory is stored.  
	 * 
	 * @return the size of this directory
	 */
	public long dirSize() {
		return new File(filename).length();
	}

	/**
	 * Returns the actual position inside the current file 
	 * (both in read and write mode). 
	 *   
	 * @return the actual position inside the current file
	 * @throws IOException
	 */
	public long getFilePointer() throws IOException {
		VirtualFile f = directory.get(currentFile);
		if (f == null) return -1;
		return ch.position() - f.pos;
	}
	
	protected static void printUsage() {
		System.out.println();
		System.out.println("List the content of a TDB file.");
		System.out.println();
		System.out.println("Input: one or more TDB files.");
		System.out.println();
		System.out.println("Optional arguments:");
		System.out.println();
		System.out.printf("   %1$-30s   %2$s\n", "-h/--help", "Print this message.");
		System.out.printf("   %1$-30s   %2$s\n", "-c/--cat <filename>", "Print the content of a file contained into the TDB file.");
		System.out.println();
	}
	
	public static void main(String[] args) throws Exception {
		
		CmdLineParser parser = new CmdLineParser(); 
		
		CmdLineParser.Option help0 = parser.addBooleanOption('h', "help");
		CmdLineParser.Option cat0 = parser.addStringOption('c', "cat");
		
		try {
			parser.parse(args);
		} 
		catch (CmdLineParser.OptionException e) {
			System.err.println(e.getMessage());
			printUsage();
			System.exit(1);
		}
		
		boolean help = (Boolean)parser.getOptionValue(help0, Boolean.FALSE);
		if (help) {
			printUsage();
			return;
		}
		
		String[] tdbfiles = parser.getRemainingArgs();
		String fileToCat = (String)parser.getOptionValue(cat0);
		
		if (tdbfiles.length == 0) {
			printUsage();
			return;
		}
		
		for (String tdbfile : tdbfiles) {
			
			if (tdbfiles.length > 1) System.out.println("\n*** File " + tdbfile + "\n");
			
			Directory dir = new Directory(tdbfile, Directory.READ_MODE);
			
			if (fileToCat != null) {
				if (dir.exist(fileToCat)) 
					dir.cat(fileToCat, System.out);
				else {
					System.out.println();
					System.out.println("The file \"" + fileToCat + "\" doesn't exist.");
					System.out.println();
					dir.ls(System.out);
					System.out.println();
				}
			}
			else {
				dir.ls(System.out);
			}
		}
	}
}
