/*
 * Copyright © 2006 - 2007 Claudio Corsi, Paolo Ferragina
 *
 * This file is part of SmallText library.
 *
 * SmallText is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SmallText is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package it.unipi.di.util;

import it.unimi.dsi.mg4j.io.FastBufferedReader;
import it.unimi.dsi.mg4j.util.MutableString;
import jargs.gnu.CmdLineParser;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeMap;
import java.util.Map.Entry;

/**
 * A multi-way external merge sort that use a TreeMap (Java Red-Black Tree) as 
 * internal sorting data structure. This implementation is 100% pure Java and 
 * is able to scale over GBs of data better than the unix <code>sort</code> command.
 * <br><br>
 * This class has a main method so that it can be used via command line in a easy way.
 * Being a standard Java class it can be instantiated and configured by means of
 * its "set" methods. The sort process can be started by the {@link #run()} method.
 * By default this class reads from stdin and write in stdout. This behavior can
 * be changed using the {@link #setInFile(String)} and the {@link #setOutFile(String)}
 * method.
 * 
 * @author Claudio Corsi
 * @author Paolo Ferragina
 * 
 */
public class ExternalSort {
	
	public static final long DEFAULT_RUN_SIZE = (long)(50 * Math.pow(2, 20)); 	// 50MB
	public static final int DEFAULT_PAGE_SIZE = 102400; 	// 100KB
	
	public static final String VERSION = "0.1.3 - run!";
	
	protected boolean verbose = false;
	
	protected HashMap<Long, Reader> runsMap;	// # run -> file reader
	protected InputStream in = System.in;
	protected PrintStream out = System.out;
	protected String outfile;
	protected String infile;
	protected int[] columns = new int[0];
	protected char sep = '\t';
	protected long runSize = DEFAULT_RUN_SIZE;
	protected int pageSize = DEFAULT_PAGE_SIZE;
	protected long elapsedSecs = 0;
	
	protected long numberOfDumpedRows = 0;
	protected long numberOfInputRows = 0;
	
	protected boolean reverse = false;
	protected boolean numeric = false;
	protected boolean uniq = false;
	protected String currKey = null;
	protected boolean dist = false;
	protected String prevKey = null;
	protected long rowsCount = 0;
	
	protected boolean EOF = false;   // End of file while reading the current run
	protected MutableString buff = new MutableString(1024);  	// 1KB buffer size
	
	protected boolean extract = false;
	
	private TreeMap<String, Tuple> map;
	
	
	/** 
	 * Create a new ExternalSort.
	 * 
	 */
	public ExternalSort() {}
	
	/**
	 * A reverse comparator.
	 * 
	 * @author data
	 *
	 */
	public class ReverseComparator implements Comparator<String> {
		public final int compare(String o1, String o2) {
			return -o1.compareTo(o2);
		}
	}
	
	/**
	 * Used to store all the Strings associated to a sorting key with their run's ids.
	 * 
	 * @author Claudio Corsi
	 *
	 */
	protected class Tuple {

		public List<String> lines = new ArrayList<String>();
		public long[] run;
		
		public Tuple() {}
		
		public final void appendRun(long runNumber) {
			
			if (run == null) {
				run = new long[] { runNumber };
				return;
			}
			
			int size = run.length + 1;
			long[] tmp = new long[size];
			System.arraycopy(run, 0, tmp, 0, run.length);
			tmp[size - 1] = runNumber;
			run = tmp;
		}
		
		public void append(String x) {
			lines.add(x);
		}
	}
	
	/**
	 * Used to compare two strings by their sorting columns (aka sorted keys).
	 * 
	 * @author Claudio Corsi
	 *
	 */
	protected class SortingKey implements Comparable<SortingKey> {
		
		public String key, row;
		
		private boolean reverse = false;
		
		public SortingKey(String row, String key) {
			this.row = row;
			this.key = key;
		}
		
		public SortingKey(String row, String cols, boolean reverse) {
			this.row = row;
			this.key = cols;
			this.reverse = reverse;
		}
		
		public void setReverse(boolean reverse) {
			this.reverse = reverse;
		}
		
		public int length() { 
			return (key != row) ? key.length() + row.length() : row.length();
		}
		
		public final int compareTo(SortingKey k) {
			return (reverse) ? -key.compareTo(k.key) : key.compareTo(k.key);
		}
	}
	
	/**
	 * Set true to have log messages on stdout during the sorting process. 
	 * 
	 * @param verbose
	 */
	public void setVerbose(boolean verbose) {
		this.verbose = verbose;
		
		System.out.println(" >> Extsort version: " + VERSION);
	}

	/**
	 * Set true to sort in reverse order (ascendent instead of descendent).
	 * 
	 * @param reverse
	 */
	public void setReverse(boolean reverse) {
		if (reverse && verbose) System.out.println(" >> Reverse order enabled.");
		this.reverse = reverse;
	}
	
	/**
	 * Compare the sorting values (rows or columns) as numerical values. In other
	 * words this flag will cause the right padding of the sorting values. 
	 * 
	 * @param numeric
	 */
	public void setNumeric(boolean numeric) {
		this.numeric = numeric;
		if (verbose) System.out.println(" >> Numeric sort enabled.");
	}
	
	/**
	 * Instead to sort, dump the frequencies of the
	 * sorting keys in their sorted order (not in the frequency values order). 
	 * 
	 * @param dist
	 */
	public void setKeysDistribution(boolean dist) {
		this.dist = dist;
		if (verbose && dist) System.out.println(" >> Keys distribution enabled, dumping frequencies.");
	}
	
	/**
	 * Remove duplicates from the result. In case of equals sorting values (rows or columns) only
	 * one of these is kept. 
	 * 
	 * @param uniq
	 */
	public void setUniq(boolean uniq) {
		this.uniq = uniq;
		if (uniq && verbose) System.out.println(" >> Uniq enabled.");
	}
	
	/**
	 * Set the size of the chunk (run) of text to sort in memory at the first stage of 
	 * the algorithm. This is the maximum size of memory available to sort. If
	 * not set the default value is 50MB.
	 * 
	 * @param runSize the size of memory available expressed in bytes
	 */
	public void setRunSize(long runSize) {
		this.runSize = runSize;
	}

	/**
	 * Set the page size to use in the second stage of the algorithm (pagination of the sorted runs).
	 * This value should be chosen depending on the disk page size. The default value is 
	 * 100KB.
	 * 
	 * @param pageSize the page size expressed in bytes.
	 */
	public void setPageSize(int pageSize) {
		this.pageSize = pageSize;
	}
	
	/**
	 * Set the columns to sort. If not specified then the input file will be sorted considering the entire content
	 * of the rows (divided by new lines). If specified each row will be divided into columns and sorted
	 * accordingly to the value of the specified rows. The order of the columns matter. For example
	 * if the columns list is [1, 0, 5] then for first the column 1 of each row is compared. In case
	 * of the same value the column 0 is compared and, at the end, the column 5. 
	 * Two rows are considered equals if all the specified columns contain the same value
	 * 
	 * @see #setUniq(boolean)
	 * @see #setSeparator(char) 
	 * 
	 * @param columns the list of columns to sort
	 */
	public void setColumns(int[] columns) {
		this.columns = columns;
		
		if (verbose) {
			StringBuffer sb = new StringBuffer();
			for (int i = 0; i < columns.length; i++) sb.append("" + columns[i] + ", ");
			System.out.println(" >> Sorting by column(s): " + sb.toString().subSequence(0, sb.length() - 2));
		}
	}

	/**
	 * Set the character to use to split the rows in columns. Default value is tab ('\t').
	 *   
	 * @param sep
	 */
	public void setSeparator(char sep) {
		this.sep = sep;
		
		if (verbose) System.out.println(" >> Field separator: '" + sep + "'");
	}
	
	/**
	 * If true, dump out only the sorting column(s) omitting the other ones.
	 * The dumped column(s) will be sorted respecting the sorting parameters.
	 * If no columns are selected (sort by the entire rows) this option
	 * doesn't have effect. 
	 * 
	 * @param extract true to dump out only the sorting column(s)
	 */
	public void setExtract(boolean extract) {
		this.extract = extract;
		
		if (verbose) System.out.println(" >> Extract the sorting column(s).");
	}

	private void resetProgressInfos() {
		// reset some class value
	    elapsedSecs = 0;
	    numberOfDumpedRows = 0;
	    numberOfInputRows = 0;
	}
	
	/**
	 * Set the output file. By default the result is written in stdout.
	 * 
	 * @param outfile
	 * @throws FileNotFoundException
	 */
	public void setOutFile(String outfile) throws FileNotFoundException {
		this.outfile = outfile;
		
		if (verbose) System.out.println(" >> Output file: " + outfile);
	}
	
	/**
	 * Set the input file to sort. By default is the stdin.  
	 * 
	 * @param infile
	 * @throws FileNotFoundException
	 */
	public void setInFile(String infile) throws FileNotFoundException {
		this.infile = infile;
		this.in = new FileInputStream(infile);
	}
	
	protected void updateProgressInfos(long start) {
		long rowsPerSec = 0;
		elapsedSecs = (System.currentTimeMillis() - start) / 1000;
		if (elapsedSecs > 0) rowsPerSec = numberOfDumpedRows / elapsedSecs;
		System.out.println(" >> Input rows: " + numberOfInputRows + ", dumped rows: " + numberOfDumpedRows + " @ " + rowsPerSec + " rows/sec.");
	}
	
	/**
	 * Start the sorting process. This method can take much time to complete.
	 * 
	 * @throws IOException
	 */
	public void run() throws IOException {
		
		if (verbose) System.out.println(" >> Input file: " + infile == null ? "<stdin>" : infile);
		if (verbose) System.out.println(" >> Page size: " + Utils.formatSize(pageSize));
		if (verbose) System.out.println(" >> Run size : " + Utils.formatSize(runSize));
		if (verbose) System.out.println(" >> Creating the sorted runs...");
		
		FastBufferedReader fbr = new FastBufferedReader(new InputStreamReader(in));
		
		long start = System.currentTimeMillis();
		
		runsMap = new HashMap<Long, Reader>();

		List<SortingKey> chunk = new ArrayList<SortingKey>();
		
		long chunkCounter = 0;
		long tmpFileSize = 0;
		rowsCount = 0;
		
		while(!EOF) {
			
			int currChunkSize = 0;
			
			while(currChunkSize <= runSize) {
				
				if (fbr.readLine(buff) == null) {
					EOF = true;
					break;
				}
				
				String line = buff.toString();
				String key = Utils.getKey(line, columns, sep, numeric);
				SortingKey s = new SortingKey(line, key, reverse);
				chunk.add(s);
				
				rowsCount++;
				currChunkSize += buff.length();  // count the number of CHARS
			} 
			
			chunkCounter++;
			
			// a chunk has been loaded, let's sort it...
			File tmpFile = createSortedRun(chunk);
			
			runsMap.put(chunkCounter, new FastBufferedReader(new FileReader(tmpFile)));
			
			tmpFileSize += tmpFile.length();
			
			if (verbose) System.out.println(" >> Created sorted run " + chunkCounter + " (" + Utils.formatSize(tmpFile.length()) + " on disk for " + chunk.size() + " strings)");
			
			chunk.clear();
		}

		fbr.close();
		
		long stop = System.currentTimeMillis();
		
		if (verbose) System.out.println(" >> Generated " + chunkCounter + " sorted runs in " + Utils.elapsedTime(start, stop));
		if (verbose) System.out.println(" >> Sorted runs size  : " + tmpFileSize + " bytes.");
		if (verbose) System.out.println(" >> Original file size: " + new File(infile).length() + " bytes.");
		
		// Init the in-memory sorting data structure
		initDataStructure();
		
		resetProgressInfos();

		// Loading the first pages
		if (verbose) System.out.print(" >> Loading initial pages... ");
		Iterator<Long> iter = runsMap.keySet().iterator();
		while (iter.hasNext()) {
			Long runNumber = (Long)iter.next();
			loadNextPage(runNumber);
		}
		if (verbose) System.out.println("done.");
		
		start = System.currentTimeMillis();
		if (verbose) System.out.println(" >> Dumping sorted rows.");

		dumpSortedRows();
		
		stop = System.currentTimeMillis();
		if (verbose) System.out.println(" >> Dump complete in " + Utils.elapsedTime(start, stop));
		
		if (verbose) System.out.print(" >> Closing streams... ");
		for (Reader r : runsMap.values()) r.close();
		if (verbose) System.out.println("done.");
		
		if (outfile != null) {
			if (verbose) System.out.println(" >> Original file size: " + new File(infile).length() + " bytes.");
			if (verbose) System.out.println(" >> Sorted file size  : " + (new File(outfile).length()) + " bytes.");
		}
	}
	
	protected File createSortedRun(List<SortingKey> chunk) throws IOException {
		
		File tmp = File.createTempFile("run", ".txt");
		tmp.deleteOnExit();
		
		Collections.sort(chunk);

		// FIXME: this is synchronized, can we avoid it?
		BufferedWriter bw = new BufferedWriter(new FileWriter(tmp));
		
		for (SortingKey e : chunk) {
			bw.write(e.row);
			bw.write('\n');
		}
		
		bw.close();
		
		return tmp;
	}
	
	protected void initDataStructure() {
		if (reverse)
			map = new TreeMap<String, Tuple>(new ReverseComparator());
		else
			map = new TreeMap<String, Tuple>();
	}
	
	protected void loadNextPage(long runNumber) throws IOException {
		
		int currPageSize = 0;
		String key = null;
		
		FastBufferedReader reader = (FastBufferedReader)runsMap.get(runNumber);
		
		for (MutableString row = reader.readLine(buff); row != null; row = reader.readLine(buff)) {
			
			String line = buff.toString();
			
			numberOfInputRows++;
			
			currPageSize += buff.length();
			
			key = Utils.getKey(line, columns, sep, numeric);
			
			Tuple tuple = new Tuple();
			Tuple oldTuple = (Tuple)map.put(key, tuple);
			
			if (oldTuple != null) {
				tuple.run = oldTuple.run;
				tuple.lines = oldTuple.lines;
			}
			
			if (!uniq || (tuple.lines.size() == 0 && !key.equals(currKey))) 
				tuple.append(line);
			
			if (currPageSize >= pageSize) {
				tuple.appendRun(runNumber);
				break;
			}
		}
	}
	
	@SuppressWarnings("unchecked")
	protected void dumpSortedRows() throws IOException {
		
		if (outfile != null) out = new PrintStream(outfile);
		
		// FIXME: synchronized Writer! Use a not synch one to speed up the I/O 
		BufferedWriter bw = new BufferedWriter(new PrintWriter(out, false), 16384); // 16KB buffer size
		
		long start = System.currentTimeMillis();

		// count how many times a sorting key has been seen
		int freq = 0;
		
		while (!map.isEmpty()) {
			long[] toLoad = null;
			Set entrySet = map.entrySet();
			Iterator iter = entrySet.iterator();
			while (iter.hasNext()) {
				
				Entry<String, Tuple> entry = (Entry)iter.next();
				iter.remove();

				currKey = entry.getKey();
				prevKey = (prevKey == null) ? currKey : prevKey;
				
				Tuple tuple = (Tuple)entry.getValue();
				List<String> lines = tuple.lines;
				
				for(int i = 0; i < lines.size(); i++) {
					
					if (dist) {  // dump the distribution
						if (!currKey.equals(prevKey)) {
							String str = (numeric) ? Utils.trimLeftZeros(prevKey) : prevKey;
							bw.write(str);
							bw.write("\t");
							bw.write("" + freq);
							bw.write("\n");
							freq = 0;
							prevKey = currKey;
							numberOfDumpedRows++;
						}
						
						freq++;
					}
					else {
						if (extract)
							bw.write((numeric) ? Utils.trimLeftZeros(currKey) : currKey);
						else
							bw.write(lines.get(i));
						
						bw.write("\n");
						numberOfDumpedRows++;
					}
					
					if (verbose && (numberOfDumpedRows % 1000000) == 0 && numberOfDumpedRows != 0) 
						updateProgressInfos(start);
				}
				
				if (tuple.run != null) {
					toLoad = tuple.run;
					break;
				}
			}
			
			if (toLoad != null) {
				for (int i = 0; i < toLoad.length; i++) {
					loadNextPage(toLoad[i]);
				}
			}
		}
		
		if (dist) {
			String str = (numeric) ? Utils.trimLeftZeros(prevKey) : prevKey;
			bw.write(str);
			bw.write("\t");
			bw.write("" + freq);
			bw.write("\n");
			numberOfDumpedRows++;
		}
		
		bw.flush();
		
		// close if it is not the stdout
		if (out != System.out) out.close();
		
		if (verbose) {
			updateProgressInfos(start);
			System.out.println(" >> Dumped rows: " + numberOfDumpedRows);
		}
	}
	
	protected static void printUsage() {
		System.out.println();
		System.out.println("A multi-way external sort that scales over GBs of text.");
		System.out.println();
		System.out.println("Input: the name of the file to sort.");
		System.out.println();
		System.out.println("Optional arguments:");
		System.out.println();
		System.out.printf("   %1$-30s   %2$s\n", "-h/--help", "Print this message.");
		System.out.printf("   %1$-30s   %2$s\n", "-v/--verbose", "Print log messages on stdout.");
		System.out.printf("   %1$-30s   %2$s\n", "-m/--run-size <size>", "Set the maximum runs size. Default value is 50MB.");
		System.out.printf("   %1$-30s   %2$s\n", "-p/--page-size <size>", "Set the page size of the sorted runs. Default value 100KB.");
		System.out.printf("   %1$-30s   %2$s\n", "-o/--output-file <filename>", "Set the output filename.");
		System.out.printf("   %1$-30s   %2$s\n", "-r/--reverse", "Sort in descendant mode.");
		System.out.printf("   %1$-30s   %2$s\n", "-u/--uniq", "Removes duplicates from the sorted file.");
		System.out.printf("   %1$-30s   %2$s\n", "-n/--numeric", "Sorts by comparing the numerical values of the strings.");
		System.out.printf("   %1$-30s   %2$s\n", "-c/--columns <col[,col]>", "Set the sorting columns.");
		System.out.printf("   %1$-30s   %2$s\n", "-s/--spearator <char>", "Set the columns separator character (default value is \'\\t\').");
		System.out.printf("   %1$-30s   %2$s\n", "--dist", "Dump the frequency distribution of the sorting row (or column) strings.");
		System.out.printf("   %1$-30s   %2$s\n", "--extract", "Dump out only the sorting column(s) in the sorted order.");
		System.out.println();
	}
	
	public static void main(String[] args) throws Exception {
		
		CmdLineParser parser = new CmdLineParser(); 
		
		CmdLineParser.Option help0 = parser.addBooleanOption('h', "help");
		CmdLineParser.Option out0 = parser.addStringOption('o', "out");
		CmdLineParser.Option pageSize0 = parser.addIntegerOption('p', "page-size");
		CmdLineParser.Option runSize0 = parser.addLongOption('m', "run-size");
		CmdLineParser.Option cols0 = parser.addStringOption('c', "columns");
		CmdLineParser.Option sep0 = parser.addStringOption('s', "separator");
		CmdLineParser.Option rev0 = parser.addBooleanOption('r', "reverse");
		CmdLineParser.Option num0 = parser.addBooleanOption('n', "numeric");
		CmdLineParser.Option uniq0 = parser.addBooleanOption('u', "uniq");
		CmdLineParser.Option verbose0 = parser.addBooleanOption('v', "verbose");
		CmdLineParser.Option dist0 = parser.addBooleanOption("dist");
		CmdLineParser.Option extract0 = parser.addBooleanOption("extract");
		
		try {
			parser.parse(args);
		} 
		catch (CmdLineParser.OptionException e) {
			System.err.println(e.getMessage());
			printUsage();
			System.exit(1);
		}
		
		boolean help = (Boolean)parser.getOptionValue(help0, Boolean.FALSE);
		if (help) {
			printUsage();
			return;
		}
		
		ExternalSort sorter = new ExternalSort();
		
		boolean verbose = (Boolean)parser.getOptionValue(verbose0, Boolean.FALSE);
		
		if (verbose) sorter.setVerbose(verbose);
		
		boolean reverse = (Boolean)parser.getOptionValue(rev0, Boolean.FALSE);
		sorter.setReverse(reverse);
		
		String out = (String)parser.getOptionValue(out0);
		if (out != null) sorter.setOutFile(out);
		
		Integer pageSize = (Integer)parser.getOptionValue(pageSize0);
		if (pageSize != null) sorter.setPageSize(pageSize);
		
		Long size = (Long)parser.getOptionValue(runSize0);
		if (size != null) sorter.setRunSize(size);
		
		boolean uniq = (Boolean)parser.getOptionValue(uniq0, Boolean.FALSE);
		sorter.setUniq(uniq);
		
		boolean dist = (Boolean)parser.getOptionValue(dist0, Boolean.FALSE);
		sorter.setKeysDistribution(dist);
		
		boolean extract = (Boolean)parser.getOptionValue(extract0, Boolean.FALSE);
		if (extract && dist) {
			System.err.println("Incompatible options \"--dist\" and \"--extract\". Use only one of them.");
			System.exit(1);
		}
		if (extract) sorter.setExtract(extract);
		
		String fields1 = (String)parser.getOptionValue(cols0);
		if (fields1 != null) {
			String[] fields = fields1.split(",");

			int dup = 0;
			for (int i = 0; i < fields.length; i++) {
				for (int j = i + 1; j < fields.length; j++) {
					if (fields[i] != null && fields[i].equals(fields[j])) {
						fields[j] = null;
						dup++;
					}
				}
			}
			
			String[] fieldsNoDup = new String[fields.length - dup];
			for (int i = 0, j = 0; i < fields.length; i++) {
				if (fields[i] != null) {
					fieldsNoDup[j] = fields[i];
					j++;
				}
			}
			
			int[] fieldsArray = new int[fieldsNoDup.length];
			try {
				for (int i = 0; i < fieldsNoDup.length; i++) fieldsArray[i] = Integer.parseInt(fieldsNoDup[i]);
			}
			catch(NumberFormatException nex) {
				System.err.println("Illegal comma separated sequence of integers for option '-c'.");
				System.exit(1);
			}
			if (fieldsArray != null) sorter.setColumns(fieldsArray);
			
			if (dup != 0) {
				System.err.println(" >> WARNING: duplicates removed from '-f' list of values.");
			}
		}
		
		// Split over a single char!!!
		String separator = (String)parser.getOptionValue(sep0);
		if (separator != null) sorter.setSeparator(separator.charAt(0));
		
		Boolean numeric = (Boolean)parser.getOptionValue(num0, Boolean.FALSE);
		if (numeric) sorter.setNumeric(numeric);

		String[] otherArgs = parser.getRemainingArgs();
		if (otherArgs.length == 1) {
			// input file
			File f = new File(otherArgs[0]);
			if (!f.exists()) {
				System.err.println("Input file \"" + otherArgs[0] + "\" not found.");
				System.exit(1);
			}
			else {
				sorter.setInFile(otherArgs[0]);
			}
		}
		else {
			if (otherArgs.length > 1) {
				printUsage();
				System.exit(1);
			}
			// else use stdin
		}
		
		long start = System.currentTimeMillis();
		if (verbose) System.out.println(" >> Process started at " + new Date(start));
		
		try {
			sorter.run();
		}
		catch(Exception e) {
			e.printStackTrace(System.out);
		}
		
		if (verbose) {
			long stop = System.currentTimeMillis();
			System.out.println(" >> Process ended at " + new Date(stop));
			System.out.println(" >> Process overall time: " + Utils.elapsedTime(start, stop));
		}
	}
}
