/*
 * Copyright © 2006 - 2007 Claudio Corsi, Paolo Ferragina
 *
 * This file is part of SmallText library.
 *
 * SmallText is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SmallText is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package it.unipi.di.util;

/**
 * Defines a range [i, j], whose bounds are included.
 * 
 * @author Claudio Corsi
 * @author Paolo Ferragina
 *
 */
public class Range implements Comparable<Range> {

	private int i, j;
	
	/**
	 * Defines the range [i, j] (bounds included). If j < i then the 
	 * bounds are swapped and the range [j, i] is defined.
	 * 
	 * @param i lower bound
	 * @param j upper bound
	 */
	public Range(int i, int j) {
		 
		if (j < i) {
			this.i = j;
			this.j = i;
		}
		else {
			this.i = i;
			this.j = j;
		}
	}
	
	public int getLowerBound() {
		return i;
	}
	
	public int getUpperBound() {
		return j;
	}
	
	/**
	 * The length of this range.
	 * 
	 * @return the length of this range.
	 */
	public int length() {
		return Math.abs(j - i) + 1;
	}

	/**
	 * Returns -1 if this range is before range r, +1 if it is after range r, or 
	 * 0 if they are overlapping.
	 * 
	 */
	public int compareTo(Range r) {
		if (r.j < i) return 1;
		if (r.i > j) return -1;
		return 0;
	}
	
	/**
	 * Check if the input point is contained in this range.
	 * 
	 * @param p the point to check.
	 * @return true if the input point is contained in this range.
	 */
	public boolean containsPoint(int p) {
		return (i <= p && p <= j);
	}
	
	public String toString() {
		return "[" + i + ", " + j + "]";
	}
	
	public static Range parseRange(String str) {
		String[] bounds = str.replace("[", "").replace("]", "").split(",");
		int i = Integer.parseInt(bounds[0].trim());
		int j = Integer.parseInt(bounds[1].trim());
		return new Range(i, j);
	}
}
