/*
 * Copyright © 2006 - 2007 Claudio Corsi, Paolo Ferragina
 *
 * This file is part of SmallText library.
 *
 * SmallText is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * SmallText is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package it.unipi.di.util;

import it.unimi.dsi.mg4j.util.Fast;
import it.unipi.di.textdb.ZipCursor;
import it.unipi.di.textdb.TextDB;

import jargs.gnu.CmdLineParser;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Random;
import java.util.Set;

/**
 * Utility class to test a TextDB (just for debug).
 * 
 * @author Claudio Corsi
 * @author Paolo Ferragina
 *
 */
public class TextDBTester {
	
	private static int OUT_BUFFER_SIZE = (int)Math.pow(2, 20);	// 1MB
	
	protected static void printUsage() {
		System.out.println();
		System.out.println("Test the access time of a TextDB (just for debug).");
		System.out.println();
		System.out.printf("   %1$-30s   %2$s\n", "-h/--help", "Print this message.");
		System.out.printf("   %1$-30s   %2$s\n", "-v/--verbose", "Enable human readable log messages");
		System.out.printf("   %1$-30s   %2$s\n", "-s/--size", "Test set size");
		System.out.printf("   %1$-30s   %2$s\n", "-t/--type", "Test type: \"rand\", \"seq\" or \"range\"");
		System.out.printf("   %1$-30s   %2$s\n", "-c/--cooler", "The file to use to warm down the system");
		System.out.printf("   %1$-30s   %2$s\n", "-e/--num-exp", "The number of times the test will be repeted (> 3). Default is 10.");
		System.out.println();
	}
	
	public static void main(String[] args) throws Exception {
		
		CmdLineParser parser = new CmdLineParser(); 
		
		CmdLineParser.Option help0 = parser.addBooleanOption('h', "help");
		CmdLineParser.Option verbose0 = parser.addBooleanOption('v', "verbose");
		CmdLineParser.Option size0 = parser.addIntegerOption('s', "size");
		CmdLineParser.Option type0 = parser.addStringOption('t', "type");
		CmdLineParser.Option cooler0 = parser.addStringOption('c', "cooler");
		CmdLineParser.Option numOfExp0 = parser.addIntegerOption('e', "num-exp");
		
		try {
			parser.parse(args);
		} 
		catch (CmdLineParser.OptionException e) {
			System.err.println(e.getMessage());
			printUsage();
			System.exit(1);
		}
		
		boolean help = (Boolean)parser.getOptionValue(help0, Boolean.FALSE);
		
		if (help) {
			printUsage();
			return;
		}
		
		boolean verbose = (Boolean)parser.getOptionValue(verbose0, Boolean.FALSE);
		
		int times = (Integer)parser.getOptionValue(size0, 0);
		
		if (times == 0) {
			System.err.println("The size option must be > 0");
			System.exit(1);
		}
		
		String type = (String)parser.getOptionValue(type0, "");
		Set<String> types = new HashSet<String>();
		types.add("rand");types.add("seq");types.add("range");
		
		if (!types.contains(type)) {
			System.err.println("Type must be one of the following: rand, seq or range");
			System.exit(1);
		}
		
		String coolingFile = (String)parser.getOptionValue(cooler0, "");
		
		String[] otherArgs = parser.getRemainingArgs();
		
		String tdbFile = otherArgs[0];
		File f = new File(tdbFile);
		if (!f.exists()) {
			System.err.println("Input file \"" + otherArgs[0] + "\" not found.");
			System.exit(1);
		}
		
		int numOfExp = (Integer)parser.getOptionValue(numOfExp0, 10);
		if (numOfExp < 4) {
			System.err.println("The value for \"-e/--num-exp\" must be > 3");
			System.exit(1);
		}
		
		//////////////// START HERE ///////////////////
		
		
		long start = 0;
		long elapsed = 0;
		
		TextDB db = TextDB.fromTDBFile(tdbFile);
		
		if (verbose) {
			System.out.println(" >> TextDB: " + tdbFile);
			System.out.println(" >> Records to fetch: " + times);
			System.out.println(" >> Type of test: " + type);
			System.out.println(" >> Compressor: " + db.getClass().getCanonicalName());
			System.out.println(" >> Cooling file: " + coolingFile);
			System.out.print(" >> Opening database... ");
		}
		else {
			System.out.println(
					"# Columns Description:\n"+
					"#   * record set size\n" +
					"#   * bytes/msecs cold\n" + 
					"#   * bytes/msecs warm\n" +
					"#   * str/msecs cold\n" +
					"#   * str/msecs warm\n" +
					"#   * elapsed time in msecs cold\n" +
					"#   * elapsed time in msecs warm\n" +
					"#   * memory occupation\n" +
					"#   * database name\n");
		}

		start = System.currentTimeMillis();
		
		Fast.compactMemory();
		long mem = Fast.availableMemory();
		
		db.open();
		
		elapsed = System.currentTimeMillis() - start;
		
		if (verbose) System.out.println("done in " + elapsed + " msecs (" + Utils.elapsedTime(elapsed) + ")");
		
		Fast.compactMemory();
		mem -= Fast.availableMemory();
		
		if (verbose) {
			System.out.println(" >> Memory usage: " + Utils.formatSize(mem) + " (" + mem + " bytes)");
			System.out.println(" >> Avail mem: " + Fast.percAvailableMemory() + "%");
		}
		
		
		int maxVal = db.size();
		if (verbose) System.out.println(" >> DB size: " + maxVal);
		
		
		int[] positions = null;
		if( !type.equals("range") ) {
			File testFile = new File("test.dat");
			
			int prevTestSize = (int)testFile.length() / 4;
			
			File prop = new File("test.txt");
			String prevTdbFile = null;
			if (prop.exists()) {
				BufferedReader br = new BufferedReader(new FileReader(prop));
				prevTdbFile = br.readLine();
				br.close();
			}
			
			if (testFile.exists() && prevTestSize == times && !tdbFile.equals(prevTdbFile)) {
				
				if (verbose) System.out.print(" >> Loading a previous test... ");
				
				int size = (int)testFile.length() / 4;
				positions = new int[size];
				FileInputStream fo = new FileInputStream(testFile);
				FileChannel ch = fo.getChannel();
				ByteBuffer buff = ByteBuffer.allocate((int)testFile.length());
				ch.read(buff);
				buff.rewind();
				int k = 0;
				for(int i = buff.getInt(); ; i = buff.getInt()) {
					positions[k++] = i;
					if (buff.position() == buff.capacity()) break;
				}
				fo.close();
				
				if (verbose) System.out.println("done. (" + k + " positions)");
			}
			else {
				// fill the random ranks vector
				if (verbose) System.out.print(" >> Generating the array of random positions... ");
				start = System.currentTimeMillis();
				Random rnd = new Random();

				positions = new int[times];
				for(int i = 0; i < times; i++) {
					positions[i] = rnd.nextInt(maxVal);
				}
				elapsed = System.currentTimeMillis() - start;
				
				if (verbose) System.out.println("done in " + elapsed + " msecs (" + Utils.elapsedTime(elapsed) + ")");

				if (verbose) System.out.print(" >> Saving current test in \"test.dat\"... ");
				FileOutputStream fo = new FileOutputStream(testFile);
				for (int i = 0; i < positions.length; i++) {
					fo.write(Utils.intToBytes(positions[i]));
				}
				fo.close();
				if (verbose) System.out.println("done.");
			}

			if (type.equals("seq")) {
				if (verbose) System.out.println(" >> Sorting positions to perform a sequential test");
				Arrays.sort(positions);
			}
		}
		else
			positions = new int[times];
		
		File result = new File("result.txt");
		
		if (verbose) System.out.println(" >> Running the test...");

		String[] performance = new String[9]; 
		performance[0] = "" + times;
		
		// start the test
		long overall = 0;
		double avgThr = 0;
		double avgThrByte = 0;
		
		int testToSkip = 3;
		
		for(int n = 0; n < numOfExp; n++) {
			
			BufferedWriter out = new BufferedWriter(new FileWriter(result), OUT_BUFFER_SIZE);
			
			start = System.currentTimeMillis();
			
			if (type.equals("range")) {
				Random rnd = new Random();				
				
				int begin = rnd.nextInt(maxVal - times + 1);
				int end = begin + times - 1;
				
				db.getRange(begin, end, -1, out);
			}
			else {
				if (type.equals("seq"))
					db.getSequential(positions, -1, out);
				else 	// rand
					for(int i = 0; i < times; i++) {
						String rec = db.get(positions[i]); 
						if (rec != null) out.write(rec); 
						out.write('\n');
					}
			}
			elapsed = System.currentTimeMillis() - start;
			
			overall += (n < testToSkip) ? 0 : elapsed;	// skip the cold tests
			
			out.close();
			
			// statistics...
			double t = positions.length / (float)elapsed;
			double throughput = Utils.round(t, 100);
			avgThr += (n < testToSkip) ? 0 : t;
			
			t = result.length() / (float)elapsed;
			double throughputByte = Utils.round(t, 10);
			avgThrByte += (n < testToSkip) ? 0 : t;
			
			String mode = (n == 0) ? "cold" : "warm";
			
			if (verbose) System.out.println(" >>   * \"" + mode  + "\" test: " + positions.length + " strings in " + elapsed + " msecs (" + Utils.elapsedTime(elapsed) + "), throughput: " + throughput + " strings/msecs, " + throughputByte + " bytes/msec");
			
			if (n == 0) {
				performance[1] = "" + throughputByte;
				performance[3] = "" + throughput;
				performance[5] = "" + elapsed; 
				performance[7] = Utils.formatSize(mem);
				performance[8] = db.getName();
			}
			
			// reset the cursor
			if (db instanceof ZipCursor) {
				db.close();
				db.open();
			}
		}
		
		numOfExp -= testToSkip;
		double throughput = Utils.round(avgThr/(float)numOfExp, 100);
		double throughputByte = Utils.round(avgThrByte/(float)numOfExp, 10);
		long elapsedAVG = (long)Utils.round(overall / (float)numOfExp, 100);
		
		if (verbose) System.out.println(" >>   * \"AVG warm\" test: " + positions.length + " strings in " + elapsedAVG + " msecs (" + Utils.elapsedTime(elapsedAVG) + "), AVG throughput: " + throughput + " strings/msecs, " + throughputByte + " bytes/msec");
		
		// store the AVG warm test only
		performance[2] = "" + throughputByte;
		performance[4] = "" + throughput;
		performance[6] = "" + elapsedAVG;
		
		if (!verbose) {
			for (String val : performance) {
				System.out.print(val + "\t");
			}
			System.out.println();
		}
		
		// always warm down before exit
		warmDown(coolingFile, verbose);
		
		File prop = new File("test.txt");
		BufferedWriter bw = new BufferedWriter(new FileWriter(prop));
		bw.write(tdbFile);
		bw.close();
	}
	
	private static void warmDown(String coolingFile, boolean verbose) throws IOException {
		
		if (!new File(coolingFile).exists()) {
			if (verbose) System.out.println(" >> Leaving the system hot!");
			return;
		}
			
		if (verbose) System.out.print(" >> Cooling the system... 0% ");
		FileInputStream fis = new FileInputStream(coolingFile);
		FileChannel ch = fis.getChannel();
		ByteBuffer buff = ByteBuffer.allocate((int)Math.pow(2, 20) * 100); // 100MB
		OutputStream os = new FileOutputStream("/dev/null");
		int percLimit = 10;
		for(ch.read(buff); ch.position() < ch.size(); ch.read(buff)) {
			
			int perc = (int)(ch.position() / (float)ch.size() * 100);
			
			if (perc >= percLimit) {
				if (verbose) System.out.print(percLimit + "% ");
				percLimit += 10;
			}
			buff.rewind();
			os.write(buff.array());
			buff.rewind();
		}
		os.close();
		fis.close();
		ch.close();
		if (verbose) System.out.println("100%, done.");
	}
}
